#define PAM_SM_AUTH

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <syslog.h>
#include <stdarg.h>
#include <utmp.h>
#include <time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#ifdef sun
#include <security/pam_appl.h>
#endif
#include <security/pam_modules.h>

#include <security/pam_appdata.h>

#include "api_adsap2.h"

#define PAM_DEBUG_ARG      1
#define PAM_FORCE_ARG      2
#define PAM_USE_FIRST_PASS 4
#define PAM_TRY_FIRST_PASS 8

#define _pam_forget(X) if (X) {memset(X, 0, strlen(X));free(X);X = NULL;}

/* internal data */
static char *pam_module_name = "pam_adsap2";

/* we need to save these from open_session to close_session, since
 * when close_session will be called we won't be root anymore and
 * won't be able to access again the radius server configuration file
 * -- cristiang */

static struct pam_appl_data *get_cyc_appl(pam_handle_t *pamh)
{
  struct pam_conv *conv;
  struct pam_appl_data *pappl;

  if (pam_get_item(pamh, PAM_CONV, (const void **)(void *) &conv) != PAM_SUCCESS) { //[GY] 2005/Sep/19  clean warning
    return(NULL);
  }

  pappl = (struct pam_appl_data *) conv->appdata_ptr;

  if (!pappl || pappl->id != 0x1234fedc) {
    return(NULL);
  }

  return(pappl);
}

/* logging */
static void _pam_log(int err, char *format, ...)
{
    va_list args;
    char buffer[1024];

    va_start(args, format);
    vsprintf(buffer, format, args);
    /* don't do openlog or closelog, but put our name in to be friendly */
    syslog(err, "%s: %s", pam_module_name, buffer);
    va_end(args);
}

/* argument parsing */
static int _pam_parse(int argc, const char **argv)
{
  int ctrl=0;

  /*
   *  If either is not there, then we can't parse anything.
   */
  if ((argc == 0) || (argv == NULL)) {
    return ctrl;
  }
  
  /* step through arguments */
  for (ctrl=0; argc-- > 0; ++argv) {
    
    /* generic options */
    if (!strcmp(*argv, "debug")) {
      ctrl |= PAM_DEBUG_ARG;
    } else if (!strcmp(*argv, "force")) {
      ctrl |= PAM_FORCE_ARG;
    } else if (!strcmp(*argv, "use_first_pass")) {
      ctrl |= PAM_USE_FIRST_PASS;
    } else if (!strcmp(*argv, "try_first_pass")) {
      ctrl |= PAM_TRY_FIRST_PASS;
    } else {
      _pam_log(LOG_WARNING, "unrecognized option '%s'", *argv);
    }
  }
  
  return ctrl;
}

#define DPRINT if (ctrl & PAM_DEBUG_ARG) _pam_log

#undef PAM_FAIL_CHECK
#define PAM_FAIL_CHECK if (retval != PAM_SUCCESS) { return retval; }

static int pam_converse(pam_handle_t *pamh, int msg_style, char *message, char **password)
{
  const struct pam_conv *conv;
  struct pam_message resp_msg;
  const struct pam_message *msg[1];
  struct pam_response *resp = NULL;
  int retval;
  
  resp_msg.msg_style = msg_style;
  resp_msg.msg = message;
  msg[0] = &resp_msg;
  
  /* grab the password */
  retval = pam_get_item(pamh, PAM_CONV, (const void **)(void *) &conv);
  PAM_FAIL_CHECK;
  
  retval = conv->conv(1, msg, &resp,conv->appdata_ptr);
  PAM_FAIL_CHECK;
  
  if (password) {		/* assume msg.type needs a response */
    /* I'm not sure if this next bit is necessary on Linux */
#ifdef sun
    /* NULL response, fail authentication */
    if ((resp == NULL) || (resp->resp == NULL)) {
      return PAM_SYSTEM_ERR;
    }
#endif
    
    *password = resp->resp;
    free(resp);
  }
  
  return PAM_SUCCESS;
}

#undef PAM_FAIL_CHECK

PAM_EXTERN int 
pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc,const char **argv)
{
  char *user, *prompt, *pass;
  int ctrl, retval;
  struct pam_appl_data *pappl;
  UnixAdsap2_msg *msg;
  int fd;

  ctrl = _pam_parse(argc, argv);
  DPRINT(LOG_DEBUG, "start");

  pappl = get_cyc_appl(pamh);

  if (prompt == NULL) prompt = "Who are you? ";
  retval = pam_get_user(pamh, (const char **)&user, prompt);

  if (retval != PAM_SUCCESS || user == NULL || *user == '\0') {
	DPRINT(LOG_DEBUG, "User name not available");
	return(PAM_AUTH_ERR);
  }
  
  DPRINT(LOG_DEBUG, "Got user name %s", user);

  /* grab the password (if any) from the previous authentication layer */
  retval = pam_get_item(pamh, PAM_AUTHTOK, (const void **)(void *)&pass);    // [GY] 2005/Sep/19  clean warning

  if (retval != PAM_SUCCESS || pass == NULL || *pass == '\0')
    pass = NULL;
  else
    pass = strdup(pass);

  /* no previous password: maybe get one from the user */
  if (!pass && (ctrl & PAM_USE_FIRST_PASS)) {
    _pam_log(LOG_ERR, "unable to re-retrieve password from previous layer");
    return PAM_AUTHINFO_UNAVAIL;
  } /* end of password == NULL */

  if (!pass) {
	retval = pam_converse(pamh, PAM_PROMPT_ECHO_OFF, "Password: ", &pass);
    if (retval != PAM_SUCCESS || pass == NULL || *pass == '\0') {
      _pam_log(LOG_ERR, "unable to obtain password" );
      return PAM_CRED_INSUFFICIENT;
    }
    pass = strdup(pass);
  }

  /* If there was a password pass it to the next layer */
  if (pass) {
    pam_set_item(pamh, PAM_AUTHTOK, pass);
  }

  if ((fd = unixAdsap2_connect()) < 0) {
    _pam_log(LOG_ERR, "unixAdsap2_connect failed\n");
    retval = PAM_AUTHINFO_UNAVAIL;
    goto exit_pass;
  }

  if ((msg = unixAdsap2_new_req()) == NULL) {
    _pam_log(LOG_ERR, "unixAdsap2_new_req failed\n");
    retval = PAM_AUTHINFO_UNAVAIL;
    goto exit_pass;
  }

  msg->req.port_num = pappl ? pappl->port : 0;

  if (inixAdsap2_put_field(strlen(user), UnixAdsap2FieldIdUsername, user)) {
    _pam_log(LOG_ERR, "unixAdsap2_put_field-1 failed\n");
    retval = PAM_AUTHINFO_UNAVAIL;
    goto exit_pass_msg;
  }
  if (inixAdsap2_put_field(strlen(pass), UnixAdsap2FieldIdPassword, pass)) {
    _pam_log(LOG_ERR, "unixAdsap2_put_field-2 failed\n");
    retval = PAM_AUTHINFO_UNAVAIL;
    goto exit_pass_msg;
  }

  if (unixAdsap2_putmsg(fd, msg)) {
    _pam_log(LOG_ERR, "unixAdsap2_putmsg failed\n");
    retval = PAM_AUTHINFO_UNAVAIL;
    goto exit_pass_msg;
  }

  unixAdsap2_drop_req();
  
  if ((msg = unixAdsap2_getmsg(fd)) == NULL) {
    _pam_log(LOG_ERR, "unixAdsap2_getmsg FIRST failed\n");
    retval = PAM_AUTHINFO_UNAVAIL;
    goto exit_pass;
  }

  DPRINT(LOG_DEBUG, "rep_msg->size      : %d\n", msg->rep.size);
  DPRINT(LOG_DEBUG, "rep_msg->rtype     : %d\n", msg->rep.rtype);

  if (msg->rep.rtype == 1) {
    DPRINT(LOG_DEBUG, "rep_msg->rcode     : %d\n", msg->rep.rcode);
    DPRINT(LOG_DEBUG, "rep_msg->privilege : %d\n", msg->rep.privilege);
    DPRINT(LOG_DEBUG, "rep_msg->preempt   : %d\n", msg->rep.preempt);
  } else {
    _pam_log(LOG_ERR, "unixAdsap2_getmsg wrong reply message\n");
    retval = PAM_AUTHINFO_UNAVAIL;
    goto exit_pass_msg;
  }

  if (msg->rep.rcode) {
    _pam_log(LOG_INFO, "unixAdsap2_getmsg authentication failed %d\n", msg->rep.rcode);
    //[RK]Jul/03/06 - check the code : 1-FAILED, 2-UNAVAILABLE
    retval =(msg->rep.rcode == 1) ?  PAM_AUTH_ERR : PAM_AUTHINFO_UNAVAIL;
    goto exit_pass_msg;
  }

  // Handles the field privilege
  if (msg->rep.privilege > 1 && pappl) { // admin
	pappl->group_name = strdup("admin");
  }

  // Yet not know how to handle field preempt

  retval = PAM_SUCCESS;

exit_pass_msg:
  free(msg);

exit_pass:
  _pam_forget(pass);

  if (fd >= 0) close (fd);

  return(retval);

}

/*
 * Return a value matching the return value of pam_sm_authenticate, for
 * greatest compatibility. 
 * (Always returning PAM_SUCCESS breaks other authentication modules;
 * always returning PAM_IGNORE breaks PAM when we're the only module.)
 */
PAM_EXTERN int
pam_sm_setcred(pam_handle_t *pamh,int flags,int argc,const char **argv)
{
  return(PAM_SUCCESS);
}

PAM_EXTERN int
pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
   return PAM_SUCCESS;
}

PAM_EXTERN int
pam_sm_open_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
   return PAM_SUCCESS;
}

PAM_EXTERN int
pam_sm_close_session(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
   return PAM_SUCCESS;
}

PAM_EXTERN int
pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
   return PAM_SUCCESS;
}

#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_auth_srv_modstruct = {
  "pam_adsap2",
  pam_sm_authenticate,
  pam_sm_setcred,
  pam_sm_acct_mgmt,
  pam_sm_open_session,
  pam_sm_close_session,
  pam_sm_chauthtok,
};
#endif

