#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "avoprint.h"
#include "key.h"
#include "ssh-buffer.h"
#include "strlcat.h"

#define ADSAP_WARNING     (AVODBG_WARNING | AVODBG_ADSAP)
#define ADSAP_FAILURE     (AVODBG_FAILURE | AVODBG_ADSAP)
#define ADSAP_INFO        (AVODBG_INFO    | AVODBG_ADSAP)
#define ADSAP_ADHOC       (AVODBG_ADHOC   | AVODBG_ADSAP)

Key *ssh_key_new(int type)
{
	Key		*k;
	RSA		*rsa;
	DSA		*dsa;
	size_t	size = sizeof(*k);

	k = malloc(size);

	k->type		= type;
	k->flags	= 0;
	k->dsa		= NULL;
	k->rsa		= NULL;

	switch (k->type)
	{
		case KEY_RSA1:
		case KEY_RSA:
			rsa = RSA_new();
			rsa->n = BN_new();
			rsa->e = BN_new();
			k->rsa = rsa;
			break;
		case KEY_DSA:
			dsa = DSA_new();
			dsa->p = BN_new();
			dsa->q = BN_new();
			dsa->g = BN_new();
			dsa->pub_key = BN_new();
			k->dsa = dsa;
			break;
		case KEY_UNSPEC:
			break;
		default:
			avoPrintf(ADSAP_FAILURE, "ssh_key_new: bad key type %d", k->type);
			break;
	}

	return k;
}

Key *ssh_key_new_private(int type)
{
	Key *k = ssh_key_new(type);
	switch (k->type)
	{
		case KEY_RSA1:
		case KEY_RSA:
			k->rsa->d = BN_new();
			k->rsa->iqmp = BN_new();
			k->rsa->q = BN_new();
			k->rsa->p = BN_new();
			k->rsa->dmq1 = BN_new();
			k->rsa->dmp1 = BN_new();
			break;
		case KEY_DSA:
			k->dsa->priv_key = BN_new();
			break;
		case KEY_UNSPEC:
			break;
		default:
			break;
	}
	return k;
}

void ssh_key_free(Key *k)
{
	switch (k->type)
	{
		case KEY_RSA1:
		case KEY_RSA:
			if (k->rsa != NULL)
				RSA_free(k->rsa);
			k->rsa = NULL;
			break;
		case KEY_DSA:
			if (k->dsa != NULL)
				DSA_free(k->dsa);
			k->dsa = NULL;
			break;
		case KEY_UNSPEC:
			break;
		default:
			avoPrintf(ADSAP_FAILURE, "key_free: bad key type %d", k->type);
			break;
	}
	free(k);
}

char *key_type(Key *k)
{
	switch (k->type)
	{
		case KEY_RSA1:
			return "RSA1";
			break;
		case KEY_RSA:
			return "RSA";
			break;
		case KEY_DSA:
			return "DSA";
			break;
	}
	return "unknown";
}

char *key_ssh_name(Key *k)
{
	switch (k->type)
	{
		case KEY_RSA:
			return "ssh-rsa";
			break;
		case KEY_DSA:
			return "ssh-dss";
			break;
	}
	return "ssh-unknown";
}

RSA *RSA_from_blob(u_char *blob, int blen)
{
	SSH_Buffer b;
	char *ktype;
	int rlen;
	RSA *rsa = NULL;

	ssh_buffer_init(&b);
	ssh_buffer_append(&b, blob, blen);
	ktype = buffer_get_string(&b, NULL);

	rsa = RSA_new();
	rsa->n = BN_new();
	rsa->e = BN_new();

	buffer_get_bignum2(&b, rsa->e);
	buffer_get_bignum2(&b, rsa->n);

	rlen = ssh_buffer_len(&b);

	if (rsa != NULL && rlen != 0)
		avoPrintf(ADSAP_FAILURE, "key_from_blob: remaining bytes in key blob %d", rlen);

	free(ktype);
	ssh_buffer_free(&b);
	return rsa;
}

int key_to_blob(Key *key, u_char **blobp, u_int *lenp)
{
	SSH_Buffer b;
	int len;
	u_char *buf;

	if (key == NULL)
	{
		avoPrintf(ADSAP_FAILURE, "key_to_blob: key == NULL");
		return 0;
	}
	ssh_buffer_init(&b);
	switch(key->type)
	{
		case KEY_DSA:
			buffer_put_cstring(&b, key_ssh_name(key));
			buffer_put_bignum2(&b, key->dsa->p);
			buffer_put_bignum2(&b, key->dsa->q);
			buffer_put_bignum2(&b, key->dsa->g);
			buffer_put_bignum2(&b, key->dsa->pub_key);
			break;
		case KEY_RSA:
			buffer_put_cstring(&b, key_ssh_name(key));
			buffer_put_bignum2(&b, key->rsa->e);
			buffer_put_bignum2(&b, key->rsa->n);
			break;
		default:
			avoPrintf(ADSAP_FAILURE, "key_to_blob: unsupported key type %d", key->type);
			ssh_buffer_free(&b);
			return 0;
	}
	len = ssh_buffer_len(&b);
	buf = malloc(len);
	memcpy(buf, ssh_buffer_ptr(&b), len);
	memset(ssh_buffer_ptr(&b), 0, len);
	ssh_buffer_free(&b);
	if (lenp != NULL)
		*lenp = len;
	if (blobp != NULL)
		*blobp = buf;
	return len;
}

static u_char *key_fingerprint_raw(Key *k, enum fp_type dgst_type, size_t *dgst_raw_length)
{
	EVP_MD *md = NULL;
	EVP_MD_CTX ctx;
	u_char *blob = NULL;
	u_char *retval = NULL;
	int len = 0;
	int nlen, elen;

	*dgst_raw_length = 0;

	switch (dgst_type)
	{
		case SSH_FP_MD5:
			md = (EVP_MD *)EVP_md5();
			break;
		case SSH_FP_SHA1:
			md = (EVP_MD *)EVP_sha1();
			break;
		default:
			avoPrintf(ADSAP_FAILURE, "key_fingerprint_raw: bad digest type %d", dgst_type);
	}
	switch (k->type)
	{
		case KEY_RSA1:
			nlen = BN_num_bytes(k->rsa->n);
			elen = BN_num_bytes(k->rsa->e);
			len = nlen + elen;
			blob = malloc(len);
			BN_bn2bin(k->rsa->n, blob);
			BN_bn2bin(k->rsa->e, blob + nlen);
			break;
		case KEY_DSA:
		case KEY_RSA:
			key_to_blob(k, &blob, &len);
			break;
		case KEY_UNSPEC:
			return retval;
			break;
		default:
			avoPrintf(ADSAP_FAILURE, "key_fingerprint_raw: bad key type %d", k->type);
			break;
	}
	if (blob != NULL)
	{
		retval = malloc(EVP_MAX_MD_SIZE);
		EVP_DigestInit(&ctx, md);
		EVP_DigestUpdate(&ctx, blob, len);
		EVP_DigestFinal(&ctx, retval, NULL);
		*dgst_raw_length = md->md_size;
		memset(blob, 0, len);
		free(blob);
	}
	else
	{
		avoPrintf(ADSAP_FAILURE, "key_fingerprint_raw: blob is null");
	}
	return retval;
}

static char *key_fingerprint_hex(u_char* dgst_raw, size_t dgst_raw_len)
{
	char *retval;
	unsigned int i;

	retval = malloc(dgst_raw_len * 3 + 1);
	retval[0] = '\0';
	for(i = 0; i < dgst_raw_len; i++)
	{
		char hex[4];
		snprintf(hex, sizeof(hex), "%02x:", dgst_raw[i]);
		strlcat(retval, hex, dgst_raw_len * 3);
	}
	retval[(dgst_raw_len * 3) - 1] = '\0';
	return retval;
}

static char *key_fingerprint_bubblebabble(u_char* dgst_raw, size_t dgst_raw_len)
{
	char vowels[] = { 'a', 'e', 'i', 'o', 'u', 'y' };
	char consonants[] = { 'b', 'c', 'd', 'f', 'g', 'h', 'k', 'l', 'm',
		'n', 'p', 'r', 's', 't', 'v', 'z', 'x' };
	u_int i, j = 0, rounds, seed = 1;
	char *retval;

	rounds = (dgst_raw_len / 2) + 1;
	retval = malloc(sizeof(char) * (rounds*6));
	retval[j++] = 'x';
	for (i = 0; i < rounds; i++)
	{
		u_int idx0, idx1, idx2, idx3, idx4;
		if ((i + 1 < rounds) || (dgst_raw_len % 2 != 0))
		{
			idx0 = (((((u_int)(dgst_raw[2 * i])) >> 6) & 3) +
					seed) % 6;
			idx1 = (((u_int)(dgst_raw[2 * i])) >> 2) & 15;
			idx2 = ((((u_int)(dgst_raw[2 * i])) & 3) +
					(seed / 6)) % 6;
			retval[j++] = vowels[idx0];
			retval[j++] = consonants[idx1];
			retval[j++] = vowels[idx2];
			if ((i + 1) < rounds)
			{
				idx3 = (((u_int)(dgst_raw[(2 * i) + 1])) >> 4) & 15;
				idx4 = (((u_int)(dgst_raw[(2 * i) + 1]))) & 15;
				retval[j++] = consonants[idx3];
				retval[j++] = '-';
				retval[j++] = consonants[idx4];
				seed = ((seed * 5) +
						((((u_int)(dgst_raw[2 * i])) * 7) +
							((u_int)(dgst_raw[(2 * i) + 1])))) % 36;
			}
		}
		else
		{
			idx0 = seed % 6;
			idx1 = 16;
			idx2 = seed / 6;
			retval[j++] = vowels[idx0];
			retval[j++] = consonants[idx1];
			retval[j++] = vowels[idx2];
		}
	}
	retval[j++] = 'x';
	retval[j++] = '\0';
	return retval;
}

char *key_fingerprint(Key *k, enum fp_type dgst_type, enum fp_rep dgst_rep)
{
	char *retval = NULL;
	u_char *dgst_raw;
	size_t dgst_raw_len;

	dgst_raw = key_fingerprint_raw(k, dgst_type, &dgst_raw_len);
	if (!dgst_raw)
		avoPrintf(ADSAP_FAILURE, "key_fingerprint: null from key_fingerprint_raw()");
	switch(dgst_rep)
	{
		case SSH_FP_HEX:
			retval = key_fingerprint_hex(dgst_raw, dgst_raw_len);
			break;
		case SSH_FP_BUBBLEBABBLE:
			retval = key_fingerprint_bubblebabble(dgst_raw, dgst_raw_len);
			break;
		default:
			avoPrintf(ADSAP_FAILURE, "key_fingerprint_ex: bad digest representation %d", dgst_rep);
			break;
	}
	memset(dgst_raw, 0, dgst_raw_len);
	free(dgst_raw);
	return retval;
}
