#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <pthread.h>

//#include "epc.h"
#include "list-item.h"
#include "avoprint.h"

/* 
** Debugging of heap memory allocations 
*/
#ifdef DMALLOC    
  #include <dmalloc.h>
#endif

// Create a recursive mutex which supports nested locks/unlocks. When locked is called, it 
// decrements the locking count of the mutex (number of pthread_mutex_lock operations 
// performed on it by the calling thread), and only when this count reaches zero is the 
// mutex actually unlocked.
static pthread_mutex_t listMutex = {0, 0, 0, PTHREAD_MUTEX_RECURSIVE_NP, {0, 0}};

//**************************************************************************************

BOOLEAN init_list_lock(PGLOBAL_LIST p_list)
{
    static int noMoreMutex = FALSE;
    
    if (noMoreMutex)
    {
		return(FALSE);
    }
    
	if(p_list)
	{
        // Not sure how to create a recursive mutex using pthread_mutex_init()
        // pthread_mutex_init( & p_list->list_lock, NULL );
        p_list->list_lock = listMutex;
        noMoreMutex = TRUE;
		return(TRUE);
	}
	else
		return(FALSE);
}

//**************************************************************************************

PGLOBAL_LIST new_list(void)
{
	PGLOBAL_LIST	p_list;
	size_t			struct_size = sizeof(GLOBAL_LIST);

	p_list = malloc(struct_size);

	if(p_list)
	{
		memset(p_list, 0, struct_size);

		if(!init_list_lock(p_list))
		{
			free(p_list);
			p_list = NULL;
		}
	}

	return(p_list);
}

//*****************************************************************************

PLIST_ITEM new_list_item(void *p_data)
{
	PLIST_ITEM	p_item;
	size_t		struct_size = sizeof(LIST_ITEM);

	p_item = malloc(struct_size);

	if(p_item)
	{
		memset(p_item, 0, struct_size);
		p_item->p_data = p_data;
        //printf("new_list_item: 0x%x\n", (unsigned int)p_item);
	}

	return(p_item);
}

//*****************************************************************************

void free_list_item(PGLOBAL_LIST p_list, PLIST_ITEM p_item, BOOLEAN list_locked)
{
	if((p_list) && (p_item))
	{
		if(!list_locked)
			pthread_mutex_lock(& p_list->list_lock);

		if(p_list->free_cb)
			p_list->free_cb((void *) p_item->p_data);
		else
			free(p_item->p_data);

		p_item->p_data = NULL;
		p_item->p_prev = NULL;
		p_item->p_next = NULL;
		free(p_item);

		if(!list_locked)
			pthread_mutex_unlock(& p_list->list_lock);
	}
}

//*****************************************************************************

WORD avail_list_item(PGLOBAL_LIST p_list, PLIST_ITEM p_item, BOOLEAN list_locked)
{
	WORD status = 0xFFFF;

	if((p_list) && (p_item))
	{
		if(!list_locked)
			pthread_mutex_lock(& p_list->list_lock);

		if(p_list->port_avail_cb)
			status = p_list->port_avail_cb((void *) p_item->p_data);

		if(!list_locked)
			pthread_mutex_unlock(& p_list->list_lock);
	}

	return(status);
}

//**************************************************************************************

void set_item_compare_callback(PGLOBAL_LIST p_list, BOOLEAN (*cb)(void *, void *, int))
{
	if((p_list) && (cb))
	{
		pthread_mutex_lock(& p_list->list_lock);
		p_list->compare_cb = cb;
		pthread_mutex_unlock(& p_list->list_lock);
	}
}

//**************************************************************************************

void set_item_valid_callback(PGLOBAL_LIST p_list, BOOLEAN (*cb)(void *))
{
	if((p_list) && (cb))
	{
		pthread_mutex_lock(& p_list->list_lock);
		p_list->valid_cb = cb;
		pthread_mutex_unlock(& p_list->list_lock);
	}
}

//**************************************************************************************

void set_item_free_callback(PGLOBAL_LIST p_list, BOOLEAN (*cb)(void *))
{
	if((p_list) && (cb))
	{
		pthread_mutex_lock(& p_list->list_lock);
		p_list->free_cb = cb;
		pthread_mutex_unlock(& p_list->list_lock);
	}
}

//**************************************************************************************

void set_item_avail_callback(PGLOBAL_LIST p_list, WORD (*cb)(void *))
{
	if((p_list) && (cb))
	{
		pthread_mutex_lock(& p_list->list_lock);
		p_list->port_avail_cb = cb;
		pthread_mutex_unlock(& p_list->list_lock);
	}
}

//*****************************************************************************

void add_list_item(PGLOBAL_LIST p_list, PLIST_ITEM p_item, BOOLEAN list_locked)
{
	if((p_list) && (p_item))
	{
		if(!list_locked)
			pthread_mutex_lock(& p_list->list_lock);

		if(p_list->list_tail == NULL)							// No items in the list
		{
			p_list->list_head			= p_item;				// Head and tail are the same
			p_list->list_tail			= p_item;
		}
		else
		{
			p_list->list_tail->p_next	= p_item;				// Add forward reference
			p_item->p_prev				= p_list->list_tail;	// Add backward reference
			p_list->list_tail			= p_item;
		}

		if(!list_locked)
			pthread_mutex_unlock(& p_list->list_lock);
	}
}

//*****************************************************************************

PLIST_ITEM del_list_item(PGLOBAL_LIST p_list, PLIST_ITEM p_item, BOOLEAN list_locked)
{
	PLIST_ITEM	p_next = NULL;

	if((p_list) && (p_item))
	{
		if(!list_locked)
			pthread_mutex_lock(& p_list->list_lock);

		if(p_item->p_prev != NULL)								// If NOT the first item
			(p_item->p_prev)->p_next	= p_item->p_next;		//    Correct the forward reference
		else													// Else, must be the first item
			p_list->list_head			= p_item->p_next;		//    Just move the head

		if(p_item->p_next != NULL)								// If NOT the last item
			(p_item->p_next)->p_prev	= p_item->p_prev;		//    Correct the backward reference
		else													// Else, must be the last item
			p_list->list_tail			= p_item->p_prev;		//    Just move the tail

		p_next = p_item->p_next;								// Return pointer to next item (if any)

		free_list_item(p_list, p_item, list_locked);			// Free list item structure and data

		if(!list_locked)
			pthread_mutex_unlock(& p_list->list_lock);
    
        //printf("del_list_item: 0x%x\n", (unsigned int)p_item);
	}

	return p_next;
}

//*****************************************************************************

PLIST_ITEM find_list_item(PGLOBAL_LIST p_list, PLIST_ITEM p_start, void *p_data, BOOLEAN forward, int data_type, BOOLEAN list_locked)
{
	PLIST_ITEM p_item = NULL;

	if((p_list) && (p_data))
	{
		if(!list_locked)
			pthread_mutex_lock(& p_list->list_lock);

		p_item = p_start;

		while(p_item != NULL) 
		{
			if(p_list->compare_cb)
			{
				if(p_list->compare_cb(p_item->p_data, p_data, data_type))
					break;
			}
			else
			{
				if(p_item->p_data == p_data)
					break;
			}

			if(forward)
				p_item = p_item->p_next;
			else
				p_item = p_item->p_prev;
		}

		if(!list_locked)
			pthread_mutex_unlock(& p_list->list_lock);
	}

	return p_item;
}

//*****************************************************************************

void purge_list(PGLOBAL_LIST p_list, BOOLEAN list_locked)
{
	PLIST_ITEM	p_item;

	if(p_list)
	{
		if(!list_locked)
			pthread_mutex_lock(& p_list->list_lock);

		p_item = p_list->list_head;

		while(p_item != NULL) 
		{
			if(!(item_valid(p_list, p_item, list_locked)))
      {
				p_item = del_list_item(p_list, p_item, list_locked);
        avoPrintf(3, "purge preauth item\n");
      }
			else
				p_item = p_item->p_next;
		}

		if(!list_locked)
			pthread_mutex_unlock(& p_list->list_lock);
	}
}

//**************************************************************************************

BOOLEAN item_valid(PGLOBAL_LIST p_list, PLIST_ITEM p_item, BOOLEAN list_locked)
{
	BOOLEAN ret_val = TRUE;

	if((p_list) && (p_item))
	{
		if(!list_locked)
			pthread_mutex_lock(& p_list->list_lock);

		if(p_list->valid_cb)
			ret_val = (p_list->valid_cb((void *) p_item->p_data));

		if(!list_locked)
			pthread_mutex_unlock(& p_list->list_lock);
	}

	return(ret_val);
}

//**************************************************************************************

void lock_unlock_list(PGLOBAL_LIST p_list, BOOLEAN lock)
{
	if (lock) {
		pthread_mutex_lock(& p_list->list_lock);
	} else {
		pthread_mutex_unlock(& p_list->list_lock);
	}
}

