#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#include "avoprint.h"
#include "list-item.h"

#include <openssl/ssl.h>
#include <openssl/rsa.h>       /* SSLeay stuff */
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/err.h>

#include "ds2mib.h"
#include "adsap2.h"
#include "ssl.h"
#include "netwrap.h"
#include "threadwrap.h"

#define _ADSAP2_SAERVER_
#include "api_adsap2.h"

#define ADSAP_WARNING     (AVODBG_WARNING | AVODBG_ADSAP)
#define ADSAP_FAILURE     (AVODBG_FAILURE | AVODBG_ADSAP)
#define ADSAP_INFO        (AVODBG_INFO    | AVODBG_ADSAP)
#define ADSAP_ADHOC       (AVODBG_ADHOC   | AVODBG_ADSAP)

static int thread_ok = 0;

#define ADSAP2_CONF		"/etc/adsap2.conf"
#define ST_SECURE_STR	"secure"
#define ST_TRUSTALL_STR	"trust-all"

static int parse_config_file(char *fconf);
static void * signalHandlerThread(void *threaddata);

void printUsage(char *pgm)
{
	printf("\nUSAGE: %s [options]\n\n", pgm);
	printf("where options are:\n\n");
	printf("\t-d <debuglevel>\n");
}

static void * unixCmdTread(void *threaddata)
{
	int fd = (int) threaddata;
	int i,j;
	unsigned long auth_ds;
	UnixAdsap2_msg *msg;
	UnixAdsap2_field *pf;
	UnixAdsap2_auth_rep rep;
	ADSAP2_AUTHORIZE_REQ AuthReq;
	ADSAP2_AUTHORIZE_REP AuthRep;

	pthreadDetach( pthread_self() );

	avoPrintf(ADSAP_INFO, "unixCmdTread called\n");
	while ((msg = unixAdsap2_getmsg(fd)) != NULL) {
		memset(&AuthReq, 0, sizeof(ADSAP2_AUTHORIZE_REQ));
		memset(&AuthRep, 0, sizeof(ADSAP2_AUTHORIZE_REP));
		memset(&rep, 0, sizeof(UnixAdsap2_auth_rep));

		avoPrintf(ADSAP_INFO, "unixCmdTread req->size : %d\n", msg->req.size);
		avoPrintf(ADSAP_INFO, "unixCmdTread req->rtype: %d\n", msg->req.rtype);
		avoPrintf(ADSAP_INFO, "unixCmdTread req->port : %d\n", msg->req.port_num);

		i = 0;
		while ((pf = unixAdsap2_get_field(i, msg)) != NULL) {
			avoPrintf(ADSAP_INFO, "unixCmdTread req->f[%d].len : %d\n", i, pf->flen);
			avoPrintf(ADSAP_INFO, "unixCmdTread req->f[%d].id  : %d\n", i, pf->fid);

			switch (pf->fid) {
			case UnixAdsap2FieldIdUsername:
				strncpy(AuthReq.username, pf->fval, ADSAP2_USERNAME_STR_LEN-1);
				avoPrintf(ADSAP_INFO, "unixCmdTread AuthReq user %s\n", AuthReq.username);
				break;
			case UnixAdsap2FieldIdPassword:
				strncpy(AuthReq.password, pf->fval, ADSAP2_PASSWORD_STR_LEN-1);
				//avoPrintf(ADSAP_INFO, "unixCmdTread AuthReq password %s\n", AuthReq.password);
				break;
			case UnixAdsap2FieldIdBlobKey:
				if ((AuthReq.p_Blob = malloc(pf->flen)) == NULL) {
					avoPrintf(AVODBG_ALWAYS, "unixCmdTread AuthReq blob: malloc failed\n");
					return(NULL);
				}
				memcpy(AuthReq.p_Blob, pf->fval, pf->flen);
				AuthReq.Blob_len = pf->flen;
				avoPrintf(ADSAP_INFO, "unixCmdTread AuthReq blob-len %d\n", AuthReq.Blob_len);
#if 0
				for (j=0; j < AuthReq.Blob_len; j++)
					avoPrintf(ADSAP_INFO, "%02x", AuthReq.p_Blob[j]);
				avoPrintf(ADSAP_INFO, "\n");
#endif
				break;
			case UnixAdsap2FieldIdStyle:
				memcpy(AuthReq.ripID, pf->fval, pf->flen);
				AuthReq.ripID[pf->flen] = 0;
				avoPrintf(ADSAP_INFO, "unixCmdTread AuthReq Style/RIP %s\n", AuthReq.ripID);
				break;
			default:
				break;
			}
			free(pf);
			i++;
		}

		AuthReq.port = msg->req.port_num;
		AuthReq.mode = ADSAP2_NORMAL;
		AuthReq.testLocalKey = msg->req.check_only;
		auth_ds = authenticate_ds(&AuthReq, &AuthRep);

		if (AuthReq.p_Blob) free(AuthReq.p_Blob);
			
		free(msg);

		rep.size = sizeof(UnixAdsap2_auth_rep);
		rep.rtype = 1;	// Reply
		//rep.rcode = (auth_ds == ADSAP2_DSAUTHENTICATION_SUCCESS)?0:1;
		//[RK]Jul/03/06 - values : 0-OK, 1-FAILED, 2-UNAVAILABLE
		rep.rcode = auth_ds;
		if (auth_ds == ADSAP2_DSAUTHENTICATION_SUCCESS) {
			rep.port_num =  AuthReq.port;
			rep.privilege = AuthRep.role;
			rep.preempt = AuthRep.preempt;
			rep.logging_dla = AuthRep.logdla;
			memcpy(rep.style, AuthRep.RIP_ID, sizeof(rep.style));
			memcpy(rep.command, AuthRep.command, sizeof(rep.command));
		}
		unixAdsap2_putmsg(fd, (UnixAdsap2_msg *)&rep);
	}
	avoPrintf(ADSAP_INFO, "unixCmdTread done\n");

	close(fd);

	return NULL;
}

int main (int argc, char *argv[])
{
	extern char *optarg;
	extern int optind;
	unsigned long level = 0;
	int opt, base = 10;
	int on = 1;
	pthread_t adsap2ListenerThreadID;
	pthread_t signalHandlerThreadID;
	pthread_t unixThreadID;
	int fd, fd_unix;
	char *intf = NULL;
	sigset_t sigs_to_block;

	/* Parse command-line arguments. */
	while ((opt = getopt(argc, argv, "d:")) != EOF) {
		switch (opt) {
		case 'd':
			if ((optarg[0] == '0' && optarg[1] == 'x') || (optarg[0] == 'x')) {
				base = 16;
			}
            level = strtoul(optarg, NULL, base);
            break;
		default:
			printf("Invalid option: '%c'\n", opt);
			printUsage(argv[0]);
			exit(-1);
			break;
		}
	}

	avoPrintfSetLevel(level);
	avoPrintf(ADSAP_INFO, "ADSAP2 starting debug level: %d\n", level);

	// Block signals
    sigemptyset(&sigs_to_block);
    sigaddset(&sigs_to_block, SIGTERM);
    sigaddset(&sigs_to_block, SIGQUIT);
    sigaddset(&sigs_to_block, SIGINT);
    sigaddset(&sigs_to_block, SIGUSR1);
    sigaddset(&sigs_to_block, SIGUSR2);
    sigaddset(&sigs_to_block, SIGPIPE);
    sigaddset(&sigs_to_block, SIGHUP);
    sigaddset(&sigs_to_block, SIGABRT);
    sigaddset(&sigs_to_block, SIGCHLD);

	pthread_sigmask(SIG_BLOCK, &sigs_to_block, NULL);
	
    pthreadCreate(&signalHandlerThreadID, NULL, &signalHandlerThread, NULL);

	while (!thread_ok) {
		sleep(1);
	}

	avoPrintf(ADSAP_INFO, "ADSAP2 initializing ssl\n");
	SSL_Init();

    /*
    ** Spawn ADSAP2 server listener
    */
    pthreadCreate(&adsap2ListenerThreadID, NULL, &adsap2ConnectServer, intf);
 
	/*
	** Create the unix socket
	*/
	fd_unix = UNIX_Listen(UNIX_SOCKET_NAME);
	fcntl(fd_unix, F_SETFD, FD_CLOEXEC);

	/*
	** Listen for authentication request from Application
	*/
	while (1) {
		avoPrintf(ADSAP_INFO, "ADSAP2 waiting UNIX socket connection\n");
		if ((fd = accept(fd_unix, NULL, 0)) >= 0) {
			/*
			** Spawn a UnixCmd thread
			*/
			avoPrintf(ADSAP_INFO, "ADSAP2 got UNIX socket connection\n");
			fcntl(fd, F_SETFD, FD_CLOEXEC);
			setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
    		pthreadCreate(&unixThreadID, NULL, &unixCmdTread, (void *)fd);
		} else {
			avoPrintf(ADSAP_INFO, "ADSAP2 UNIX socket accept error %s\n", strerror(errno));
		}
	}
	return 0;
}

static void  adsap_exit(void)
{
	avoPrintf(ADSAP_INFO, "ADSAP2 finished\n");
}

// Read config file. Read values will be stored by the  DsmibInitAuth.. routine
static int parse_config_file(char *fconf)
{
	struct stat pbuf;
	FILE *fc;
	char *state = NULL;
	char *authserver = NULL;
	char *buf = NULL;
	char *as, *tok, *tokptr;
	int ix, ix_server = 0;
	IP_ADDR serverauth;
	int st = DSMIB_SECURITY_MODE_SECURE;		// default state mode

// Check config file name. Open file if OK.
	if (!fconf || (stat(fconf, &pbuf))) {
		goto err_exit;
	}
	if ((fc = fopen(fconf, "r")) == NULL) {
		goto err_exit;
	}

// Alloc buffer to read the file,
	if ((buf = calloc(1, pbuf.st_size+1)) == NULL) {
		fclose(fc);
		goto err_exit;
	}

// Read file, line by line and check for valid parameters.
	while (fgets(buf, pbuf.st_size, fc)) {
		if (!state && !strncmp(buf, "state ", 6)) {
// Save configured state string
			state = strdup(buf+6);
		}
		if (!authserver && !strncmp(buf, "authserver ", 11)) {
// Save configured authserver string
			authserver = strdup(buf+11);
		}
	}
	if (buf) free(buf);

// Set state mode according to state string.
	if (state && *state) {
		if (!strncmp(state, ST_SECURE_STR, strlen(ST_SECURE_STR))) {
			st = DSMIB_SECURITY_MODE_SECURE;
		} else if (!strncmp(state, ST_TRUSTALL_STR, strlen(ST_TRUSTALL_STR))) {
			st = DSMIB_SECURITY_MODE_TRUSTALL;
		}
	}
	DsmibInitSecurityMode(st);

// Alloc buffer to parse authserver
	if (authserver && *authserver) {
		for (ix = 0, as = authserver; 
				ix < DSMIB_AUTHSERVER_ADDRESS_COUNT; 
				ix++, as = NULL) {
// Separate authserver string by its delimiter.
			if ((tok = strtok_r(as, ",", &tokptr))) {
// Check if string has a valid IP address and stores it in the struct.
				if (inet_aton(tok, &serverauth)) {
					DsmibInitAuthServerAddress(ix_server++, &serverauth);
				}
			} else {
				break;
			}
		}
	}

// Reset left DsmibInitAuthServerAddress entries.
	inet_aton("0.0.0.0", &serverauth);
	for (ix = ix_server; ix < DSMIB_AUTHSERVER_ADDRESS_COUNT; ix++) {
		DsmibInitAuthServerAddress(ix, &serverauth);
	}

// Release memory and close file. 
	if (state) free(state);
	if (authserver) free(authserver);
	fclose(fc);

	return 0;

err_exit:
	avoPrintf(AVODBG_ALWAYS,"Configuration file %s parse error.\n", fconf);
	return (-1);
}

static void * signalHandlerThread(void *threaddata)
{
	sigset_t sigs_to_catch;
	int sig_caught;

	sigemptyset(&sigs_to_catch);
	sigaddset(&sigs_to_catch, SIGTERM);
	sigaddset(&sigs_to_catch, SIGQUIT);
	sigaddset(&sigs_to_catch, SIGINT);
	sigaddset(&sigs_to_catch, SIGUSR1);
	sigaddset(&sigs_to_catch, SIGUSR2);
	sigaddset(&sigs_to_catch, SIGPIPE);
	sigaddset(&sigs_to_catch, SIGHUP);
	sigaddset(&sigs_to_catch, SIGABRT);
    sigaddset(&sigs_to_catch, SIGCHLD);

	// Read the configuration file and fill in the adsap2_conf struct
	parse_config_file(ADSAP2_CONF);

	atexit(adsap_exit);

	thread_ok = 1;

	while (1) {
		sigwait(&sigs_to_catch, &sig_caught);

		switch (sig_caught) {
		case SIGQUIT:
			avoPrintf(ADSAP_INFO, "ADSAP2 receive signal: QUIT\n");
			exit(-1);
			break;
		case SIGTERM:
		case SIGINT:
		case SIGABRT:
			avoPrintf(ADSAP_INFO, "ADSAP2 receive signal: INT\n");
			exit(0);
			break;
		case SIGCHLD:
			break;
		case SIGHUP:
			avoPrintf(ADSAP_INFO, "ADSAP2 receive signal: HUP\n");
			parse_config_file(ADSAP2_CONF);
			break;
		default:
			/* Do nothing for now */
			avoPrintf(ADSAP_INFO, "ADSAP2 receive signal: %d\n", sig_caught);
			continue;
		}
	}

	return(NULL);
}

void adsap2_system(char *cmd)
{
	pid_t pid;
	sigset_t sigs_to_unblock;
	int st;

	if ((pid = fork()) < 0) {
		avoPrintf(ADSAP_INFO,"Could not fork to execute %s\n", cmd);
		return;
	}

	if (pid) {  // parent
		if (waitpid(pid, &st, 0) == pid) {
			avoPrintf(ADSAP_INFO,"execl %s return %d\n", cmd, WEXITSTATUS(st));
		}
		return;
	}

	// Child

    sigaddset(&sigs_to_unblock, SIGTERM);
    sigaddset(&sigs_to_unblock, SIGQUIT);
    sigaddset(&sigs_to_unblock, SIGINT);
    sigaddset(&sigs_to_unblock, SIGUSR1);
    sigaddset(&sigs_to_unblock, SIGUSR2);
    sigaddset(&sigs_to_unblock, SIGPIPE);
    sigaddset(&sigs_to_unblock, SIGHUP);
    sigaddset(&sigs_to_unblock, SIGABRT);
    sigaddset(&sigs_to_unblock, SIGCHLD);

	pthread_sigmask(SIG_UNBLOCK, &sigs_to_unblock, NULL);

	execl(cmd, NULL);
	
	avoPrintf(ADSAP_INFO,"Could not execute %s\n", cmd);
}
