/*
 *  Module:  aidp-ts.c (Avocent Install and Discovery Protocol for TS family)
 *
 *  Author:  Marcelo Peccin <marcelo.peccin@avocent.com>
 *
 *  Copyright:  (c) 2006 Avocent Corporation
 *
 *  Date:  May 10, 2006
 *
 *  Comments:  Avocent Install and Discovery Protocol product-specific functions.
 *             This modules complements "aidpaux" module for TS case.
 *             See also aidpaux.c and aidp.c files.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <asm/prxk.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "aidp.h"
#include <rot_nports.h>


/////////////////////////////////////////////////////////////////////////////
//
//  Get appliance's Model code (as understood by DSView) and MAC address.
//
//  PARAMETERS: pointer to write appliance's Model
//              pointer to write appliance's 6-digit MAC address
//  
//  RETURN: 0 for success, -1 for error
//
/////////////////////////////////////////////////////////////////////////////
int appGetModelandMAC(unsigned short *appID, unsigned char* appMAC)
{
	int i;
	struct MachineType pmt;

	if (!get_machine_type(&pmt)) {
		printDEBUG("appGetModelandMAC: problem getting machine type\n");
		return -1;
	}

	switch(pmt.board_type)
	{
	case BOARD_TS1H:
		*appID = APP_TS100;
		break;
	case BOARD_TS4H:
		*appID = APP_TS400;
		break;
	case BOARD_TS8H:
		*appID = APP_TS800;
		break;
	case BOARD_TS1K:
		*appID = APP_TS1000;
		break;
	case BOARD_TS2K:
		*appID = APP_TS2000;
		break;
	case BOARD_TS3K:
		*appID = APP_TS3000;
		break;
	case BOARD_TS110:
		*appID = APP_TS110;
		break;
	default:
		printDEBUG("appGetModelandMAC: unknown TS type\n");
		return -1;
	}

	pmt.mac_address[12] = '\0';
	for (i = 5; i >= 0; i--) {
		appMAC[i] = (char)strtol(&pmt.mac_address[2*i], NULL, 16);
		pmt.mac_address[2*i] = '\0';
	}

	printDEBUG("appGetModelandMAC: model=%d MAC=%02x.%02x.%02x.%02x.%02x.%02x\n",
			   *appID, appMAC[0], appMAC[1], appMAC[2], appMAC[3], appMAC[4], appMAC[5]);

	return 0;
}


/////////////////////////////////////////////////////////////////////////////
//
//  Configure ethernet interface as desired, writing configuration
//  to apropriate configuration files (ramdisk only)
//
//  PARAMETERS: pointer to new IP address,
//              pointer to new network mask
//              pointer to new default Gateway address
//  
//  RETURN: 0 for success, -1 for error
//
/////////////////////////////////////////////////////////////////////////////
int appConfigEthInterface(const unsigned char *ip_addr,
						  const unsigned char *netmask,
						  const unsigned char *gw_addr)
{
	FILE *fp, *tempfp;
	char line[256];
	int  found_dhcp=FALSE, found_ip=FALSE, found_mask=FALSE;
	int lines_near_dhcp=0, retval=0;

	if ((fp = fopen("/etc/portslave/pslave.conf", "r")) < 0) {
		printDEBUG("appConfigEthInterface: %s opening config file\n", strerror(errno));
		return -1;
	}
	if ((tempfp = fopen("/tmp/pslave.conf", "w")) < 0) {
		printDEBUG("appConfigEthInterface: %s opening temp config file\n", strerror(errno));
		goto error_exit1;
	}

	while (fgets(line, 256, fp) != NULL) {
		if (line[0] != '#') {
			if (lines_near_dhcp) lines_near_dhcp--;
			if (strstr(line, "conf.dhcp_client")) {
				sprintf (line, "conf.dhcp_client\t0\n");
				found_dhcp = TRUE;
				lines_near_dhcp = 5; //will search next 5 lines for conf.eth_ip and
				                     //conf.eth_mask even if they are commented...
			} 
			else if (strstr(line, "conf.eth_ip")) {
				if (!strstr(line, "conf.eth_ip_alias")) {
					sprintf (line, "conf.eth_ip\t%d.%d.%d.%d\n",
							 ip_addr[0], ip_addr[1], ip_addr[2], ip_addr[3]);
					found_ip = TRUE;
				}
			}
			else if (strstr(line, "conf.eth_mask")) {
				if (!strstr(line, "conf.eth_mask_alias")) {
					sprintf (line, "conf.eth_mask\t%d.%d.%d.%d\n",
							 netmask[0], netmask[1], netmask[2], netmask[3]);
					found_mask = TRUE;
				}
			}
		} else {
			if (lines_near_dhcp) {
				lines_near_dhcp--;

				if (strstr(line, "conf.eth_ip")) {
					if (!strstr(line, "conf.eth_ip_alias")) {
						sprintf (line, "conf.eth_ip\t%d.%d.%d.%d\n",
								 ip_addr[0], ip_addr[1], ip_addr[2], ip_addr[3]);
						found_ip = TRUE;
					}
				}
				else if (strstr(line, "conf.eth_mask")) {
					if (!strstr(line, "conf.eth_mask_alias")) {
						sprintf (line, "conf.eth_mask\t%d.%d.%d.%d\n",
								 netmask[0], netmask[1], netmask[2], netmask[3]);
						found_mask = TRUE;
					}
				}
			}
		}
		if (fputs(line, tempfp) == EOF) {
			printDEBUG("appConfigEthInterface: s% writting config file\n", strerror(errno));
			goto error_exit2;
		}
	}

	if (!found_dhcp) {
		if (fputs("conf.dhcp_client\t0\n", tempfp) == EOF) {
			printDEBUG("appConfigEthInterface: s% writting config file\n", strerror(errno));
			goto error_exit2;
		}
	}
	if (!found_ip) {
		sprintf (line, "conf.eth_ip\t%d.%d.%d.%d\n",
				 ip_addr[0], ip_addr[1], ip_addr[2], ip_addr[3]);
		if (fputs(line, tempfp) == EOF) {
			printDEBUG("appConfigEthInterface: s% writting config file\n", strerror(errno));
			goto error_exit2;
		}
	}
	if (!found_mask) {
		sprintf (line, "conf.eth_mask\t%d.%d.%d.%d\n",
				 netmask[0], netmask[1], netmask[2], netmask[3]);
		if (fputs(line, tempfp) == EOF) {
			printDEBUG("appConfigEthInterface: s% writting config file\n", strerror(errno));
			goto error_exit2;
		}
	}

	fclose(tempfp);
	fclose(fp);

	if (gw_addr[0] + gw_addr[1] + gw_addr[2] + gw_addr[3]) {

		if ((fp = fopen("/etc/network/st_routes", "w")) < 0) {
			printDEBUG("appConfigEthInterface: %s opening st_routes file\n", strerror(errno));
			return -1;
		}

		sprintf(line, "/sbin/route add default gw %d.%d.%d.%d\n",
				gw_addr[0], gw_addr[1], gw_addr[2], gw_addr[3]);

		if (fputs(line, fp) == EOF) {
			printDEBUG("appConfigEthInterface: %s writting st_routes file\n", strerror(errno));
			goto error_exit1;
		}
	}

	fclose(fp);

	if (mknod("/var/aidp.lock", 0600, 0)) //makes "signal_ras" not to kill us!
		return -1;		
	if (system("mv /tmp/pslave.conf /etc/portslave > /dev/null 2>&1") ||
		system("/bin/signal_ras hup > /dev/null 2>&1")) {
		printDEBUG("appConfigEthInterface: problem applying configuration\n");
		retval = -1;
	}
	if (unlink("/var/aidp.lock")) //releases lock
		return -1;		
	
	return retval;

error_exit2:
	fclose(tempfp);
error_exit1:
	fclose(fp);

	return -1;
}


/////////////////////////////////////////////////////////////////////////////
//
//  Write new ethernet interface configuration to non-vol storage.
//
//  PARAMETERS: None
//  
//  RETURN: 0 for success, -1 for error
//
/////////////////////////////////////////////////////////////////////////////
int appSaveEthConfiguration(void)
{

	if (system("/bin/saveconf > /dev/null 2>&1")) {
		printDEBUG("appSaveEthConfiguration: problem saving configuration\n");
		return -1;
	}
	else
		return 0;
}
