/*
 *  Module:  aidpaux.c (Avocent Install and Discovery Protocol)
 *
 *  Author:  Marcelo Peccin <marcelo.peccin@avocent.com>
 *
 *  From the original:  Avocent Install and Discovery Protocol for DSR
 *             Author:  Jamie Collins (January 2002)
 *
 *  Copyright:  (c) 2006 Avocent Corporation
 *
 *  Date:  May 10, 2006
 *
 *  Comments:  Avocent Install and Discovery Protocol auxiliary functions.
 *             This module is complemented by another product-specific
 *             module that "resolves" some functions called here.
 *             See also aidp-XXX.c product-specific file.
 */

#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <linux/if_ether.h>  //ETH_P_IP
#include <net/if.h>          //struct ifreq
#include <linux/if_packet.h> //struct sockaddr_ll

#include "aidp.h"
#include "aidp-ts.h"

word appID;
byte appMAC[6];
byte appIP[]     = {0, 0, 0, 0};
byte appSubNet[] = {0, 0, 0, 0};
byte appGw[]     = {0, 0, 0, 0};

static byte dummySrcAddr[4] = {1,2,3,4};
static byte destEthAddr[6]; //loaded at process_socket_message()


//-----------------------------------------------------------------------
//    FUNCTION: initAppVars
//
// DESCRIPTION: Initialize appliance-related global variables.
//
//  PARAMETERS: None
//
//     RETURNS: 0 for success, -1 for error
//
//       NOTES: At AIDP daemon startup, some appliance's parameters must
//              be retrieved for future use.
//-----------------------------------------------------------------------
int initAppVars(void)
{
	// For this implementation, we only need to get appliance's model
	// and MAC address. IP, subnet mask and Gw are already
	// initialized to 0.0.0.0

	return appGetModelandMAC(&appID, appMAC);

/* #include <stdlib.h> */
/* #include <unistd.h> */
/* #include <time.h> */

/* byte ip_addr[]={172,26,39,82}; */
/* byte netmask[]={255,255,255,0}; */
/* byte gw_addr[]={172,26,39,5}; */
/* time_t timenow; */

/* 	sleep(60); */

/* 	if (time(&timenow) < 0) timenow = 0; */
/* 	printlog("DEBUG Session ------------------ %s", ctime(&timenow)); */

/* 	if (!appConfigEthInterface(ip_addr, netmask, gw_addr)) */
/* 		printlog("configure eth interface went ok\n"); */
/* 	else */
/* 		printlog("configure eth interface run into trouble\n"); */

/* 	if (time(&timenow) < 0) timenow = 0; */
/* 	printlog("DEBUG Session ------------------ %s", ctime(&timenow)); */

/* 	if (!appSaveEthConfiguration()) */
/* 		printlog("save to flash OK\n"); */
/* 	else */
/* 		printlog("save to flash not OK\n"); */

/* 	if (time(&timenow) < 0) timenow = 0; */
/* 	printlog("DEBUG Session ------------------ %s*** END ***\n", ctime(&timenow)); */

/* 	exit(0); */
}


void printPacketHdr(byte *recvBuf)
{
	int j;

	printlog("Dest Eth Addr: %02x %02x %02x %02x %02x %02x\n",
			 recvBuf[0], recvBuf[1], recvBuf[2], recvBuf[3], recvBuf[4], recvBuf[5]);
	printlog("Src  Eth Addr: %02x %02x %02x %02x %02x %02x\n",
			 recvBuf[6], recvBuf[7], recvBuf[8], recvBuf[9], recvBuf[10], recvBuf[11]);
	printlog("Ethernet Type: %02x %02x\n", recvBuf[12], recvBuf[13]);

	for (j = 14; j < 34; j += 4)
		printlog("IP Header (%d): %02x %02x %02x %02x\n",
				 (j-14)/4, recvBuf[j], recvBuf[j+1], recvBuf[j+2], recvBuf[j+3]);

	printlog("UDP Src  Port: %02x %02x\n", recvBuf[34], recvBuf[35]);
	printlog("UDP Dest Port: %02x %02x\n", recvBuf[36], recvBuf[37]);
	printlog("UDP Packt Len: %02x %02x\n", recvBuf[38], recvBuf[39]);
	printlog("UDP  checksum: %02x %02x\n", recvBuf[40], recvBuf[41]);
}


//-----------------------------------------------------------------------
//    FUNCTION: process_socket_message
//
// DESCRIPTION: Receive a raw Ethernet packet and provide the source MAC
//              address, source IP address, source UDP port and
//              the AIDP message from the packet.
//
//  PARAMETERS: psocket_data - "raw" Ethernet packet info
//
//     RETURNS: 0 for success, -1 for error
//
//       NOTES: Designed to operate the AIDP protocol over an
//              unconfigured network.
//-----------------------------------------------------------------------
int process_socket_message(AIDP_SOCKET_MESSAGE *psocket_data)
{
	int   retval;
	byte  recvBuf[500];
	word  dest_port;

	// Receive raw packet
	if ((retval = recvfrom(psocket_data->udp_socket, recvBuf, 500, 0,
						   (struct sockaddr *)NULL, (int *)NULL)) < 0)
		return -1;

	// If the destination port is not AIDP port or the protocol is not UDP,
	// then discard this message.
	dest_port = MAKE_WORD(((UDP_IP_PacketHdr *)recvBuf)->destPort[0],
						  ((UDP_IP_PacketHdr *)recvBuf)->destPort[1]);
	if (dest_port != AIDP_PORT)
		return -1;
	if (((UDP_IP_PacketHdr *)recvBuf)->protocol != UDP_PROTOCOL)
		return -1;

	if (debug) {
		printlog("process_socket_message: received %d bytes from raw socket\n", retval);
		printPacketHdr(recvBuf);
	}

	// Get the source IP Address and UDP port from the packet
	memcpy(psocket_data->recvfrom_ip, &((UDP_IP_PacketHdr *)recvBuf)->sourceIpAddr, 4);
	psocket_data->recvfrom_port = MAKE_WORD(((UDP_IP_PacketHdr *)recvBuf)->srcPort[0],
											((UDP_IP_PacketHdr *)recvBuf)->srcPort[1]);

	// Store the source ethernet address to be used as the
	// destination ethernet address in the reply.
	memcpy(destEthAddr, ((UDP_IP_PacketHdr *)recvBuf)->sourceEthAddr, 6);

	// Get the length of the data
	if ((psocket_data->reclen = MAKE_WORD(((UDP_IP_PacketHdr *)recvBuf)->pktLen[0],
										  ((UDP_IP_PacketHdr *)recvBuf)->pktLen[1]) - 8) <= 0) {
		printDEBUG("process_socket_message: no UDP data received\n");
		return -1;
	}
	if (psocket_data->reclen > 400) {
		printDEBUG("process_socket_message: data length too big\n");
		return -1;
	}

	printDEBUG("process_socket_message: reclen=%d from %x %x\n", psocket_data->reclen,
			   ((UDP_IP_PacketHdr *)recvBuf)->pktLen[0], ((UDP_IP_PacketHdr *)recvBuf)->pktLen[1]);

	memcpy(psocket_data->MsgData, &recvBuf[PACK_HDR_LEN], psocket_data->reclen);
	return 0;
}


//------------------------------------------------------------------------
//    FUNCTION: set_appliance_ip
//
// DESCRIPTION: Set appliance's IP Address, subnet mask and gateway IP
//              address for the Ethernet interface.
//
//  PARAMETERS: ip_addr, netmask, gw_addr
//
//     RETURNS: 0 for success, -1 for error
//
//       NOTES: Called by AIDP TESTIP messages to set appliance's
//              network parameters.
//------------------------------------------------------------------------
int set_appliance_ip(byte *ip_addr, byte *netmask, byte *gw_addr)
{
	int retval;

	if (debug) {
		printlog("set_appliance_ip:\n");
		printlog("IP Address: %d.%d.%d.%d\n", ip_addr[0], ip_addr[1], ip_addr[2], ip_addr[3]);
		printlog("Netmask   : %d.%d.%d.%d\n", netmask[0], netmask[1], netmask[2], netmask[3]);
		printlog("Gw Address: %d.%d.%d.%d\n", gw_addr[0], gw_addr[1], gw_addr[2], gw_addr[3]);
	}

	if ((retval = appConfigEthInterface(ip_addr, netmask, gw_addr))) {

		printDEBUG("set_appliance_ip: could not configure eth interface\n");
		
	} else {

		//store new values
		memcpy(appIP, ip_addr, 4);
		memcpy(appSubNet, netmask, 4);
		memcpy(appGw, gw_addr, 4);

	}

	return retval;
}


//------------------------------------------------------------------------
//    FUNCTION: update_appliance_ip
//
// DESCRIPTION: Store appliance's IP Address, subnet mask and gateway
//              IP address in non-vol storage.
//
//  PARAMETERS: ip_addr, netmask, gw_addr
//
//     RETURNS: 0 for success, -1 for error, -2 for inconsistency
//
//       NOTES: Called by AIDP SETIP message to inform the appliance that
//              permanent network parameters are now configured.
//------------------------------------------------------------------------
int update_appliance_ip(byte *ip_addr, byte *netmask, byte *gw_addr)
{
	// check if parameters are the expected ones
	if (memcmp(appIP, ip_addr, 4) ||
		memcmp(appSubNet, netmask, 4) ||
		memcmp(appGw, gw_addr, 4)) {

		printDEBUG("update_appliance_ip: not expected values\n");
		return -2;
	}

	printDEBUG("update_appliance_ip: writing network config to non-vol storage\n");

	return appSaveEthConfiguration();
}


static word ipCksum(word *addr, int len)
{
  int nleft = len;
  int sum = 0;
  word * w = addr;
  word answer = 0;
  
  /*
  ** Using a 32 bit sum, add sequential 16 bit words to it and at the
  ** end fold back all the carry bits from the top 16 bits into the
  ** lower 16 bits
  */
  while ( nleft > 1 )
  {
    sum += *w++;
    nleft -= 2;
  }
  
  // Mop up an odd byte, if necessary
  if ( nleft == 1 )
  {
    *(byte *)(& answer ) = *(byte *) w;
    sum += answer;
  }
  
  // Add back carry outs from top 16 bits to low 16 bits
  sum = (sum >> 16) + (sum & 0xffff); // Add hi 16 to low 16
  sum += (sum >> 16);                 // add carry
  answer = ~sum;                      // truncate to 16 bits
  return (answer);
}


word udpCksum(byte *buf, int datalen)
{
  byte cksbuf[255];
  word checksum;
  
  /*
  ** Create a buffer including the UDP pseudo-header in order to 
  ** calculate the UDP checksum.
  */
  
  // Add the UDP pseudo-header
  memcpy ( & cksbuf[ 0 ], & buf[ 26 ], 4 );  // source IP
  memcpy ( & cksbuf[ 4 ], & buf[ 30 ], 4 );  // dest IP
  cksbuf[8] = 0;                             // zero
  cksbuf[9] = buf[ 23 ];                     // protocol
  cksbuf[10] = buf[ 38 ];                    // UDP length
  cksbuf[11] = buf[ 39 ];
  
  // Add the UDP header and data
  memcpy ( & cksbuf[12], & buf[34], 8+datalen );
  
  // Pad with a zero at the end if needed
  if ( datalen % 2 )
  {
    cksbuf[ 20 + datalen ] = 0x0;
    checksum = ipCksum( (word *)cksbuf, 20+datalen+1 );
  }
  else
  {
    checksum = ipCksum( (word *)cksbuf, 20+datalen );
  }
  
  //avoPrintf(3, "UDP checksum is %x \n", checksum);
  return ( checksum );
}


int getIfIndex(int sock, char *ifname)
{
	struct ifreq ifr;

	memset(&ifr, 0, sizeof(struct ifreq));

	sprintf(ifr.ifr_name,"%s",ifname);

    if (ioctl(sock, SIOCGIFINDEX, (char *)&ifr) < 0) {
		printDEBUG("getIfIndex: %s getting interface index\n", strerror(errno));
		return -1;
	}

	return ifr.ifr_ifindex;
}


//---------------------------------------------------------------------
//    FUNCTION: process_send
//
// DESCRIPTION: Send raw packet using an unconfigured network.
//
//  PARAMETERS: psocket_data - "raw" Ethernet packet info
//
//     RETURNS: number of bytes sent for success, -1 for error
//
//       NOTES:  
//---------------------------------------------------------------------
int process_send(AIDP_SOCKET_MESSAGE *psocket_data, byte *sendbuf, int sendlen)
{
	int retval;
	static int ifindex = -1;
	struct sockaddr_ll eth_addr;
	UDP_IP_PacketHdr hdr;
	byte sendBuffer[256];
	word IP_cksum, UDP_cksum;

	memset(sendBuffer, 0, 256);
	memset(&eth_addr, 0, sizeof(struct sockaddr_ll));

	if (ifindex == -1) {
		if ((ifindex = getIfIndex(psocket_data->udp_socket, APP_ETHER_DEV_NAME)) < 0) {
			printDEBUG("process_send: could not get interface index\n");
			return -1;
		}
	}

	// Fill in the link layer header information
	eth_addr.sll_family = AF_PACKET;
	eth_addr.sll_protocol = ETH_P_IP;
	eth_addr.sll_ifindex = ifindex;

	// DLC header
	memcpy(hdr.destEthAddr, destEthAddr, 6 ); // dest MAC addr
	memcpy(hdr.sourceEthAddr, appMAC, 6 ); // source MAC addr
	hdr.ethType[0] = 0x08; // ETH_P_IP
	hdr.ethType[1] = 0x00;

	// IP Header
	hdr.vers_hdr_len = 0x45;   // IP version 4, header length 20 bytes
	hdr.tos = 0x00;            // normal service
	hdr.len[0] = 0x00;         // IP packet length, not including DLC hdr
	hdr.len[1] = 20+8+sendlen;
	hdr.ID[0] = 0x34;          // Identification
	hdr.ID[1] = 0x01;
	hdr.flags_offs[0] = 0x00;  // flags/fragment offset
	hdr.flags_offs[1] = 0x00;
	hdr.ttl = 0x80;             // time to live = 128 seconds/hops
	hdr.protocol = 0x11;        // UDP protocol
	hdr.checksum[0] = 0x00;     // IP checksum (see below)
	hdr.checksum[1] = 0x00;
	memcpy(hdr.sourceIpAddr, dummySrcAddr, 4);            // source IP address
	memcpy(hdr.destIpAddr, psocket_data->recvfrom_ip, 4); // dest IP address

	// UDP Header
	hdr.srcPort[0] = (AIDP_PORT) >> 8;    // UDP source port = AIDP_PORT
	hdr.srcPort[1] = (AIDP_PORT) & 0x00FF;
	hdr.destPort[0] = htons( psocket_data->recvfrom_port ) >> 8;
	hdr.destPort[1] = htons( psocket_data->recvfrom_port ) & 0x00FF;
	hdr.pktLen[0] = 0x00; // UDP packet length including UDP hdr
	hdr.pktLen[1] = ((byte)sendlen) + 8;
	hdr.UDPchecksum[0] = 0x00; // UDP checksum (see below)
	hdr.UDPchecksum[1] = 0x00;

	// Copy in the Header
	memcpy(sendBuffer, &hdr, PACK_HDR_LEN);

	// Calculate the IP checksum
	IP_cksum = ipCksum((word *)&sendBuffer[14], 20);
	sendBuffer[24] = (byte)(IP_cksum >> 8);
	sendBuffer[25] = (byte)(IP_cksum & 0x00FF); // IP checksum

	// Copy in the Data
	memcpy(&sendBuffer[PACK_HDR_LEN], sendbuf, sendlen);

	// Calculate the UDP checksum
	UDP_cksum = udpCksum(sendBuffer, sendlen);
	sendBuffer[40] = (byte)(UDP_cksum >> 8);
	sendBuffer[41] = (byte)(UDP_cksum & 0x00FF); // UDP checksum

	if (debug) {
		int i,j;
		printlog("---------------\n");
		printPacketHdr(sendBuffer);
		for (i=0; i<sendlen; ) {
			printlog("\n");
			for (j=0; j<8; j++) {
				printlog("%02x ", sendBuffer[PACK_HDR_LEN+(i++)]);
				if (i==sendlen) break;
			}
		}
		printlog("\n-----------------------------------------\n");
	}

	// Send the raw packet
	if ((retval = sendto(psocket_data->udp_socket, sendBuffer, PACK_HDR_LEN+sendlen, 0,
						 (struct sockaddr *)&(eth_addr), sizeof(struct sockaddr_ll))) < 0) {
		printDEBUG("process_send: error sending data: %s\n", strerror(errno));
		return -1;
	}

	return (retval-PACK_HDR_LEN);
}


//-----------------------------------------------------------------------
//    FUNCTION: kill_other_dynamicIP_daemons
//
// DESCRIPTION: Kill other concurrent "dinamic IP daemons" that
//              may be running in the appliance.
//
//  PARAMETERS: None
//
//     RETURNS: 0 for success, -1 for error
//
//       NOTES: Called when AIDP daemon is exiting. Since appliance's IP
//              is now set, we do not want other daemons like "dhcpcd"
//              keep running.
//-----------------------------------------------------------------------
int kill_other_dynamicIP_daemons(void)
{
	// For this implementation, we only need to call a product-specific
	// function that knows how to solve this problem.

	return appKillDynamicIPdaemons();
}
