
#include <stddef.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/select.h>

// PortSlave includes
#include <server.h>
#include <rot_nports.h>

#define EVENT_NOTIF_C	1
#include "trap_oids.h"
#include "event_msgs.h"
#include "event_conf.h"

//[RK]Aug/07/06 - port from Onboard
#define BUFFERSIZE	2048+255 // size of buffer to read data from SNMP_PIPE

int snmp_pipe;
int evt_pipe;

// syslog message base
// <166><sp><timestamp><sp><ACS IP addr><sp>EVT[<evt number>]:
unsigned char syslog_base[] = "<166> %s %s EVT[%d]: ";

// appliance type
unsigned char appliance_type;

// signal interruption
int got_signal;

int debug = 0; 

struct cyc_info *cyc = NULL;

static void got_interruption(int signal)
{
	struct sigaction act;
	memset((char*)&act,0,sizeof(struct sigaction));

	switch (signal) {
		case SIGCHLD :
		case SIGHUP :
		case SIGUSR1 :
			got_signal = signal;
			break;
			
                default :
                        syslog(LOG_NOTICE,"Receive Signal[%d] %d", getpid(),signal);
                        got_signal = signal;
                        break;
        }

	act.sa_handler = got_interruption;
	sigaction(signal, &act, 0);
	unblock(signal);
}

// read the EVENT_CONF_FILE and set the global variables
static void read_event_conf(void)
{
	FILE *fd;
	char *line, *pli, *pe;
	int i, destn, trapn;

	// default values 
	syslog_msg_enabled = 1; // syslog enabled
	memset(generate_status,0, NUM_ENTERPRISE_TRAPS+1);
	if (community) {
		free(community);
	}
	
	for (i=0; i < MAX_TRAP_DESTINATIONS; i++) {
		if (dest_traps[i]) {
			free(dest_traps[i]);
			dest_traps[i] = NULL;
		}
	}

	if ((fd = fopen(EVENT_CONF_FILE, "r")) == NULL) {
		if (debug) syslog(4,"%s: fopen EVENT_CONF_FILE : %m",__func__);
		return;
	}

	destn = 0;

	// allocate memory for one line with 80 chars
	line = calloc(81,1);
	
	// read a line
	while (fgets(line, 81, fd) != NULL) {
		if (line[0] == '#') continue;
		if (strncmp("SYSLOG_MSG_ENABLE",line,17) == 0) {
			pli = strstr(line,"=");
			if (pli == NULL) continue;
			while (*pli && isspace(*pli)) pli++;
			if (strncmp(pli+1,"YES",3) == 0) {
				syslog_msg_enabled = 1; // syslog enabled
			} else {
				syslog_msg_enabled = 0; // syslog disabled
			}
			continue;
		}
		if (strncmp("destination",line,11) == 0) {
			pe = strstr(line,"=");
			if (!pe) continue;
			pe++;
			while (*pe && (isspace(*pe))) pe++;
			if (*pe == 0x00) continue;
			pli=pe;
			while (*pe && !(isspace(*pe))) pe++;
			if (*pe) *pe=0x00;
			dest_traps[destn++] = strdup(pli); 
		}
		if (strncmp("trapCommunity",line,13) == 0) {
			pe = strstr(line,"=");
			if (!pe) continue;
			pe++;
			while (*pe && (isspace(*pe))) pe++;
			pli=pe;
			while (*pe && !(isspace(*pe))) pe++;
			if (*pe) *pe=0x00;
			community = strdup(pli); 
		}
		if ((pli = strstr(line,"="))) {
			*pli++ = 0x00;
			trapn = atoi(line);
			while (*pli && !(isdigit(*pli))) pli++;
			if (!*pli) continue;
			if ((trapn > 0) && (trapn <= NUM_ENTERPRISE_TRAPS)) {
				generate_status[trapn] = atoi(pli);
			}
		}
	}
	fclose (fd);
}

// mount the syslog message and write it in the evt_pipe fifo device
// format : <166><sp><data yyyy-mm-ddThh:mm:ssZ><sp>< ACS IP ><sp>EVT[<event number>]:<event message>\n
static void send_syslog_msg(int evtn, time_t evttime, int args[])
{
	char * msg;
	char timestamp[21];
	struct tm * tms;
	char ipaddr[16];
	int nargs = event_snmptrap[evtn-1].variables_len;
	struct EVENT_SYSLOG *psys = &event_syslog[evtn - 1];
	int size,max;
	char format[100],*fmt;
	int i;
	
	if ((tms = gmtime(&evttime))) {
		sprintf(timestamp,"%d-%d-%dT%d:%d:%dZ",(tms->tm_year+1900),(tms->tm_mon+1),
			tms->tm_mday,tms->tm_hour, tms->tm_min, tms->tm_sec);
	} else {
		timestamp[0] = 0;
	}

	{
		union {
			unsigned int naddr;
			char baddr[4];
		} myaddr;
		myaddr.naddr = cy_shm->main_conf.ipno;
		sprintf(ipaddr,"%d.%d.%d.%d",myaddr.baddr[0],myaddr.baddr[1],myaddr.baddr[2],myaddr.baddr[3]);
	}
	
	msg = calloc(MAX_SYSLOG_MSG+1,1);

	size = sprintf(msg,syslog_base,timestamp,ipaddr,evtn);

	max = MAX_SYSLOG_MSG - size;

	if (!syslog_msg_enabled) {
		memset(format,0,100);
		for (i=0; i < nargs; i++) {
			if (i) strcat(format,",");
			switch (event_snmptrap[evtn-1].variable_type[i]) {
				case TYPE_NUMBER :
					strcat(format,"%d");
					break;
				case TYPE_STRING :
					strcat(format,"%s");
					break;
			}
		}
		fmt = format;
	} else {
		fmt = psys->summary;
	}

	if (debug) syslog(4,"%s : format[%s]",__func__,fmt);

	i = 0;  //[RK]Aug/07/06 - port from Onboard

	switch (nargs) {

		case 0 : break; // no arguments

		case 1 : // 1 argument
			i = snprintf(msg+size,max,fmt, args[0]);
			break;
		case 2 :
			i = snprintf(msg+size,max,fmt, args[psys->arguments[0]],
					args[psys->arguments[1]]);
			break;
		case 3 :
			i = snprintf(msg+size,max,fmt, args[psys->arguments[0]],
					args[psys->arguments[1]],
					args[psys->arguments[2]]);
			break;
		case 4 :
			i = snprintf(msg+size,max,fmt, args[psys->arguments[0]],
					args[psys->arguments[1]],
					args[psys->arguments[2]],
					args[psys->arguments[3]]);
			break;
		case 5 :
			i = snprintf(msg+size,max,fmt, args[psys->arguments[0]],
					args[psys->arguments[1]],
					args[psys->arguments[2]],
					args[psys->arguments[3]],
					args[psys->arguments[4]]);
			break;
		case 6 :
			i = snprintf(msg+size,max,fmt, args[psys->arguments[0]],
					args[psys->arguments[1]],
					args[psys->arguments[2]],
					args[psys->arguments[3]],
					args[psys->arguments[4]],
					args[psys->arguments[5]]);
			break;
	}

	size += (i >= max) ? max : i; //[RK]Aug/07/06 - port from onboard

	// include the ended of the syslog message : \n
	if ((size + 2) >= MAX_SYSLOG_MSG) {
		msg[size - 2] = '\n';
		msg[size - 1] = 0x00;
	} else {
		msg[size++] = '\n';
		msg[size] = 0x00;
	}	
	
	// write the message in the evt_pipe fifo device
	write(evt_pipe,msg,size);
	fsync(evt_pipe);
	free(msg);
}

// mount the command with the snmptrap message
// snmptrap -v 1 -c <community> <destination> <ACS model> "" 6 <specific trap> '' [OIDs]
static void send_snmptrap_msg(int evtn, int args[])
{
	struct EVENT_SNMPTrapOID *ptrap;
	int i,j,cnt;
	char *argv[100];
	char oid[50];
	int chlpid;
	
	ptrap = &event_snmptrap[evtn-1];

	// fill the args 
	argv[0] = strdup("/bin/snmptrap");
	argv[1] = strdup("-v");
	argv[2] = strdup("1");
	argv[3] = strdup("-c");
	argv[4] = strdup(community);

	// argv[5] ==> destination

	// enterprise OID = cyMgmt.cyModel.cyAcsXX
	sprintf(oid, "%s.%d.%d", cyMgmt, cyModel,appliance_type);

	argv[6] = strdup(oid);

	argv[7] = strdup(""); // hostname 
	argv[8] = strdup("6"); // enterprise trap
	
	sprintf(oid,"%d",evtn);
	argv[9] = strdup(oid);	// specific trap

	argv[10] = strdup("");	// up time

	cnt = 11;

	// append OIDs
	for (i=0; i < ptrap->variables_len ; i++) {
		sprintf(oid,"%s.%d.%d",cyMgmt,cyTrapObject,ptrap->variables[i]);
		argv[cnt++] = strdup(oid);
		switch (ptrap->variable_type[i]) {
			case TYPE_NUMBER :
				argv[cnt++] = strdup("i");
				sprintf(oid,"%d",(int)args[i]);
				argv[cnt++] = strdup(oid);
				break;
			case TYPE_STRING :
				argv[cnt++] = strdup("s");
				argv[cnt++] = strdup((char *)args[i]);
				break;
		}
	}
	
	argv[cnt] = NULL;

	for (i=0; i < MAX_TRAP_DESTINATIONS; i++) {
		if (dest_traps[i] == NULL) break;
		argv[5] = strdup(dest_traps[i]);
		if ((chlpid=fork()) == 0) {
			for (j=3; j < 64; j++) close(j);
			j = execve("/bin/snmptrap",argv,NULL);
			if (j == -1)
				syslog(4,"event_gen execve [%m]");
			exit(1);
		}
		if (chlpid == -1) 
			syslog(3,"%s fork error : %m",__func__);
		wait(NULL);
	}
	for (i=0; i < cnt; i++) free(argv[i]);
}

// return the number pointed by pnum
// pnum will be pointer to the first position after the space char
// return -1 when find the EOL 
static int parse_number(char **num)
{
	int number;
	char *pnum = *num;

	number = atoi(pnum);
	
	while (*pnum && !isspace(*pnum)) pnum++;
	if (*pnum == 0) return -1;
	pnum++; // skip the space char	
	*num = pnum;
	return (number);
}

// return a pointer to the string
// pstr will be pointer to the first position after the space char
// return NULL when EOL
static char * parse_string(char ** str)
{
	int size, cnt;
	char * pb;
	char * pstr = *str;

	if ((size = parse_number(&pstr)) == -1) {
		return NULL; // do not get the string size
	}

	pb = pstr;
	
	cnt = 0;

	while (*pstr && (cnt < size) && isascii(*pstr)) {
		pstr++;
		cnt++;
	}

	if ((*pstr != ' ') || (cnt != size))  {
		return NULL;
	}

	*pstr++ = 0x00; // change space by null

	*str = pstr;
	return (pb);
}

// treat the event line
// do the parse of the event
// format : EVT<event number><sp><time of the event><sp>[<parameters>]<sp>
// send the notification

static void treat_event_line(char * pevt_line)
{
	char * pevt = pevt_line;
	int evtn;
	time_t evttime;

	struct EVENT_SNMPTrapOID *ptrap;

	int argn, aux;
	int pargs[MAX_ARGUMENTS] ;

	if (strncmp(pevt,"EVT",3)) {
		goto TREAT_EVENT_ERROR; 
	}
	pevt += 3;

	// event number
	if (((evtn = parse_number(&pevt)) == -1) ||
	    ((evtn == 0) && (evtn > sizeof(event_internal)))) {
		goto TREAT_EVENT_ERROR; 
	}

	// find time of the event
	if ((evttime = parse_number(&pevt)) == -1) { 
		goto TREAT_EVENT_ERROR; 
	}

	evtn = event_internal[evtn-1];

	if (evtn == -1) {
		goto TREAT_EVENT_ERROR;
	}

	ptrap = &event_snmptrap[evtn-1];

	if (debug) syslog(4,"%s : evtn=%d time=%d nvar=%d generate_status=%d",__func__,evtn,(int)evttime,ptrap->variables_len,generate_status[evtn]);

	// read the arguments
	for (argn=0; argn < ptrap->variables_len ; argn++) {

		if (*pevt == 0x00) {
			// the event need more arguments
			goto TREAT_EVENT_ERROR;
		}

		switch (ptrap->variable_type[argn]) {
			case TYPE_NUMBER :
				if (( aux  = parse_number(&pevt)) == -1) {
					goto TREAT_EVENT_ERROR;
				}
				pargs[argn] = aux;
				break;

			case TYPE_STRING :
				if (((char *)pargs[argn] = parse_string(&pevt)) == NULL) {
					goto TREAT_EVENT_ERROR;
				}
				break;
		}
	}

	// send the notifications ==> syslog message
	if (generate_status[evtn] & GENERATE_SYSLOG_MESSAGE) {
		send_syslog_msg(evtn,evttime, pargs);
	}

	// send the notifications ==> snmptrap message
	if ((generate_status[evtn] & GENERATE_SNMPTRAP) && (dest_traps[0]!=NULL)) {
		send_snmptrap_msg(evtn, pargs);
	}
	
	return;

TREAT_EVENT_ERROR:
	// invalid event ==> ignore line
	syslog(4,"%s : invalid event [%s]\n",__func__,pevt_line);
	return;
}

int main(int argc, char **argv)
{
	int count,offset;
	char *buffer, *pline=0, *pend;

	fd_set rfds;
	int retval;

	// redefines interruptions
	struct sigaction sa;

	if (argc > 1) {
		if (strcmp(argv[1],"-d")== 0) 
			debug = 1;
	}

	memset((char*)&sa,0,sizeof(sigaction));
	sa.sa_handler = got_interruption;
	for (count = 1; count < 32; count++) {
		if (count == SIGPIPE) continue;
		if (count == SIGWINCH) continue;
		if (count == SIGSEGV) continue;
		if (count == SIGFPE) continue;
		if (count == SIGILL) continue;
		if (count == SIGKILL) continue;
		if (count == SIGCHLD) continue;
		sigaction(count,  &sa, NULL);
	}	
	signal(SIGCHLD,SIG_IGN);
	
	got_signal = 0;

	// shared used to get the Appliance IP address
	if (shm_init(-1) < 0) {
		return -1;
	}

	// read the machine type
	{
		struct MachineType mt;
		get_machine_type(&mt);
		appliance_type = mt.nports;
	}

	// read the cpnfiguration files
	read_event_conf();

	// open fifo devices
	//snmp_pipe = open(SNMP_PIPE, O_RDWR | O_NONBLOCK | O_NOCTTY);
	//snmp_pipe = open(SNMP_PIPE, O_RDWR | O_NOCTTY);
	snmp_pipe = open(SNMP_PIPE, O_RDWR | O_NOCTTY | O_NONBLOCK);
	evt_pipe  = open(EVT_PIPE, O_RDWR | O_NONBLOCK | O_NOCTTY);

	buffer = calloc(BUFFERSIZE,1);
	count = offset = 0;

	// loop waiting messages from applications
	while(1) {
		FD_ZERO(&rfds);
 		FD_SET(snmp_pipe,&rfds);
  
 		retval = select(snmp_pipe+1, &rfds,NULL, NULL, NULL);
 		
 		//[RK]Jun/02/06 - fixed problem with restart
 		if (got_signal) {
 			switch (got_signal) {
 				case SIGHUP :
 					// read the configuration files
 					read_event_conf();
 					got_signal = 0;
 					close(evt_pipe);
 					evt_pipe  = open(EVT_PIPE, O_RDWR | O_NONBLOCK | O_NOCTTY);
 					break;
 				
 				case SIGUSR1 :
 					debug = !debug;
 					got_signal = 0;
 					break;
 
 				case SIGCHLD : 
 					got_signal = 0;
 					break;
 			
				case SIGTERM :
					exit(0);

 				default :
 					syslog(4," select error [%m]");
 					exit(-1);
 			}
 		}
 	
 		if (retval <= 0) continue;
 		
 		count = read (snmp_pipe,&buffer[offset], BUFFERSIZE-1-offset);
 		buffer[BUFFERSIZE] = 0x00;

		if (count <= 0) continue;

		// treat the data received from applications
		// format : EVT###<sp><time><sp>[<parameters>]<sp>\n
		
		count += offset; // included the initial buffer

		offset = 0;
		while (count > 0) {
			// find the event
			pend = pline = &buffer[offset];
			while (*pend && (*pend != '\n')) pend++;
			if (*pend == 0) {
				// event is not completed
				offset = count;
				break;
			}
			*pend = 0x00;
			if (debug) syslog(4,"event_gen : line=[%s]",pline);
			treat_event_line(pline);
			offset += (pend - pline) + 1;
			count -= (pend - pline) + 1;
		}
		if (count > 0) {
			// reallocate the data to the begin of the buffer
			strncpy(buffer,pline,count);
			offset = count;
		} else {
			offset = 0;
		}
	}
}

