/*
 *  8/00  -  CPM SPI works, so just use that
 *
 *  Copyright (c) 2000  Montavista Software, Inc <source@mvista.com>
 *  based on the following
 *
 */

/*-----------------------------------------------------------------------------
 *
 * Copyright (C) Cyclades Corporation, 1997 - 2000
 *
 * This file :      cy_spi.c
 *
 * Description:     This file contains the routines to access for SPI protocol
 *
 * Complies with Cyclades SW coding Standard rev 1.2
 *-----------------------------------------------------------------------------
 */

 /*----------------------------------------------------------------------------
  * Change History
  *----------------------------------------------------------------------------
  * 11/8/01   Arnaldo Zimmermann     Initial implementation
  *----------------------------------------------------------------------------
  * Aug/09/02 Edson Seabra           Port to linux kernel
  *----------------------------------------------------------------------------
  */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/time.h>

#include <asm/uaccess.h>
#include <asm/mpc8xx.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/prxk.h>

#ifdef CONFIG_SM100
#include <linux/mc146818rtc.h>
#include <linux/delay.h>
#include "commproc.h"
#else
#include <asm/delay.h>
#include <asm/commproc.h>
#include <asm/time.h>
#include <asm/todc.h>
#endif

struct spi_ctl {
	int		active;
	cbd_t		*rbdf;
	cbd_t		*tbdf;
};

static struct spi_ctl tspi;


static void
msdelay(uint t) 
{
	uint i;
	/* delay t miliseconds */
	for (i=0; i<t; i++) {
		udelay(1000);
	}
}

void __init
cyc_cpm_spi_init(void)
{
	uint	mem_addr, dp_addr, reloc;
	volatile cpm8xx_t	*cp;
	volatile spi_t		*spp;
	volatile immap_t	*immap;

	cp = cpmp;	/* Get pointer to Communication Processor */
	immap = (immap_t *)IMAP_ADDR;	/* and to internal registers */

	spp = (spi_t *)&cp->cp_dparam[PROFF_SPI];

	/* Check for and use a microcode relocation patch.
	*/
	if ((reloc = spp->spi_rpbase))
		spp = (spi_t *)&cp->cp_dpmem[spp->spi_rpbase];
		

	/* Initialize Port B SPI pins.
	*/
#ifdef CONFIG_SM100
	cp->cp_pbpar |=  0x0000000E;
	cp->cp_pbdir |=  0x0000000F;
	cp->cp_pbodr &= ~0x0000000E;
	cp->cp_pbdat |=  0x00000001;
#else
	cp->cp_pbpar &= ~0x00020000;
	cp->cp_pbpar |=  0x0000000E;
	cp->cp_pbdir |=  0x0002000F;
	cp->cp_pbodr &= ~0x0000000E;
	cp->cp_pbdat |=  0x00000001;
#endif

	/* Initialize the parameter ram.
	 * We need to make sure many things are initialized to zero,
	 * especially in the case of a microcode patch.
	 */
	spp->spi_rstate = 0;
	spp->spi_rdp = 0;
	spp->spi_rbptr = 0;
	spp->spi_rbc = 0;
	spp->spi_rxtmp = 0;
	spp->spi_tstate = 0;
	spp->spi_tdp = 0;
	spp->spi_tbptr = 0;
	spp->spi_tbc = 0;
	spp->spi_txtmp = 0;


	/* Allocate space for one transmit and one receive buffer
	 * descriptor in the DP ram.
	 */
	dp_addr = m8xx_cpm_dpalloc(sizeof(cbd_t) * 2);
	if (dp_addr == CPM_DP_NOSPACE) {
		panic("Warning! Failed to allocate tx/rx buffer on DP ram for SPI");
	}
	
	/* Allocate space for receive buffer in the host memory.
	 * 
	 */

	/* Set up the SPI parameters in the parameter ram.
	*/
	spp->spi_tbase = dp_addr;
	spp->spi_rbase = dp_addr + sizeof(cbd_t);

	/*
	 *  manually init if using microcode patch
	 */
	if(reloc)
	{
		spp->spi_rbptr = spp->spi_rbase;
		spp->spi_tbptr = spp->spi_tbase;
	}

	spp->spi_tfcr = 0x18;
	spp->spi_rfcr = 0x18;

	/* Set maximum receive size.
	*/
	spp->spi_mrblr = 32;

	/* Initialize Tx/Rx parameters.
	*/
	if (reloc == 0) {
		cp->cp_cpcr =
			mk_cr_cmd(CPM_CR_CH_SPI, CPM_CR_INIT_TRX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
	}
	/* The notes on the SPI relocation patch indicate the the init rx and 
	   tx command doesn't work.  It isn't clear if this also includes 
	   the individual rx and tx init commands, but at any rate it seems 
	   to work fine if we just skip these commands.
	*/
	else {
		cp->cp_cpcr =
			mk_cr_cmd(CPM_CR_CH_SPI, CPM_CR_INIT_TX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);

		cp->cp_cpcr =
			mk_cr_cmd(CPM_CR_CH_SPI, CPM_CR_INIT_RX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
	}

	/* Set the buffer address.
	*/
	mem_addr = m8xx_cpm_hostalloc(32);
	if (!mem_addr) {
		panic("Warning! Failed to allocate rx buffer on CPM ram for SPI");
	}
	tspi.rbdf = (cbd_t *)&cp->cp_dpmem[spp->spi_rbase];
	tspi.rbdf->cbd_bufaddr = __pa(mem_addr);

	mem_addr = m8xx_cpm_hostalloc(32);
	if (!mem_addr) {
		panic("Warning! Failed to allocate tx buffer on CPM ram for SPI");
	}
	tspi.tbdf = (cbd_t *)&cp->cp_dpmem[spp->spi_tbase];
	tspi.tbdf->cbd_bufaddr = __pa(mem_addr);

	/* Clear event
	*/
	cp->cp_spie = ~0;

	/* No events 
	*/
	cp->cp_spim = 0;


	/*
     *  62.5KHz clock for the A/D
     *  8 bit 
     *  The datasheet specs for the ADS7843 indicate that it should work 
     *  with a clock rate of up to 2MHz, but we don't seem to be able to 
     *  get reliable touchpanel samples with a clock much faster than 
     *  62.5KHz.  This translates to a sample time of 768 microseconds 
     *  to sample both the X and Y coordinates.
     */

    cp->cp_spmode = SPMODE_CI | SPMODE_CP | SPMODE_REV | SPMODE_MSTR | 
			SPMODE_DIV16 | ((8-1) << 4) | 0xF | SPMODE_EN;

	printk("SPI init done.\r\n");
}


/* Read from SPI.
 * This is a two step process.  First, we send the "dummy" write
 * to set the device offset for the read.  Second, we perform
 * the read operation.
 */

ssize_t
cyc_cpm_spi_read(char *buf, size_t count)
{
	volatile spi_t *spp;
	volatile cbd_t	*tbdf, *rbdf;
	volatile cpm8xx_t	*cp;
	unsigned long flags;
	uint	reloc, i;
 	volatile sysconf8xx_t *psiu = &(((immap_t *)IMAP_ADDR)->im_siu_conf);

	cp = cpmp;	/* Get pointer to Communication Processor */
	spp = (spi_t *)&cp->cp_dparam[PROFF_SPI];

	/* Check for and use a microcode relocation patch.
	*/
	if (reloc = spp->spi_rpbase) {
		spp = (spi_t *)&cp->cp_dpmem[spp->spi_rpbase];
	}

	if (count >= spp->spi_mrblr) {
		return 0;
	}

	tbdf = (cbd_t *)&cp->cp_dpmem[spp->spi_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[spp->spi_rbase];

	memcpy(__va(tbdf->cbd_bufaddr), buf, count);
	*(char*)(__va(tbdf->cbd_bufaddr)) &= ~0x80;		/* read command */
	memset(__va(rbdf->cbd_bufaddr),0,count);

	tbdf->cbd_datlen = count;
	tbdf->cbd_sc = BD_SC_READY|BD_SC_WRAP|BD_SC_INTRPT|BD_SC_LAST; 

	rbdf->cbd_datlen = 0;
	rbdf->cbd_sc = BD_SC_EMPTY|BD_SC_WRAP|BD_SC_INTRPT;


	/* set enable here.
	*/
	cp->cp_pbdat &=  ~0x00000001;
	msdelay(2);

	save_flags(flags); cli();

	cp->cp_spie = ~0;
	cp->cp_spcom |= 0x80;	/* Start Transmit */

	/* Wait for SPI transfer.
	*/
	i = 10000000;
	while(!(cp->cp_spie & 0x35)) {
	 	psiu->sc_swsr = 0x556c;
		psiu->sc_swsr = 0xaa39;
		i--;
		if (!i) {
			break;
		}
	}

	restore_flags(flags);

	cp->cp_pbdat |=   0x00000001;

	if (cp->cp_spie != 0x03) {
		printk("\r\nSPI error on reading RTC. Status = 0x%04x", cp->cp_spie);
		return 0;
	}

	if (signal_pending(current))
		return -ERESTARTSYS;

	if (rbdf->cbd_datlen != count) {
		printk("cyc_cpm_spi_read %d of %d\n", rbdf->cbd_datlen, count);
	}

	if (rbdf->cbd_sc & BD_SC_EMPTY) {
		printk("SPI read complete but rbuf empty\n");
	}
	
	memcpy(buf,__va(rbdf->cbd_bufaddr), count);

	//printk("RTC read: ");
	//for (i = 0; i<count; i++) printk("%02x", (unsigned char)buf[i]);
	//printk("\r\n");
	
	return rbdf->cbd_datlen;
}


ssize_t
cyc_cpm_spi_write(char *buf, size_t count)
{
	volatile spi_t *spp;
	volatile cbd_t	*tbdf, *rbdf;
	volatile cpm8xx_t	*cp;
	unsigned long flags;
	uint	reloc, i, status;
 	volatile sysconf8xx_t *psiu = &(((immap_t *)IMAP_ADDR)->im_siu_conf);

	cp = cpmp;	/* Get pointer to Communication Processor */
	spp = (spi_t *)&cp->cp_dparam[PROFF_SPI];

	/* Check for and use a microcode relocation patch.
	*/
	if ((reloc = spp->spi_rpbase)) {
		spp = (spi_t *)&cp->cp_dpmem[spp->spi_rpbase];
	}

	if (count >= spp->spi_mrblr) {
		return 0;
	}

	tbdf = (cbd_t *)&cp->cp_dpmem[spp->spi_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[spp->spi_rbase];

	//printk("RTC write: ");
	//for (i = 0; i<count; i++) printk("%02x", (unsigned char)buf[i]);
	//printk("\r\n");

	memcpy(__va(tbdf->cbd_bufaddr), buf, count);
	*(char*)(__va(tbdf->cbd_bufaddr)) |= 0x80;	/* write command */
	memset(__va(rbdf->cbd_bufaddr),0,count);

	tbdf->cbd_datlen = count;		/* Length */
	rbdf->cbd_datlen = 0;			/* Length */

	tbdf->cbd_sc = BD_SC_READY|BD_SC_WRAP|BD_SC_INTRPT|BD_SC_LAST; 
	rbdf->cbd_sc = BD_SC_EMPTY|BD_SC_WRAP|BD_SC_INTRPT;

	cp->cp_pbdat &=  ~0x00000001;
	msdelay(2);

	save_flags(flags); cli();

	cp->cp_spie = ~0;
	cp->cp_spcom = 0x80;

	/* Wait for SPI transfer.
	*/
	i = 100000;
	while(!(status = (cp->cp_spie & 0x32))) {
	 	psiu->sc_swsr = 0x556c;
		psiu->sc_swsr = 0xaa39;
		i--;
		if (!i) {
			break;
		}
	}

	restore_flags(flags);

	cp->cp_pbdat |=   0x00000001;

	if (signal_pending(current))
		return -ERESTARTSYS;

	if (status != 0x02) {				// check if TxB is set w/o other errors
		printk("\r\nSPI error when reading RTC");
		return 0;
	}

	if (tbdf->cbd_sc & BD_SC_READY)
		printk("SPI ra complete but tbuf ready\n");

	return count;
}

