/*
 * Fast Ethernet Controller (FEC) driver for Motorola MPC8xx.
 * Copyright (c) 1997 Dan Malek (dmalek@jlc.net)
 *
 * This version of the driver is specific to the FADS implementation,
 * since the board contains control registers external to the processor
 * for the control of the LevelOne LXT970 transceiver.  The MPC860T manual
 * describes connections using the internal parallel port I/O, which
 * is basically all of Port D.
 *
 * Right now, I am very watseful with the buffers.  I allocate memory
 * pages and then divide them into 2K frame buffers.  This way I know I
 * have buffers large enough to hold one frame within one buffer descriptor.
 * Once I get this working, I will use 64 or 128 byte CPM buffers, which
 * will be much more memory efficient and will easily handle lots of
 * small packets.
 *
 */
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <asm/8xx_immap.h>
#include <asm/pgtable.h>
#include <asm/mpc8xx.h>
#include <asm/irq.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>
#include "commproc.h"

/* The number of Tx and Rx buffers.  These are allocated from the page
 * pool.  The code may assume these are power of two, so it it best
 * to keep them that size.
 * We don't need to allocate pages for the transmitter.  We just use
 * the skbuffer directly.
 */
#if 0
#define FEC_ENET_RX_PAGES	4
#define FEC_ENET_RX_FRSIZE	2048
#define FEC_ENET_RX_FRPPG	(PAGE_SIZE / FEC_ENET_RX_FRSIZE)
#define RX_RING_SIZE		(FEC_ENET_RX_FRPPG * FEC_ENET_RX_PAGES)
#define TX_RING_SIZE		8	/* Must be power of two */
#define TX_RING_MOD_MASK	7	/*   for this to work */
#else
#define FEC_ENET_RX_PAGES	16
#define FEC_ENET_RX_FRSIZE	2048
#define FEC_ENET_RX_FRPPG	(PAGE_SIZE / FEC_ENET_RX_FRSIZE)
#define RX_RING_SIZE		(FEC_ENET_RX_FRPPG * FEC_ENET_RX_PAGES)
#define TX_RING_SIZE		16	/* Must be power of two */
#define TX_RING_MOD_MASK	15	/*   for this to work */
#endif

static int MAX_INTERRUPT_WORK = 0;
static int max_interrupt_work;

#ifdef CONFIG_PRxK
#define FEC_INTERRUPT   SIU_LEVEL5
#endif
 
/* Interrupt events/masks.
*/
#define FEC_ENET_HBERR	((uint)0x80000000)	/* Heartbeat error */
#define FEC_ENET_BABR	((uint)0x40000000)	/* Babbling receiver */
#define FEC_ENET_BABT	((uint)0x20000000)	/* Babbling transmitter */
#define FEC_ENET_GRA	((uint)0x10000000)	/* Graceful stop complete */
#define FEC_ENET_TXF	((uint)0x08000000)	/* Full frame transmitted */
#define FEC_ENET_TXB	((uint)0x04000000)	/* A buffer was transmitted */
#define FEC_ENET_RXF	((uint)0x02000000)	/* Full frame received */
#define FEC_ENET_RXB	((uint)0x01000000)	/* A buffer was received */
#define FEC_ENET_MII	((uint)0x00800000)	/* MII interrupt */
#define FEC_ENET_EBERR	((uint)0x00400000)	/* SDMA bus error */

/* The FEC stores dest/src/type, data, and checksum for receive packets.
 */
#define PKT_MAXBUF_SIZE		1518
#define PKT_MINBUF_SIZE		64
#define PKT_MAXBLR_SIZE		1520

/* The FEC buffer descriptors track the ring buffers.  The rx_bd_base and
 * tx_bd_base always point to the base of the buffer descriptors.  The
 * cur_rx and cur_tx point to the currently available buffer.
 * The dirty_tx tracks the current buffer that is being sent by the
 * controller.  The cur_tx and dirty_tx are equal under both completely
 * empty and completely full conditions.  The empty/ready indicator in
 * the buffer descriptor determines the actual condition.
 */
struct fec_enet_private {
	/* The saved address of a sent-in-place packet/buffer, for skfree(). */
	struct	sk_buff* tx_skbuff[TX_RING_SIZE];
	ushort	skb_cur;
	ushort	skb_dirty;

	/* CPM dual port RAM relative addresses.
	*/
	cbd_t	*rx_bd_base;		/* Address of Rx and Tx buffers. */
	cbd_t	*tx_bd_base;
	cbd_t	*cur_rx, *cur_tx;		/* The next free ring entry */
	cbd_t	*dirty_tx;	/* The ring entries to be free()ed. */
	scc_t	*sccp;
	struct	net_device_stats stats;
	char	tx_full;
	unsigned long lock;
};

static	void mii_startup_cmds(void);
static int fec_enet_open(struct device *dev);
static int fec_enet_start_xmit(struct sk_buff *skb, struct device *dev);
static int fec_enet_rx(struct device *dev);
static void fec_enet_mii(struct device *dev);
static	void fec_enet_interrupt(int irq, void * dev_id, struct pt_regs * regs);
static int fec_enet_close(struct device *dev);
static struct net_device_stats *fec_enet_get_stats(struct device *dev);
static void set_multicast_list(struct device *dev);
static void restart_fec(struct device *dev);

/* MII processing.  We keep this as simple as possible.  Requests are
 * placed on the list (if there is room).  When the request is finished
 * by the MII, an optional function may be called.
 */
typedef struct mii_list {
	uint	mii_regval;
	void	(*mii_func)(uint val, struct device *dev);
	struct	mii_list *mii_next;
} mii_list_t;

#define		NMII	64
mii_list_t	mii_cmds[NMII];
mii_list_t	*mii_free;
mii_list_t	*mii_head;
mii_list_t	*mii_tail;

static	int	phyaddr;
static	uint	phytype;

static int	mii_queue(int request, void (*func)(uint, struct device *));

/* Make MII read/write commands for the FEC.
*/

#define mk_mii_phyaddr(ADDR)	(0x60020000 | ((ADDR) << 23) | (2 << 18))

#define mk_mii_read(REG)	(0x60020000 | ((phyaddr << 23) | \
						(REG & 0x1f) << 18))

#define mk_mii_write(REG, VAL)	(0x50020000 | ((phyaddr << 23) | \
						(REG & 0x1f) << 18) | \
						(VAL & 0xffff))

//#define DEBUG_FEC_INTR
//#define DEBUG_FEC_INTR_TX
//#define DEBUG_FEC_INTR_RX
#define final_version

#if defined(DEBUG_FEC_INTR_RX) || defined(DEBUG_FEC_INTR_TX)

#if 0
unsigned long seq1 = 0;
unsigned long seq2 = 0;
unsigned long seqa = 0;
#endif

void dump_buf(unsigned char *buf, int pkt_len, int flag)
{
	int i;
#if 0
	unsigned long seq;

	if (buf[0] != 0x45 || buf[9] != 0x06)
		return;

	if (*(unsigned short *)&buf[20] != 20  && (unsigned short *)&buf[22] != 20)
		return;
 
   	if (flag == 1) {
		seq = *(unsigned long *)((char *)&buf[28]);
		if (seq == seqa)
   			printk("\nEth0T[%d] lost tx ack %x", pkt_len, seq);
		seqa = seq;
		return;
	}

	if (flag == 0) {
		seq = *(unsigned long *)((char *)&buf[24]);
		if (seq == seq1 || seq == seq2) {
   			printk("\nEth0R[%d] retrans seq %x seq1 %x seq2 %x", pkt_len, seq, seq1, seq2);
			return;
		}
		if (seq < seq1 || seq < seq2) {
   			printk("\nEth0R[%d] lost seq %x seq1 %x seq2 %x", pkt_len, seq, seq1, seq2);
		}
		seq1 = seq2;
		seq2 = seq;
		return;
	}

	seq = *(unsigned long *)((char *)&buf[24]);
   	printk("\nEth0T[%d] rx error seq %x", pkt_len, seq);
#endif
   	if (flag) {
   		printk("\nEth0T[%d]:", pkt_len);
   	} else {
   		printk("\nEth0R[%d]:", pkt_len);
   	}
   	i = pkt_len > 80 ? 80 : pkt_len;
   	for( ; i--; )
   	printk("%02x", *buf++);
}
#endif

static int
fec_enet_open(struct device *dev)
{

	restart_fec(dev);

	/* I should reset the ring buffers here, but I don't yet know
	 * a simple way to do that.
	 */

	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;

	return 0;					/* Always succeed */
}

static int
fec_enet_start_xmit(struct sk_buff *skb, struct device *dev)
{
	struct fec_enet_private *fep = (struct fec_enet_private *)dev->priv;
	volatile cbd_t	*bdp;
	unsigned long flags;

#ifdef DEBUG_FEC_INTR_TX
dump_buf(skb->data+14, skb->len, 1);
#endif
	/* Transmitter timeout, serious problems. */
	if (dev->tbusy) {
		int tickssofar = jiffies - dev->trans_start;
		if (tickssofar < 20)
			return 1;
		printk("%s: transmit timed out.\n", dev->name);
		fep->stats.tx_errors++;
#ifndef final_version
		{
			int	i;
			cbd_t	*bdp;
			printk(" Ring data dump: cur_tx %x%s cur_rx %x.\n",
				   (uint)(fep->cur_tx),
				   fep->tx_full ? " (full)" : "",
				   (uint)(fep->cur_rx));
			bdp = fep->tx_bd_base;
			for (i = 0 ; i < TX_RING_SIZE; i++)
				printk("%04x %04x %08x\n",
					bdp->cbd_sc,
					bdp->cbd_datlen,
					bdp->cbd_bufaddr);
			bdp = fep->rx_bd_base;
			for (i = 0 ; i < RX_RING_SIZE; i++)
				printk("%04x %04x %08x\n",
					bdp->cbd_sc,
					bdp->cbd_datlen,
					bdp->cbd_bufaddr);
		}
#endif

		dev->tbusy=0;
		dev->trans_start = jiffies;

		return 0;
	}

	/* Block a timer-based transmit from overlapping.  This could better be
	   done with atomic_swap(1, dev->tbusy), but set_bit() works as well. */
	if (test_and_set_bit(0, (void*)&dev->tbusy) != 0) {
		printk("%s: Transmitter access conflict.\n", dev->name);
		return 1;
	}

	if (test_and_set_bit(0, (void*)&fep->lock) != 0) {
		printk("%s: tx queue lock!.\n", dev->name);
		/* don't clear dev->tbusy flag. */
		return 1;
	}

	/* Fill in a Tx ring entry */
	bdp = fep->cur_tx;

#ifndef final_version
	if (bdp->cbd_sc & BD_ENET_TX_READY) {
		/* Ooops.  All transmit buffers are full.  Bail out.
		 * This should not happen, since dev->tbusy should be set.
		 */
		printk("%s: tx queue full!.\n", dev->name);
		fep->lock = 0;
		return 1;
	}
#endif

	/* Clear all of the status flags.
	 */
	bdp->cbd_sc &= ~BD_ENET_TX_STATS;

	/* Set buffer length and buffer pointer.
	*/
	bdp->cbd_bufaddr = __pa(skb->data);
	bdp->cbd_datlen = skb->len;

	/* Save skb pointer.
	*/
	fep->tx_skbuff[fep->skb_cur] = skb;

	fep->stats.tx_bytes += skb->len;
	fep->skb_cur = (fep->skb_cur+1) & TX_RING_MOD_MASK;
	
	/* Push the data cache so the CPM does not get stale memory
	 * data.
	 */
	flush_dcache_range((uint)(skb->data), (uint)(skb->data + skb->len));

	/* Send it on its way.  Tell CPM its ready, interrupt when done,
	 * its the last BD of the frame, and to put the CRC on the end.
	 */
	save_flags(flags);
	cli();

	bdp->cbd_sc |= (BD_ENET_TX_READY | BD_ENET_TX_INTR | BD_ENET_TX_LAST | BD_ENET_TX_TC);

	dev->trans_start = jiffies;
	(&(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec))->fec_x_des_active = 0x01000000;

	/* If this was the last BD in the ring, start at the beginning again.
	*/
	if (bdp->cbd_sc & BD_ENET_TX_WRAP)
		bdp = fep->tx_bd_base;
	else
		bdp++;

	fep->lock = 0;
	if (bdp->cbd_sc & BD_ENET_TX_READY)
		fep->tx_full = 1;
	else
		dev->tbusy=0;
	fep->cur_tx = (cbd_t *)bdp;
	restore_flags(flags);
	return 0;
}

/* The interrupt handler.
 * This is called from the MPC core interrupt.
 */
static	void
fec_enet_interrupt(int irq, void * dev_id, struct pt_regs * regs)
{
	struct	device *dev = dev_id;
	struct	fec_enet_private *fep;
	volatile cbd_t	*bdp;
	volatile fec_t	*ep;
	uint	int_events;

	fep = (struct fec_enet_private *)dev->priv;
	ep = &(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec);
	if (dev->interrupt)
		printk("%s: Re-entering the interrupt handler.\n", dev->name);
	dev->interrupt = 1;

#ifdef DEBUG_FEC_INTR
	printk("FEC %s intr.", dev->name);
#endif

	max_interrupt_work = MAX_INTERRUPT_WORK;

	/* Get the interrupt events that caused us to be here.
	*/
	while ((int_events = ep->fec_ievent) != 0) {
 	if (MAX_INTERRUPT_WORK) {
		if (max_interrupt_work > 0)
			max_interrupt_work--;
		else
			break;
	}
	ep->fec_ievent = int_events;
	if ((int_events & (FEC_ENET_HBERR | FEC_ENET_EBERR)) != 0)
				printk("FEC ERROR %x\n", int_events);
	if ((int_events & FEC_ENET_BABR) != 0)
		fep->stats.rx_errors++;
	if ((int_events & FEC_ENET_BABT) != 0)
		fep->stats.tx_errors++;

	/* Handle receive event in its own function.
	*/
	if (int_events & (FEC_ENET_RXF | FEC_ENET_RXB)) {
#ifdef DEBUG_FEC_INTR
		printk("R.");
#endif
		fec_enet_rx(dev_id);
	}

	/* Transmit OK, or non-fatal error.  Update the buffer descriptors.
	 * FEC handles all errors, we just discover them as part of the
	 * transmit process.
	 */
	if (int_events & (FEC_ENET_TXF | FEC_ENET_TXB)) {
#ifdef DEBUG_FEC_INTR
		printk("T.");
#endif
		if (max_interrupt_work) ++max_interrupt_work;
	    bdp = fep->dirty_tx;
	    while ((bdp->cbd_sc&BD_ENET_TX_READY)==0) {

		if (MAX_INTERRUPT_WORK) {
			if (max_interrupt_work > 0)
				max_interrupt_work--;
			else
				break;
		}
#if 1
		if ((bdp==fep->cur_tx) && (fep->tx_full == 0))
		    break;
#endif

#ifdef CONFIG_PRxK
		{
		volatile sysconf8xx_t *psiu = &(((immap_t *)IMAP_ADDR)->im_siu_conf);
		psiu->sc_swsr = 0x556c;
		psiu->sc_swsr = 0xaa39;
		}
#endif

		if (bdp->cbd_sc & (BD_ENET_TX_HB|BD_ENET_TX_LC|BD_ENET_TX_RL|
							BD_ENET_TX_UN|BD_ENET_TX_CSL|BD_ENET_TX_DEF)) {
#ifdef DEBUG_FEC_INTR
		printk("E.");
#endif
		/* Deferred means some collisions occurred during transmit,
		 * but we eventually sent the packet OK.
		 */
		if (bdp->cbd_sc & BD_ENET_TX_DEF)
		    fep->stats.collisions++;
		if (bdp->cbd_sc & BD_ENET_TX_HB)	/* No heartbeat */
		    fep->stats.tx_heartbeat_errors++;
		if (bdp->cbd_sc & BD_ENET_TX_LC)	/* Late collision */
		    fep->stats.tx_window_errors++;
		if (bdp->cbd_sc & BD_ENET_TX_RL)	/* Retrans limit */
		    fep->stats.tx_aborted_errors++;
		if (bdp->cbd_sc & BD_ENET_TX_UN)	/* Underrun */
		    fep->stats.tx_fifo_errors++;
		if (bdp->cbd_sc & BD_ENET_TX_CSL)	/* Carrier lost */
		    fep->stats.tx_carrier_errors++;

		//fep->stats.tx_errors++;
		}
	    
		fep->stats.tx_packets++;
		
#ifndef final_version
		if (bdp->cbd_sc & BD_ENET_TX_READY)
		    printk("HEY! Enet xmit interrupt and TX_READY.\n");
#endif
	    
		/* Free the sk buffer associated with this last transmit.
		 */

#if 0
printk("TXI: %x %x %x\n", bdp, fep->tx_skbuff[fep->skb_dirty], fep->skb_dirty);
#endif

		dev_kfree_skb(fep->tx_skbuff[fep->skb_dirty]/*, FREE_WRITE*/);
		fep->tx_skbuff[fep->skb_dirty] = NULL;
		fep->skb_dirty = (fep->skb_dirty + 1) & TX_RING_MOD_MASK;
	    
		/* Update pointer to next buffer descriptor to be transmitted.
		 */
		if (bdp->cbd_sc & BD_ENET_TX_WRAP)
		    bdp = fep->tx_bd_base;
		else
		    bdp++;
	    
		/* Since we have freed up a buffer, the ring is no longer
		 * full.
		 */
		if (fep->tx_full && dev->tbusy) {
		    fep->tx_full = 0;
		    dev->tbusy = 0;
		    mark_bh(NET_BH);
		}

		fep->dirty_tx = (cbd_t *)bdp;
#if 0
		if (bdp==fep->cur_tx)
		    break;
#endif
	    }/*while (bdp->cbd_sc&BD_ENET_TX_READY)==0*/
	 } /* if tx events */

	if (int_events & FEC_ENET_MII) {
#ifdef DEBUG_FEC_INTR
		printk("I.");
#endif
		fec_enet_mii(dev_id);
	}
	
	} /* while any events */

	dev->interrupt = 0;

#ifdef DEBUG_FEC_INTR
	printk("done");
#endif

	return;
}

/* During a receive, the cur_rx points to the current incoming buffer.
 * When we update through the ring, if the next incoming buffer has
 * not been given to the system, we just set the empty indicator,
 * effectively tossing the packet.
 */
static int
fec_enet_rx(struct device *dev)
{
	struct	fec_enet_private *fep;
	volatile cbd_t *bdp;
	struct	sk_buff	*skb;
	ushort	pkt_len;
	volatile fec_t	*ep;
	int	rx_error = 0;

	fep = (struct fec_enet_private *)dev->priv;
	ep = &(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec);

	/* First, grab all of the stats for the incoming packet.
	 * These get messed up if we get called due to a busy condition.
	 */
	bdp = fep->cur_rx;
	if (max_interrupt_work) ++max_interrupt_work;

for (;;) {
		if (MAX_INTERRUPT_WORK) {
			if (max_interrupt_work > 0)
				max_interrupt_work--;
			else
				break;
		}

#ifdef CONFIG_PRxK
		{
		volatile sysconf8xx_t *psiu = &(((immap_t *)IMAP_ADDR)->im_siu_conf);
		psiu->sc_swsr = 0x556c;
		psiu->sc_swsr = 0xaa39;
		}
#endif

	if (bdp->cbd_sc & BD_ENET_RX_EMPTY)
		break;
		
#ifndef final_version
	/* Since we have allocated space to hold a complete frame,
	 * the last indicator should be set.
	 */
	if ((bdp->cbd_sc & BD_ENET_RX_LAST) == 0)
		printk("FEC ENET: rcv is not +last\n");
#endif

	/* Frame too long or too short.
	*/
	if (bdp->cbd_sc & BD_ENET_RX_LAST) { /* Last in a frame */
		if (bdp->cbd_sc & (BD_ENET_RX_LG | BD_ENET_RX_SH))
			fep->stats.rx_length_errors++;
		if (bdp->cbd_sc & BD_ENET_RX_NO) {	/* Frame alignment */
			rx_error = 1;
			fep->stats.rx_frame_errors++;
		}
		if (bdp->cbd_sc & BD_ENET_RX_CR)	/* CRC Error */
			fep->stats.rx_crc_errors++;
		if (bdp->cbd_sc & BD_ENET_RX_OV)	/* FIFO overrun */
			fep->stats.rx_crc_errors++;
	}

	/* Report late collisions as a frame error.
	 * On this error, the BD is closed, but we don't know what we
	 * have in the buffer.  So, just drop this frame on the floor.
	 */
	if (bdp->cbd_sc & BD_ENET_RX_CL) {
		fep->stats.rx_frame_errors++;
		rx_error = 1;
	}
	if (!rx_error) {

		/* Process the incoming frame.
		*/
		fep->stats.rx_packets++;
		pkt_len = bdp->cbd_datlen;
		fep->stats.rx_bytes += pkt_len;

		/* This does 16 byte alignment, exactly what we need.
		*/
		skb = dev_alloc_skb(pkt_len);

		if (skb == NULL) {
			printk("%s: Memory squeeze, dropping packet.\n", dev->name);
			fep->stats.rx_dropped++;
		}
		else {
			skb->dev = dev;
			skb_put(skb,pkt_len);	/* Make room */
			eth_copy_and_sum(skb,
				(unsigned char *)__va(bdp->cbd_bufaddr),
				pkt_len, 0);
			skb->protocol=eth_type_trans(skb,dev);
#ifdef DEBUG_FEC_INTR_RX
dump_buf(skb->data, skb->len, 0);
#endif
			netif_rx(skb);
		}
	}

	/* Clear the status flags for this buffer.
	*/
	bdp->cbd_sc &= ~BD_ENET_RX_STATS;

	/* Mark the buffer empty.
	*/
	bdp->cbd_sc |= BD_ENET_RX_EMPTY;

	/* Update BD pointer to next entry.
	*/
	if (bdp->cbd_sc & BD_ENET_RX_WRAP)
		bdp = fep->rx_bd_base;
	else
		bdp++;
	
#if 1
	/* Doing this here will keep the FEC running while we process
	 * incoming frames.  On a heavily loaded network, we should be
	 * able to keep up at the expense of system resources.
	 */
	ep->fec_r_des_active = 0x01000000;
#endif
   }
	fep->cur_rx = (cbd_t *)bdp;

#if 0
	/* Doing this here will allow us to process all frames in the
	 * ring before the FEC is allowed to put more there.  On a heavily
	 * loaded network, some frames may be lost.  Unfortunately, this
	 * increases the interrupt overhead since we can potentially work
	 * our way back to the interrupt return only to come right back
	 * here.
	 */
	ep->fec_r_des_active = 0x01000000;
#endif

	return 0;
}

static void
fec_enet_mii(struct device *dev)
{
	struct	fec_enet_private *fep;
	volatile fec_t	*ep;
	mii_list_t	*mip;
	uint		mii_reg;

	fep = (struct fec_enet_private *)dev->priv;
	ep = &(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec);
	mii_reg = ep->fec_mii_data;
	
	if ((mip = mii_head) == NULL) {
		printk("MII and no head!\n");
		return;
	}

	if (mip->mii_func != NULL) {
		(*(mip->mii_func))(mii_reg, dev);
	} else {
//printk("fec event 0x%02x mii_reg 0x%04x\n", (mii_reg >> 18) & 0x1f, mii_reg&0xffff) ;
	}

	mii_head = mip->mii_next;
	mip->mii_next = mii_free;
	mii_free = mip;

	if ((mip = mii_head) != NULL)
		ep->fec_mii_data = mip->mii_regval;
}

static int
mii_queue(int regval, void (*func)(uint, struct device *))
{
	unsigned long	flags;
	mii_list_t	*mip;
	int		retval;

	retval = 0;

	save_flags(flags);
	cli();

	if ((mip = mii_free) != NULL) {
		mii_free = mip->mii_next;
		mip->mii_regval = regval;
		mip->mii_func = func;
		mip->mii_next = NULL;
		if (mii_head) {
			mii_tail->mii_next = mip;
			mii_tail = mip;
		}
		else {
			mii_head = mii_tail = mip;
			(&(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec))->fec_mii_data = regval;
		}
	}
	else {
		retval = 1;
	}

	restore_flags(flags);

	return(retval);
}

static	volatile uint	full_duplex = 1;

static int fec_neg_complete = 0;

static int prev_mii_reg = 0;

static void
mii_status(uint mii_reg, struct device *dev)
{
	volatile uint	prev_duplex;

//printk("fec mii_status 0x%02x mii_reg 0x%04x\n", (mii_reg >> 18) & 0x1f, mii_reg&0xffff) ;
#ifdef CONFIG_PRxK
#ifdef CONFIG_SM100
	/* LXT972A PHY 
	*/

	if (((mii_reg >> 18) & 0x1f) == 17) {
		/* Extended chip status register.
		*/

		if (prev_mii_reg == (mii_reg & 0x4680) && fec_neg_complete == 2) {
			return;
		}

		prev_mii_reg = mii_reg & 0x4680;
		prev_duplex = full_duplex;

		if ((mii_reg & 0x0080) || ((bd_t *)res)->mii_operation) {
			/* auto-negotiation completed or disabled 
			*/
			fec_neg_complete = 2;
			printk("fec: ");
			if (mii_reg & 0x0400) {
				// Signals link up to the device
				//dev->flags |= IFF_RUNNING;
				printk("link UP");
			} else {
				// Signals link down to the device
				//dev->flags &= ~IFF_RUNNING;
				printk("link DOWN");
			}
			if (((bd_t *)res)->mii_operation)
				printk(", Configured as: ");
			else
				printk(", ");
			if (mii_reg & 0x4000)
				printk("100 Mbps");
			else
				printk("10 Mbps");

			if (mii_reg & 0x0200) {
				full_duplex = 1;
				printk(", Full-Duplex");
			} else {
				full_duplex = 0;
				printk(", Half-Duplex");
			}
			if (MAX_INTERRUPT_WORK) {
				printk(", Intr-Max %d", MAX_INTERRUPT_WORK);
			}
		} else {
			if (fec_neg_complete == 1) {
				return;
			}
			fec_neg_complete = 1;

			// Signals link down to the device
			//dev->flags &= ~IFF_RUNNING;
			printk("fec: link down");
		}
		printk("\n");

		if (prev_duplex != full_duplex) {
			volatile fec_t *ep = &(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec);
			/* Enable MII mode.
			*/
			if (full_duplex) {
				ep->fec_r_cntrl &= ~0x02;	/* MII enable */
				ep->fec_x_cntrl |= 0x04;	/* FD enable */
			}
			else {
				ep->fec_r_cntrl |= 0x02;	// MII enable|No Rcv on Xmit
				ep->fec_x_cntrl &= ~0x04;
			}
			//restart_fec(dev);
		}
	}
#else
	/* DP83846A DsPHYTER 
	*/

	if (((mii_reg >> 18) & 0x1f) == 0x10) {
		/* Extended chip status register.
		*/

		if (prev_mii_reg == (mii_reg & 0x5f) && fec_neg_complete == 2) {
			return;
		}

		prev_mii_reg = mii_reg & 0x57;
		prev_duplex = full_duplex;

		if ((mii_reg & 0x0010) || ((bd_t *)res)->mii_operation) {
			fec_neg_complete = 2;
			printk("fec: ");
			if (mii_reg & 0x0001) {
				// Signals link up to the device
				//dev->flags |= IFF_RUNNING;
				printk("link UP");
			} else {
				// Signals link down to the device
				//dev->flags &= ~IFF_RUNNING;
				printk("link DOWN");
			}
			if (((bd_t *)res)->mii_operation)
				printk(", Configured as: ");
			else
				printk(", ");
			if (mii_reg & 0x0002)
				printk("10 Mbps");
			else
				printk("100 Mbps");
	
			if (mii_reg & 0x0004) {
				full_duplex = 1;
				printk(", Full-Duplex");
			} else {
				full_duplex = 0;
				printk(", Half-Duplex");
			}
			if (MAX_INTERRUPT_WORK) {
				printk(", Intr-Max %d", MAX_INTERRUPT_WORK);
			}
		} else {
			if (fec_neg_complete == 1) {
				return;
			}
			fec_neg_complete = 1;
			// Signals link down to the device
			//dev->flags &= ~IFF_RUNNING;
			printk("fec: link down");
		}

		if (mii_reg & 0x0060)
			printk("; ");
		if (mii_reg & 0x0020)
			printk("loopback");
		if (mii_reg & 0x0040) {
			if (!(mii_reg & 0x0020))
				printk(", ");
			printk(", remote fault");
		}

		printk("\n");

		if (prev_duplex != full_duplex) {
			volatile fec_t *ep = &(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec);
			/* Enable MII mode.
			*/
			if (full_duplex) {
				ep->fec_r_cntrl &= ~0x02;	/* MII enable */
				ep->fec_x_cntrl |= 0x04;	/* FD enable */
			}
			else {
				ep->fec_r_cntrl |= 0x02;	// MII enable|No Rcv on Xmit
				ep->fec_x_cntrl &= ~0x04;
			}
			//restart_fec(dev);
		}
	}
#endif
#else
	if (((mii_reg >> 18) & 0x1f) == 1) {
		/* status register.
		*/
		printk("fec: ");
		if (mii_reg & 0x0004)
			printk("link up");
		else
			printk("link down");

		if (mii_reg & 0x0010)
			printk(", remote fault");
		if (mii_reg & 0x0020)
			printk(", auto complete");
		printk("\n");
	}
	if (((mii_reg >> 18) & 0x1f) == 0x14) {
		/* Extended chip status register.
		*/
		prev_duplex = full_duplex;
		printk("fec: ");
		if (mii_reg & 0x0800)
			printk("100 Mbps");
		else
			printk("10 Mbps");

		if (mii_reg & 0x1000) {
			printk(", Full-Duplex\n");
			full_duplex = 1;
		}
		else {
			printk(", Half-Duplex\n");
			full_duplex = 0;
		}
		if (prev_duplex != full_duplex)
			restart_fec(dev);
	}
	if (((mii_reg >> 18) & 0x1f) == 31) {
		/* QS6612 PHY Control/Status.
		 * OK, now we have it all, so figure out what is going on.
		 */
		prev_duplex = full_duplex;
		printk("fec: ");

		mii_reg = (mii_reg >> 2) & 7;

		if (mii_reg & 1)
			printk("10 Mbps");
		else
			printk("100 Mbps");

		if (mii_reg > 4) {
			printk(", Full-Duplex\n");
			full_duplex = 1;
		}
		else {
			printk(", Half-Duplex\n");
			full_duplex = 0;
		}

		if (prev_duplex != full_duplex)
			restart_fec(dev);
	}
#endif
}

#define COUNT_MOD 1

static int cnt_check = COUNT_MOD;

void check_fec_neg_complete(void)
{
#ifdef CONFIG_SM100
	int status_reg = 17;
#else
	int status_reg = 0x10;
#endif

	if ((cnt_check = (cnt_check + 1) & COUNT_MOD))
		return;

	if (fec_neg_complete) {
		mii_queue(mk_mii_read(status_reg), mii_status);
	}
}

static	uint	phyno;

static void
mii_discover_phy3(uint mii_reg, struct device *dev)
{
//printk("fec mii_phy3 0x%02x mii_reg 0x%04x\n", (mii_reg >> 18) & 0x1f, mii_reg&0xffff) ;
	phytype <<= 16;
	phytype |= (mii_reg & 0xffff);
	printk("fec: Phy @ 0x%x, type 0x%08x\n", phyno, phytype);
	mii_startup_cmds();
}

static void
mii_discover_phy(uint mii_reg, struct device *dev)
{
//printk("fec mii_phy 0x%02x mii_reg 0x%04x\n", (mii_reg >> 18) & 0x1f, mii_reg&0xffff) ;
	if (phyno < 32) {
		if ((phytype = (mii_reg & 0xffff)) != 0xffff) {
			phyaddr = phyno;
			mii_queue(mk_mii_read(3), mii_discover_phy3);
		}
		else {
			phyno++;
			mii_queue(mk_mii_phyaddr(phyno), mii_discover_phy);
		}
	}
	else {
		printk("FEC: No PHY device found.\n");
	}
}

#ifdef CONFIG_PRxK
static void
mii_reset(uint mii_reg, struct device *dev)
{
	bd_t	*bd = (bd_t *)res;
	int		cmd;

	if (mii_reg & 0x8000) {
		mii_queue(mk_mii_read(0), mii_reset);
		return;
	}

#ifdef CONFIG_SM100
	/* LXT972A PHY 
	*/

	prev_mii_reg = -1;

	/* Force configuration as defined.
	*/
	if (!bd->mii_operation) {
		cmd = 0x1200; //Autonegotiate
	} else {
		cmd  = (bd->mii_operation & 1) ? 0x2000 : 0x0000;	// 100bt or 10bt
		cmd |= (bd->mii_operation & 2) ? 0x0100 : 0x0000;	// Full or half duplex
	}
	mii_queue(mk_mii_write(0x00, cmd), NULL);

	/* Read status registers to clear any pending interrupt.
	*/
	mii_queue(mk_mii_read(17), mii_status);
	//mii_queue(mk_mii_read(19), mii_status);

	/* Enable Link status change interrupts.
	*/
	//mii_queue(mk_mii_write(18, 0x00f2), NULL);

	/* Configure leds 
	*/
	mii_queue(mk_mii_write(0x14, 0x0d01), NULL);

#else
	/* DP83846A DsPHYTER 
	*/

	prev_mii_reg = -1;

	/* disable heartbeat for 10BT 
	*/
	mii_queue(mk_mii_write(0x1a, 0x0106), NULL);

	/* Enable some link change interrupts.
	*/
	if (!bd->mii_operation) {
		cmd = 0x1200; //Autonegotiate
	} else {
		cmd  = (bd->mii_operation & 1) ? 0x2000 : 0x0000;	// 100bt or 10bt
		cmd |= (bd->mii_operation & 2) ? 0x0100 : 0x0000;	// Full or half duplex
	}
	mii_queue(mk_mii_write(0x00, cmd), NULL);

	/* Enable Link status change interrupts.
	*/
	//mii_queue(mk_mii_write(0x12, 0x3600), NULL);
	//mii_queue(mk_mii_write(0x11, 0x0002), NULL);

	mii_queue(mk_mii_read(0x10), mii_status);

#endif
}
#endif

static	void
mii_startup_cmds(void)
{
#ifdef CONFIG_RPXCLASSIC
	/* QS6612 PHY.
	*/

	/* Set BCSR2 so FEC LEDs show whats happening
	*/
	*((uint *) RPX_CSR_ADDR) &= ~BCSR2_FETHLEDMODE;

	/* Write some registers with bits that the documentation
	 * claims are both reserved and should be zero.
	 */

	/* The PHY powers up isolated on the RPX, so send a command to
	 * allow operation.
	 */
	mii_queue(mk_mii_write(31, 0x0dc0), NULL);

	/* Enable some link change interrupts.
	*/
	mii_queue(mk_mii_write(30, 0x50), NULL);
	mii_queue(mk_mii_write(0, 0x1000), NULL); //Autonegotiate

	/* Read status registers to clear any pending interrupt.
	*/
	mii_queue(mk_mii_read(0), mii_status);
	mii_queue(mk_mii_read(1), mii_status);
	mii_queue(mk_mii_read(29), mii_status);

	/* Get actual link status.
	*/
	mii_queue(mk_mii_read(31), mii_status);

#else
#ifdef CONFIG_PRxK
	/* DP83846A DsPHYTER or LXT972A PHY 
	*/
	mii_queue(mk_mii_write(0x00, 0x8000), mii_reset);

#else
	/* Level One PHY.
	*/

	/* Read status registers to clear any pending interrupt.
	*/
	mii_queue(mk_mii_read(1), mii_status);
	mii_queue(mk_mii_read(18), mii_status);

	/* Read extended chip status register.
	*/
	mii_queue(mk_mii_read(0x14), mii_status);

	/* Set default operation of 100-TX....for some reason
	 * some of these bits are set on power up, which is wrong.
	 */
	mii_queue(mk_mii_write(0x13, 0), NULL);

	/* Enable Link status change interrupts.
	*/
	mii_queue(mk_mii_write(0x11, 0x0002), NULL);

	/* Don't advertize Full duplex.
	mii_queue(mk_mii_write(0x04, 0x0021), NULL);
	*/
#endif
#endif

}

#if defined(CONFIG_FADS) || defined(CONFIG_RPXCLASSIC)

/* This supports the mii_link interrupt below.
 * We should get called three times.  Once for register 1, once for
 * register 18, and once for register 20.
 */
static	uint mii_saved_reg1;

static void
mii_relink(uint mii_reg, struct device *dev)
{
	volatile uint	prev_duplex;
	unsigned long	flags;

	if (((mii_reg >> 18) & 0x1f) == 1) {
		/* Just save the status register and get out.
		*/
		mii_saved_reg1 = mii_reg;
		return;
	}
	if (((mii_reg >> 18) & 0x1f) == 18) {
		/* Not much here, but has to be read to clear the
		 * interrupt condition.
		 */
		if ((mii_reg & 0x8000) == 0)
			printk("fec: re-link and no IRQ?\n");
		if ((mii_reg & 0x4000) == 0)
			printk("fec: no PHY power?\n");
	}
	if (((mii_reg >> 18) & 0x1f) == 20) {
		/* Extended chip status register.
		 * OK, now we have it all, so figure out what is going on.
		 */
		prev_duplex = full_duplex;
		printk("fec: ");
		if (mii_saved_reg1 & 0x0004)
			printk("link up");
		else
			printk("link down");

		if (mii_saved_reg1 & 0x0010)
			printk(", remote fault");
		if (mii_saved_reg1 & 0x0020)
			printk(", auto complete");

		if (mii_reg & 0x0800)
			printk(", 100 Mbps");
		else
			printk(", 10 Mbps");

		if (mii_reg & 0x1000) {
			printk(", Full-Duplex\n");
			full_duplex = 1;
		}
		else {
			printk(", Half-Duplex\n");
			full_duplex = 0;
		}
		if (prev_duplex != full_duplex) {
			save_flags(flags);
			cli();
			restart_fec(dev);
			restore_flags(flags);
		}
	}
	if (((mii_reg >> 18) & 0x1f) == 31) {
		/* QS6612 PHY Control/Status.
		 * OK, now we have it all, so figure out what is going on.
		 */
		prev_duplex = full_duplex;
		printk("fec: ");
		if (mii_saved_reg1 & 0x0004)
			printk("link up");
		else
			printk("link down");

		if (mii_saved_reg1 & 0x0010)
			printk(", remote fault");
		if (mii_saved_reg1 & 0x0020)
			printk(", auto complete");

		mii_reg = (mii_reg >> 2) & 7;

		if (mii_reg & 1)
			printk(", 10 Mbps");
		else
			printk(", 100 Mbps");

		if (mii_reg > 4) {
			printk(", Full-Duplex\n");
			full_duplex = 1;
		}
		else {
			printk(", Half-Duplex\n");
			full_duplex = 0;
		}

		if (prev_duplex != full_duplex) {
			save_flags(flags);
			cli();
			restart_fec(dev);
			restore_flags(flags);
		}
	}
}

/* This interrupt occurs when the LTX970 detects a link change.
*/
static	void
#ifdef CONFIG_RPXCLASSIC
mii_link_interrupt(void *dev_id)
#else
mii_link_interrupt(int irq, void * dev_id, struct pt_regs * regs)
#endif
{
	struct	device *dev = dev_id;
	struct	fec_enet_private *fep;
	volatile fec_t	*ep;

	fep = (struct fec_enet_private *)dev->priv;
	ep = &(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec);

	/* We need to sequentially read registers 1 and 18 to clear
	 * the interrupt.  We don't need to do that here because this
	 * is an edge triggered interrupt that has already been acknowledged
	 * by the top level handler.  We also read the extended status
	 * register 20.  We just queue the commands and let them happen
	 * as part of the "normal" processing.
	 */
#ifdef CONFIG_RPXCLASSIC
	mii_queue(mk_mii_read(1), mii_relink);
	mii_queue(mk_mii_read(29), mii_relink);
	mii_queue(mk_mii_read(31), mii_relink);
#else
	mii_queue(mk_mii_read(1), mii_relink);
	mii_queue(mk_mii_read(18), mii_relink);
	mii_queue(mk_mii_read(20), mii_relink);
#endif

}
#endif

static int
fec_enet_close(struct device *dev)
{
	volatile	fec_t	*fecp;

	fecp = (fec_t *)dev->base_addr;

	/* Whack a reset.  We should wait for this.
	*/
	fecp->fec_ecntrl = 1;
	udelay(10);

	/* Don't know what to do yet.
	*/
	dev->start = 0;
	dev->tbusy = 1;

	return 0;
}

static struct net_device_stats *fec_enet_get_stats(struct device *dev)
{
	struct fec_enet_private *fep = (struct fec_enet_private *)dev->priv;

	return &fep->stats;
}

/* Set or clear the multicast filter for this adaptor.
 * Skeleton taken from sunlance driver.
 * The CPM Ethernet implementation allows Multicast as well as individual
 * MAC address filtering.  Some of the drivers check to make sure it is
 * a group multicast address, and discard those that are not.  I guess I
 * will do the same for now, but just remove the test if you want
 * individual filtering as well (do the upper net layers want or support
 * this kind of feature?).
 */

static void set_multicast_list(struct device *dev)
{
	struct	fec_enet_private *fep;
	volatile fec_t *ep;

	fep = (struct fec_enet_private *)dev->priv;
	ep = &(((immap_t *)IMAP_ADDR)->im_cpm.cp_fec);

	if (dev->flags&IFF_PROMISC) {
	  
		/* Log any net taps. */
		printk("%s: Promiscuous mode enabled.\n", dev->name);
		ep->fec_r_cntrl |= 0x0008;
	} else {

		ep->fec_r_cntrl &= ~0x0008;

		if (dev->flags & IFF_ALLMULTI) {
			/* Catch all multicast addresses, so set the
			 * filter to all 1's.
			 */
			ep->fec_hash_table_high = 0xffffffff;
			ep->fec_hash_table_low = 0xffffffff;
		}
#if 0
		else {
			/* Clear filter and add the addresses in the list.
			*/
			ep->sen_gaddr1 = 0;
			ep->sen_gaddr2 = 0;
			ep->sen_gaddr3 = 0;
			ep->sen_gaddr4 = 0;

			dmi = dev->mc_list;

			for (i=0; i<dev->mc_count; i++) {
				
				/* Only support group multicast for now.
				*/
				if (!(dmi->dmi_addr[0] & 1))
					continue;

				/* The address in dmi_addr is LSB first,
				 * and taddr is MSB first.  We have to
				 * copy bytes MSB first from dmi_addr.
				 */
				mcptr = (u_char *)dmi->dmi_addr + 5;
				tdptr = (u_char *)&ep->sen_taddrh;
				for (j=0; j<6; j++)
					*tdptr++ = *mcptr--;

				/* Ask CPM to run CRC and set bit in
				 * filter mask.
				 */
				cpmp->cp_cpcr = mk_cr_cmd(CPM_CR_CH_SCC1, CPM_CR_SET_GADDR) | CPM_CR_FLG;
				/* this delay is necessary here -- Cort */
				udelay(10);
				while (cpmp->cp_cpcr & CPM_CR_FLG);
			}
		}
#endif
	}
}

static int fec_set_mac_address(struct device *dev, void *p)
{
        struct sockaddr *addr=p;
	volatile	fec_t	*fecp;

        if(dev->start)
                return -EBUSY;
        memcpy(dev->dev_addr, addr->sa_data,dev->addr_len);

	fecp = (fec_t *)dev->base_addr;

	/* Set station address.
	*/
	fecp->fec_addr_low =	(dev->dev_addr[0] << 24) |
				(dev->dev_addr[1] << 16) |
				(dev->dev_addr[2] <<  8) |
				dev->dev_addr[3];
	fecp->fec_addr_high = (dev->dev_addr[4] << 8) | dev->dev_addr[5];

	phyno = 0;
	mii_queue(mk_mii_phyaddr(0), mii_discover_phy);

	return(0);
}

#ifdef CONFIG_PRxK
static int fec_do_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
{
	switch (cmd) {
	case SIOCDEVPRIVATE:
		((bd_t *)res)->mii_operation = ifr->ifr_ifru.ifru_ivalue;

		if (ifr->ifr_ifru.ifru_flags)
			return(0);

		phyno = 0;
		mii_queue(mk_mii_phyaddr(0), mii_discover_phy);

		return(0);
	case SIOCDEVPRIVATE+1:
		((bd_t *)res)->max_int_work = ifr->ifr_ifru.ifru_ivalue;
		MAX_INTERRUPT_WORK = ifr->ifr_ifru.ifru_ivalue;
		return(0);
	}

	return -EINVAL;
}
#endif

/* Initialize the FEC Ethernet on 860T.
 */
__initfunc(int fec_enet_init(void))
{
	struct device *dev;
	struct fec_enet_private *fep;
	int i, j;
	unsigned char	*iap;
	unsigned long	mem_addr;
	pte_t		*pte;
	volatile	cbd_t	*bdp;
	cbd_t		*cbd_base;
	volatile	immap_t	*immap;
	volatile	fec_t	*fecp;
	bd_t		*bd;
#ifdef CONFIG_RPXCLASSIC
	unsigned char	tmpaddr[6];
#endif

	immap = (immap_t *)IMAP_ADDR;	/* pointer to internal registers */

	bd = (bd_t *)res;

	/* Allocate some private information.
	*/
	fep = (struct fec_enet_private *)kmalloc(sizeof(*fep), GFP_KERNEL);
	if (!fep) panic("FEC: null pointer");

	__clear_user(fep,sizeof(*fep));

	/* Create an Ethernet device instance.
	*/
	dev = init_etherdev(0, 0);

	fecp = &(immap->im_cpm.cp_fec);

	/* Whack a reset.  We should wait for this.
	*/
	fecp->fec_ecntrl = 1;
	udelay(10);

#ifdef CONFIG_PRxK
	/* reset 100BaseT through PC4 */
	immap->im_ioport.iop_pcdir |= 0x0800;
	immap->im_ioport.iop_pcpar &= ~(0x0800);
	immap->im_ioport.iop_pcdat &= ~(0x0800);
#endif

	/* Set the Ethernet address.  If using multiple Enets on the 8xx,
	 * this needs some work to get unique addresses.
	 */
	iap = bd->bi_enetaddr;

#ifdef CONFIG_RPXCLASSIC
	/* The Embedded Planet boards have only one MAC address in
	 * the EEPROM, but can have two Ethernet ports.  For the
	 * FEC port, we create another address by setting one of
	 * the address bits above something that would have (up to
	 * now) been allocated.
	 */
	for (i=0; i<6; i++)
		tmpaddr[i] = *iap++;
	tmpaddr[3] |= 0x80;
	iap = tmpaddr;
#endif

	for (i=0; i<6; i++)
		dev->dev_addr[i] = *iap++;

	/* Allocate memory for buffer descriptors.
	*/
	if (((RX_RING_SIZE + TX_RING_SIZE) * sizeof(cbd_t)) > PAGE_SIZE) {
		printk("FEC init error.  Need more space.\n");
		printk("FEC initialization failed.\n");
		return 1;
	}
	if (!(mem_addr = __get_free_page(GFP_KERNEL))) {
		printk("FEC init error.  Null pointer for BD.\n");
		return 1;
	}
	cbd_base = (cbd_t *)mem_addr;

	/* Make it uncached.
	*/
	pte = va_to_pte(&init_task, (int)mem_addr);
	pte_val(*pte) |= _PAGE_NO_CACHE;
	flush_tlb_page(current->mm->mmap, mem_addr);

	/* Set receive and transmit descriptor base.
	*/
	fep->rx_bd_base = cbd_base;
	fep->tx_bd_base = cbd_base + RX_RING_SIZE;

	fep->dirty_tx = fep->cur_tx = fep->tx_bd_base;
	fep->cur_rx = fep->rx_bd_base;

	fep->skb_cur = fep->skb_dirty = 0;

	/* Initialize the receive buffer descriptors.
	*/
	bdp = fep->rx_bd_base;
	for (i=0; i<FEC_ENET_RX_PAGES; i++) {

		/* Allocate a page.
		*/
		if (!(mem_addr = __get_free_page(GFP_KERNEL))) {
			printk("FEC init error.  Null pointer for PAGE.\n");
			return 1;
		}

		/* Make it uncached.
		*/
		pte = va_to_pte(&init_task, mem_addr);
		pte_val(*pte) |= _PAGE_NO_CACHE;
		flush_tlb_page(current->mm->mmap, mem_addr);

		/* Initialize the BD for every fragment in the page.
		*/
		for (j=0; j<FEC_ENET_RX_FRPPG; j++) {
			bdp->cbd_sc = BD_ENET_RX_EMPTY;
			bdp->cbd_bufaddr = __pa(mem_addr);
			mem_addr += FEC_ENET_RX_FRSIZE;
			bdp++;
		}
	}

	/* Set the last buffer to wrap.
	*/
	bdp--;
	bdp->cbd_sc |= BD_SC_WRAP;

	/* Install our interrupt handler.
	*/
	if (request_8xxirq(FEC_INTERRUPT, fec_enet_interrupt, 0, "fec", dev) != 0)
		panic("Could not allocate FEC IRQ!");

#ifdef CONFIG_RPXCLASSIC
	/* Make Port C, bit 15 an input that causes interrupts.
	*/
	immap->im_ioport.iop_pcpar &= ~0x0001;
	immap->im_ioport.iop_pcdir &= ~0x0001;
	immap->im_ioport.iop_pcso &= ~0x0001;
	immap->im_ioport.iop_pcint |= 0x0001;
	cpm_install_handler(CPMVEC_PIO_PC15, mii_link_interrupt, dev);
#endif
#ifdef CONFIG_FADS
	if (request_8xxirq(SIU_IRQ2, mii_link_interrupt, 0, "mii", dev) != 0)
		panic("Could not allocate MII IRQ!");
#endif
#ifdef CONFIG_SM100
//	if (request_8xxirq(SIU_IRQ6, mii_link_interrupt, 0, "mii", dev) != 0)
//		panic("Could not allocate MII IRQ!");
#endif

	dev->base_addr = (unsigned long)fecp;
	dev->priv = fep;

	/* The FEC Ethernet specific entries in the device structure. */
	dev->open = fec_enet_open;
	dev->hard_start_xmit = fec_enet_start_xmit;
	dev->stop = fec_enet_close;
	dev->get_stats = fec_enet_get_stats;
	dev->set_multicast_list = set_multicast_list;
	dev->set_mac_address = fec_set_mac_address;
#ifdef CONFIG_PRxK
	dev->do_ioctl = fec_do_ioctl;
#endif

#ifdef CONFIG_PRxK
	for (i=0; i<NMII-1; i++)
		mii_cmds[i].mii_next = &mii_cmds[i+1];
	mii_free = mii_cmds;
#endif

	restart_fec(dev);

	printk("%s: FEC ENET Version 0.1, ", dev->name);
	for (i=0; i<5; i++)
		printk("%02x:", dev->dev_addr[i]);
	printk("%02x\n", dev->dev_addr[5]);

	return 0;
}

/* This function is called to start or restart the FEC during a link
 * change.  This only happens when switching between half and full
 * duplex.
 */
void
restart_fec(struct device *dev)
{
	struct fec_enet_private *fep;
	int i;
	volatile	cbd_t	*bdp;
	volatile	immap_t	*immap;
	volatile	fec_t	*fecp;
	extern		uint	_get_IMMR();
#ifdef CONFIG_PRxK
	bd_t		*bd = (bd_t *)res;
#endif

	immap = (immap_t *)IMAP_ADDR;	/* pointer to internal registers */

	fecp = &(immap->im_cpm.cp_fec);

	fep = dev->priv;

	/* Whack a reset.  We should wait for this.
	*/
	fecp->fec_ecntrl = 1;
	udelay(10);

#ifdef CONFIG_PRxK
	for (i=0; i<NMII-1; i++)
		mii_cmds[i].mii_next = &mii_cmds[i+1];
	mii_free = mii_cmds;
	mii_head = NULL;
#endif

	/* Enable interrupts we wish to service.
	*/
	fecp->fec_imask = (FEC_ENET_TXF | FEC_ENET_TXB |
				FEC_ENET_RXF | FEC_ENET_RXB | FEC_ENET_MII);

	/* Clear any outstanding interrupt.
	*/
	fecp->fec_ievent = 0xffc0;

	fecp->fec_ivec = (FEC_INTERRUPT/2) << 29;

	/* Set station address.
	*/
	fecp->fec_addr_low =	(dev->dev_addr[0] << 24) |
				(dev->dev_addr[1] << 16) |
				(dev->dev_addr[2] <<  8) |
				dev->dev_addr[3];
	fecp->fec_addr_high = (dev->dev_addr[4] << 8) | dev->dev_addr[5];

	/* Reset all multicast.
	*/
	fecp->fec_hash_table_high = 0;
	fecp->fec_hash_table_low = 0;

	/* Set maximum receive buffer size.
	*/
	fecp->fec_r_buff_size = PKT_MAXBLR_SIZE;
	fecp->fec_r_hash = PKT_MAXBUF_SIZE;

	/* Set receive and transmit descriptor base.
	*/
	fecp->fec_r_des_start = __pa((uint)(fep->rx_bd_base));
	fecp->fec_x_des_start = __pa((uint)(fep->tx_bd_base));

	fep->dirty_tx = fep->cur_tx = fep->tx_bd_base;
	fep->cur_rx = fep->rx_bd_base;

	/* Reset SKB transmit buffers.
	*/
	fep->skb_cur = fep->skb_dirty = 0;
	for (i=0; i<=TX_RING_MOD_MASK; i++) {
		if (fep->tx_skbuff[i] != NULL) {
			dev_kfree_skb(fep->tx_skbuff[i]);
			fep->tx_skbuff[i] = NULL;
		}
	}

	/* Initialize the receive buffer descriptors.
	*/
	bdp = fep->rx_bd_base;
	for (i=0; i<RX_RING_SIZE; i++) {

		/* Initialize the BD for every fragment in the page.
		*/
		bdp->cbd_sc = BD_ENET_RX_EMPTY;
		bdp++;
	}

	/* Set the last buffer to wrap.
	*/
	bdp--;
	bdp->cbd_sc |= BD_SC_WRAP;

	/* ...and the same for transmmit.
	*/
	bdp = fep->tx_bd_base;
	for (i=0; i<TX_RING_SIZE; i++) {

		/* Initialize the BD for every fragment in the page.
		*/
		bdp->cbd_sc = 0;
		bdp->cbd_bufaddr = 0;
		bdp++;
	}

	/* Set the last buffer to wrap.
	*/
	bdp--;
	bdp->cbd_sc |= BD_SC_WRAP;

	MAX_INTERRUPT_WORK = bd->max_int_work;
#ifdef CONFIG_PRxK
	if (bd->mii_operation & 2)		// Full or half duplex
		full_duplex = 1;
#endif

	/* Enable MII mode.
	*/
	if (full_duplex) {
		fecp->fec_r_cntrl = 0x04;	/* MII enable */
		fecp->fec_x_cntrl = 0x04;	/* FD enable */
	}
	else {
		//	fecp->fec_r_cntrl = 0x04;
		fecp->fec_r_cntrl = 0x06;	// MII enable|No Rcv on Xmit
		fecp->fec_x_cntrl = 0x00;
	}

	/* Enable big endian and don't care about SDMA FC.
	*/
	fecp->fec_fun_code = 0x78000000;

	/* Set MII speed (50 MHz core).
	*/
#ifdef CONFIG_PRxK
	fecp->fec_mii_speed = bd->bi_busfreq * 2 / 5 + 2;
#else
	fecp->fec_mii_speed = 0x14;
#endif

#ifdef CONFIG_PRxK
	immap->im_ioport.iop_pcdat &= ~(0x0800);
	udelay(30000);
	immap->im_ioport.iop_pcdat |= 0x0800;
	udelay(1000);
#endif

	/* Configure all of port D for MII.
	*/
	immap->im_ioport.iop_pdpar = 0x1fff;

	/* Bits moved from Rev. D onward.
	*/
	if ((_get_IMMR() & 0xffff) < 0x0501)
		immap->im_ioport.iop_pddir = 0x1c58;	/* Pre rev. D */
	else
		immap->im_ioport.iop_pddir = 0x1fff;	/* Rev. D and later */

#if 0
	mii_startup_cmds();
#else
	phyno = 0;
	mii_queue(mk_mii_phyaddr(0), mii_discover_phy);
#endif

	/* And last, enable the transmit and receive processing.
	*/
	fecp->fec_ecntrl = 6;
	fecp->fec_r_des_active = 0x01000000;
}

