
/* MPC8xx CPM I2C interface. Copyright (c) 1999 Dan Malek (dmalek@jlc.net).
 * This is a read only interface, primarily to test the microcode
 * patches.
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include <asm/mpc8xx.h>

#include "commproc.h"

/* This is the I2C device number, but I use the minors to indicate
 * the I2C device address.
 */
#define CPM_IIC_CDEV_MAJOR 89

void	cpm_iic_init(void);
static int cpm_iic_open(struct inode *, struct file *);
static int cpm_iic_close(struct inode *, struct file *);
static ssize_t cpm_iic_read(struct file *, char *, size_t, loff_t *);
static void cpm_iic_interrupt(void *);

static struct file_operations cpm_iic_fops = {
	NULL,		/* lseek */
	cpm_iic_read,		/* read */
	NULL,		/* write */
	NULL,		/* readdir */
	NULL,		/* select */
	NULL,		/* ioctl */
	NULL,		/* mmap */
	cpm_iic_open,
	cpm_iic_close,
	NULL		/* fsync */
};

/* Only one user at a time.
*/
static int	cpm_iic_flags;
#define CPM_IIC_OPEN	0x01

static	dev_t	iic_device;
static	struct wait_queue *iic_wait;

static	ushort r_tbase, r_rbase;

void
cpm_iic_init()
{
	uint	mem_addr, dp_addr, reloc;
	volatile cbd_t		*bdp;
	volatile cpm8xx_t	*cp;
	volatile iic_t		*iip;
	volatile i2c8xx_t	*i2c;
	volatile immap_t	*immap;

	cp = cpmp;	/* Get pointer to Communication Processor */
	immap = (immap_t *)IMAP_ADDR;	/* and to internal registers */

	iip = (iic_t *)&cp->cp_dparam[PROFF_IIC];

	/* Check for and use a microcode relocation patch.
	*/
	if (reloc = iip->iic_rpbase)
		iip = (iic_t *)&cp->cp_dpmem[iip->iic_rpbase];
		
	i2c = (i2c8xx_t *)&(immap->im_i2c);

	/* Initialize Port B IIC pins.
	*/
	cp->cp_pbpar |= 0x00000030;
	cp->cp_pbdir |= 0x00000030;
	cp->cp_pbodr |= 0x00000030;

	/* Initialize the parameter ram.
	 * We need to make sure many things are initialized to zero,
	 * especially in the case of a microcode patch.
	 */
	iip->iic_rstate = 0;
	iip->iic_rdp = 0;
	iip->iic_rbptr = 0;
	iip->iic_rbc = 0;
	iip->iic_rxtmp = 0;
	iip->iic_tstate = 0;
	iip->iic_tdp = 0;
	iip->iic_tbptr = 0;
	iip->iic_tbc = 0;
	iip->iic_txtmp = 0;

	/* Allocate space for one transmit and one receive buffer
	 * descriptor in the DP ram.
	 */
	dp_addr = m8xx_cpm_dpalloc(sizeof(cbd_t) * 2);

	/* Allocate space for receive buffer in the host memory.
	 * The most you can read at once from IIC is 128 bytes.
	 */
	mem_addr = m8xx_cpm_hostalloc(128);

	/* Set up the IIC parameters in the parameter ram.
	*/
	iip->iic_tbase = r_tbase = dp_addr;
	iip->iic_rbase = r_rbase = dp_addr + sizeof(cbd_t);

	/* Set the buffer address.
	*/
	bdp = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];
	bdp->cbd_bufaddr = __pa(mem_addr);

	iip->iic_tfcr = SMC_EB;
	iip->iic_rfcr = SMC_EB;

	/* Set maximum receive size.
	*/
	iip->iic_mrblr = 128;

	/* Initialize Tx/Rx parameters.
	*/
	if (reloc == 0) {
		cp->cp_cpcr =
			mk_cr_cmd(CPM_CR_CH_I2C, CPM_CR_INIT_TRX) | CPM_CR_FLG;
		while (cp->cp_cpcr & CPM_CR_FLG);
	}

	/* Select an arbitrary address.  Just make sure it is unique.
	*/
	i2c->i2c_i2add = 0x34;

	/* Make clock run maximum slow.
	*/
	i2c->i2c_i2brg = 7;

	/* Disable interrupts.
	*/
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;

	/* Install interrupt handler.
	*/
	cpm_install_handler(CPMVEC_I2C, cpm_iic_interrupt, (void *)iip);

	if (register_chrdev(CPM_IIC_CDEV_MAJOR,"iic",&cpm_iic_fops))
		printk("unable to get major %d for IIC devs\n",
				CPM_IIC_CDEV_MAJOR);
}

/* Open does not have to do much, just make sure only one user.
*/
static int
cpm_iic_open(struct inode *ip, struct file *fp)
{

	/* We allow only one user of the device.
	*/
	if (cpm_iic_flags & CPM_IIC_OPEN)	/* device is busy */
		return(EBUSY);

	fp->f_op = &cpm_iic_fops;

	iic_device = MINOR(ip->i_rdev);

	cpm_iic_flags |= CPM_IIC_OPEN;

	return(0);
}

static int
cpm_iic_close(struct inode *ip, struct file *fp)
{
	volatile i2c8xx_t *i2c;

	/* Shut down IIC.
	*/
	i2c = (i2c8xx_t *)&(((immap_t *)IMAP_ADDR)->im_i2c);
	i2c->i2c_i2mod = 0;
	i2c->i2c_i2cmr = 0;
	i2c->i2c_i2cer = 0xff;

	cpm_iic_flags &= ~CPM_IIC_OPEN;

	return(0);
}


/* Read from IIC.
 * This is a two step process.  First, we send the "dummy" write
 * to set the device offset for the read.  Second, we perform
 * the read operation.
 */
static	u_char	iitemp[32];

static ssize_t
cpm_iic_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	volatile iic_t *iip;
	volatile i2c8xx_t *i2c;
	volatile cbd_t	*tbdf, *rbdf;
	volatile cpm8xx_t	*cp;
	u_char			*tb;
	loff_t	offset;
	unsigned long flags;
	uint	reloc;

	cp = cpmp;	/* Get pointer to Communication Processor */
	iip = (iic_t *)&cp->cp_dparam[PROFF_IIC];

	/* Check for and use a microcode relocation patch.
	*/
	if (reloc = iip->iic_rpbase) {
		iip = (iic_t *)&cp->cp_dpmem[iip->iic_rpbase];

		iip->iic_tbase = r_tbase;
		iip->iic_rbase = r_rbase;

		iip->iic_tfcr = SMC_EB;
		iip->iic_rfcr = SMC_EB;

		iip->iic_mrblr = 128;

		iip->iic_rstate = 0;
		iip->iic_rdp = 0;
		iip->iic_rbptr = 0;
		iip->iic_rbc = 0;
		iip->iic_rxtmp = 0;
		iip->iic_tstate = 0;
		iip->iic_tdp = 0;
		iip->iic_tbptr = 0;
		iip->iic_tbc = 0;
		iip->iic_txtmp = 0;
	}

	i2c = (i2c8xx_t *)&(((immap_t *)IMAP_ADDR)->im_i2c);

	tbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_tbase];
	rbdf = (cbd_t *)&cp->cp_dpmem[iip->iic_rbase];

	/* Send a "dummy write" operation.  This is a write request with
	 * only the offset sent, followed by another start condition.
	 * This will ensure we start reading from the first location
	 * of the EEPROM.
	 */
	offset = *ppos;
	tb = iitemp;
	tb = (u_char *)(((uint)tb + 15) & ~15);
	tbdf->cbd_bufaddr = __pa(tb);
	*tb = iic_device & 0xfe;	/* Device address */
	*(tb+1) = offset;		/* Offset */
	tbdf->cbd_datlen = 2;		/* Length */
	tbdf->cbd_sc =
	    BD_SC_READY | BD_SC_INTRPT | BD_SC_LAST | BD_SC_WRAP | BD_IIC_START;

	save_flags(flags); cli();
	i2c->i2c_i2cmr = 0x13;	/* Enable some interupts */
	i2c->i2c_i2cer = 0xff;
	i2c->i2c_i2com = 0x81;	/* Start master */
	i2c->i2c_i2mod = 1;	/* Enable */

	/* Wait for IIC transfer.
	*/
	interruptible_sleep_on(&iic_wait);
	restore_flags(flags);
	if (signal_pending(current))
		return -ERESTARTSYS;

	if (tbdf->cbd_sc & BD_SC_READY)
		printk("IIC ra complete but tbuf ready\n");
	
	/* To read, we need an empty buffer of the proper length.
	 * All that is used is the first byte for address, the remainder
	 * is just used for timing (and doesn't really have to exist).
	 */
	if (reloc) {
		iip->iic_tbase = r_tbase;
		iip->iic_rbase = r_rbase;

		iip->iic_tfcr = SMC_EB;
		iip->iic_rfcr = SMC_EB;

		iip->iic_mrblr = 128;
		iip->iic_rstate = 0;
		iip->iic_rdp = 0;
		iip->iic_rbptr = 0;
		iip->iic_rbc = 0;
		iip->iic_rxtmp = 0;
		iip->iic_tstate = 0;
		iip->iic_tdp = 0;
		iip->iic_tbptr = 0;
		iip->iic_tbc = 0;
		iip->iic_txtmp = 0;
	}
	tbdf->cbd_bufaddr = __pa(tb);
	*tb = iic_device | 1;	/* Device address */
	tbdf->cbd_datlen = rbdf->cbd_datlen = count + 1;	/* Length */
	tbdf->cbd_sc =
	    BD_SC_READY | BD_SC_INTRPT | BD_SC_LAST | BD_SC_WRAP | BD_IIC_START;
	rbdf->cbd_sc = BD_SC_EMPTY | BD_SC_WRAP;

	/* Chip bug, set enable here.
	*/
	save_flags(flags); cli();
	i2c->i2c_i2cer = 0xff;
	i2c->i2c_i2com = 0x81;	/* Start master */
	i2c->i2c_i2mod = 1;	/* Enable */

	/* Wait for IIC transfer.
	*/
	interruptible_sleep_on(&iic_wait);
	restore_flags(flags);
	if (signal_pending(current))
		return -ERESTARTSYS;

	if (rbdf->cbd_sc & BD_SC_EMPTY)
		printk("IIC read complete but rbuf empty\n");
	
	if (copy_to_user(buf, __va(rbdf->cbd_bufaddr), count))
		return -EFAULT;
	
	*ppos += count;

	return count;
}

static void
cpm_iic_interrupt(void *dev_id)
{
	volatile iic_t *iip;
	volatile i2c8xx_t *i2c;

	i2c = (i2c8xx_t *)&(((immap_t *)IMAP_ADDR)->im_i2c);

	iip = (iic_t *)dev_id;

	/* Chip errata, clear enable.
	*/
	i2c->i2c_i2mod = 0;

	/* Clear interrupt.
	*/
	i2c->i2c_i2cer = 0xff;

	/* Get 'me going again.
	*/
	wake_up_interruptible(&iic_wait);
}
