/*----------------------------------------------------------------------------
 *      Copyright (C) Cyclades Corporation, 1998.
 *      All rights reserved
 *      TS110
 *
 *      This file:      cy_io.c
 *      Description:    Routines to handle digital/analog I/O ports.
 *---------------------------------------------------------------------------
 *---------------------------------------------------------------------------
 *      Change History
 *---------------------------------------------------------------------------
 *      Nov/27/2002     V 1.0.0   Arnaldo Zimmermann - Initial Implementation
 *--------------------------------------------------------------------------*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/miscdevice.h>
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/mpc8xx.h>
#include <asm/8xx_immap.h>
#include <linux/cy_sm.h>
#include <asm-ppc/irq.h>
#include <asm/prxk.h>

static int gpio_is_open=0;
static int adc1_is_open=0;
static int adc2_is_open=0;
static int adc_phy_addr[2];
extern ssize_t cpm_iic_read(char *buf, size_t count, size_t flag);
extern ssize_t cpm_iic_write(char *buf, size_t count);

#define ADC_CONF_REG		0x8c	/* ADS1100 ADC in 8 SPS and gain = 1 */
#define ADC_CONV_BIT		0x80	/* conversion bit */
#define ADC1				0
#define ADC2				1

static void msdelay(uint t) 
{
	uint i;
	/* delay t miliseconds */
	for (i=0; i<t; i++) {
		udelay(1000);
	}
}

static long long cy_io_llseek(struct file *file, long long offset, int origin)
{
	return -ESPIPE;
}

/*
 *	Write to I/O Ports.
 */

static ssize_t cy_io_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
    struct inode *inode;
	volatile immap_t *pm = (volatile immap_t *)IMAP_ADDR;
	int status;

	if (!count) {
		return -ESPIPE;
	}

    inode = file->f_dentry->d_inode;
	switch(MINOR(inode->i_rdev))
	{
		case GPIO_MINOR:
			if(!gpio_is_open) {
				return -ESPIPE;
			}
			/* read only. No write action. */			
			break;
		case ADC1_MINOR:
			if(!adc1_is_open) {
				return -ESPIPE;
			}
			status = cpm_iic_write(buf, count);
			return status;
		case ADC2_MINOR:
			if(!adc2_is_open) {
				return -ESPIPE;
			}
			status = cpm_iic_write(buf, count);
			return status;
		default:
			return -ENODEV;
	}
	return 1;
}

/*
 *	Read I/O Port.
 */
 
static ssize_t cy_io_read(struct file *file, char *buf, size_t count, loff_t *ptr)
{
    struct inode *inode;
	volatile immap_t *pm = (volatile immap_t *)IMAP_ADDR;
	int status;

    inode = file->f_dentry->d_inode;

	if(!count) {
		return -ESPIPE;
	}

	switch(MINOR(inode->i_rdev))
	{
		case GPIO_MINOR:
			if(!gpio_is_open) {
				return -ESPIPE;
			}

			/* read GPIO digital ports */
			/* PA6  -> D7 */
			/* PA7  -> D6 */
			/* PA11 -> D5 */
			/* PA5  -> D4 */
			/* PA4  -> D3 */
			/* PA2  -> D2 */
			/* PA1  -> D1 */
			/* PA0  -> D0 */

			status = pm->im_ioport.iop_padat;
			buf[0] = (((status >> 15) & 0x01) | \
					  ((status >> 13) & 0x02) | \
					  ((status >> 11) & 0x04) | \
					  ((status >> 8)  & 0x08) | \
					  ((status >> 6)  & 0x10) | \
					  ((status << 1)  & 0x20) | \
					  ((status >> 2)  & 0x40) | \
					  ((status >> 2)  & 0x80));

			return 1;
		case ADC1_MINOR:
			if(!adc1_is_open) {
				return -ESPIPE;
			}
			/* check if physical address is not present */
			if (!buf[0]) {
				buf[0] = adc_phy_addr[ADC1];
			}
			status = cpm_iic_read(buf, count, I2C_NO_OFFSET|I2C_SET_MASTER);
			return status;
		case ADC2_MINOR:
			if(!adc2_is_open) {
				return -ESPIPE;
			}
			/* check if physical address is not present */
			if (!buf[0]) {
				buf[0] = adc_phy_addr[ADC2];
			}
			status = cpm_iic_read(buf, count, I2C_NO_OFFSET|I2C_SET_MASTER);
			return status;
		default:						
			return -ESPIPE;
	}

	return 1;
}

/*
 *	Open I/O Port.
 */

static int cy_io_open(struct inode *inode, struct file *file)
{
	bd_t    *bd = (bd_t *)res;

	if (bd->hw_info.board_type != BOARD_TS110) {
		return 0;
	}
	switch(MINOR(inode->i_rdev))
	{
		case GPIO_MINOR:
			if(gpio_is_open)
				return -EBUSY;
			gpio_is_open=1;
			break;
		case ADC1_MINOR:
			if(adc1_is_open)
				return -EBUSY;
			if(!adc_phy_addr[ADC1])
				return -ENODEV;
			adc1_is_open=1;
			break;
		case ADC2_MINOR:
			if(adc2_is_open)
				return -EBUSY;
			if(!adc_phy_addr[ADC2])
				return -ENODEV;
			adc2_is_open=1;
			break;
		default:
			return -ENODEV;
	}
	return 0;
}

/*
 *	Release I/O Port.
 */

static int cy_io_release(struct inode *inode, struct file *file)
{
	switch(MINOR(inode->i_rdev))
	{
		case GPIO_MINOR:
			if(!gpio_is_open)
				return -ESPIPE;
			gpio_is_open=0;
			break;
		case ADC1_MINOR:
			if(!adc1_is_open)
				return -ESPIPE;
			adc1_is_open=0;
			break;
		case ADC2_MINOR:
			if(!adc2_is_open)
				return -ESPIPE;
			adc2_is_open=0;
			break;
		default:
			return -ENODEV;
	}
	return 0;
}

/*
 *	Kernel Interfaces
 */
 
 
static struct file_operations cy_io_fops = {
	cy_io_llseek,
	cy_io_read,
	cy_io_write,
	NULL,		/* No Readdir */
	NULL,		/* No Select */
	NULL,		/* No ioctl */
	NULL,		/* No mmap */
	cy_io_open,
	NULL,		/* flush */
	cy_io_release
};

static struct miscdevice cy_gpio_miscdev=
{
	GPIO_MINOR,
	"CY_GPIO",
	&cy_io_fops
};

static struct miscdevice cy_adc1_miscdev=
{
	ADC1_MINOR,
	"CY_ADC1",
	&cy_io_fops
};

static struct miscdevice cy_adc2_miscdev=
{
	ADC2_MINOR,
	"CY_ADC2",
	&cy_io_fops
};


int cy_io_init(void)
{
	bd_t    *bd = (bd_t *)res;
	unsigned char buf[8] = {0};
	int i;
	

	if (bd->hw_info.board_type != BOARD_TS110) {
		return 0;
	}
	
	cy_gpio_init();
	cpm_iic_init();

	misc_register(&cy_gpio_miscdev);
	misc_register(&cy_adc1_miscdev);
	misc_register(&cy_adc2_miscdev);

	buf[1] = ADC_CONF_REG;
	
	/* WARNING: ADC1 physical address MUST BE 
	   lower than ADC2 physical address */
	 
	/* scan first I2C address */
	for (i=ADC_I2C_ADDR_FIRST; i<=ADC_I2C_ADDR_LAST; i+=2) {
		buf[0] = i;
		if (cpm_iic_write(buf, 2)) {
			break;	/* ADC1 found! */
		}
	}	
	if (i > ADC_I2C_ADDR_LAST) {
		adc_phy_addr[ADC1] = 0;		/* ADC1 not found */
		i = ADC_I2C_ADDR_FIRST;		/* next scan starts on first possible phy addr */
		printk("Warning! Failed to init ADC1.\r\n");
	}
	else {
		adc_phy_addr[ADC1] = i;		/* save ADC1 phy addr found */
		i = i+2;					/* next scan starts on next possible phy addr */
		printk("ADC1 at physical address 0x%02x.\r\n", adc_phy_addr[ADC1]);
	}

	/* scan second I2C address */
	buf[1] = ADC_CONF_REG;
	for (; i<=ADC_I2C_ADDR_LAST; i+=2) {
		buf[0] = i;
		if (cpm_iic_write(buf, 2)) {
			break;	/* ADC2 found */
		}
	}	
	if (i > ADC_I2C_ADDR_LAST) {
		adc_phy_addr[ADC2] = 0;		/* ADC2 not found */
		printk("Warning! Failed to init ADC2.\r\n");
	}
	else {
		adc_phy_addr[ADC2] = i;		/* save ADC2 phy addr found */
		printk("ADC2 at physical address 0x%02x.\r\n", adc_phy_addr[ADC2]);
	}

	//printk("I/O init done.\r\n");
	return 0;
}

int cy_gpio_init(void)
{
	volatile immap_t *pm = (volatile immap_t *)IMAP_ADDR;

    pm->im_ioport.iop_padir &= ~0xEF10;
    pm->im_ioport.iop_paodr &= ~0xEF10;
    pm->im_ioport.iop_papar &= ~0xEF10;
	return 0;
}

