/*----------------------------------------------------------------------------
 *      Copyright (C) Cyclades Corporation, 1998.
 *      All rights reserved
 *      PR-3000 TS
 *
 *      This file:      cy_zbus.c
 *      Description:    Routines to initialize the ZBUS card
 *---------------------------------------------------------------------------
 *---------------------------------------------------------------------------
 *      Change History
 *---------------------------------------------------------------------------
 *      Aug/30/2000     V 1.0.0    Edson Ap. Seabra - Initial Implementation
 *--------------------------------------------------------------------------*/

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/tty.h>
#include <linux/major.h>
#include <linux/interrupt.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <linux/blk.h>
#include <linux/ioport.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/mmu.h>
#include <asm/processor.h>
#include <asm/residual.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/ide.h>
#include <asm/machdep.h>

#include <asm/mpc8xx.h>
#include <asm/8xx_immap.h>

#include <asm/fpga.h>
#include <asm/flash.h>

#include <asm/cy_zbus.h>

#include <asm/zbus_fpga.h>

#define ZBUS_PROG_SIZE sizeof(zbus_program)

#define cyc_read_byte(x) *(volatile unsigned char *)(x)

static ZBUS_INFO zInfo[NUMBER_OF_BOARDS] = {{0}};

unsigned char board_detect[NUMBER_OF_BOARDS];

static unsigned char wait_table[] = {3,3,4,4,5,5,6,6,7,3,7,7,7,7,7,7};

unsigned char dummy;

extern int rs_initialize_zbus_tty(ZBUS_INFO *);

/*-----------------------------------------------------------------------------
 * cyc_zbus_detect_power()
 *-----------------------------------------------------------------------------
 * Description: this routine detects the power suply type (Externla/Internal)
 * Parameters:
 *
 * board_addr - board address
 *
 *  PS. This routine do not affect programmed ports.
 *
 *-----------------------------------------------------------------------------
 */

int cyc_zbus_detect_power(unsigned char *pboard)	
{
	volatile ZBUS_REGS *zbus = (ZBUS_REGS *) pboard;
	volatile SEX_REGS *sexr;

	sexr = (SEX_REGS *) (pboard + 0x10000L);	// For the first SE	

	// Forces changes in ZBUS high address
	dummy = cyc_read_byte(pboard + 0x10000L + 0x400L);

	if (sexr->id == SEX_ID) {
		return (zbus->sense);
	}
	return (Z_NO_EXTERNAL_POWER);
}

/*-----------------------------------------------------------------------------
 * cyc_zbus_reset_uart()
 *-----------------------------------------------------------------------------
 * Description: this routine reset and check if uart is OK
 * Parameters:
 *
 * uart - UART registers
 *
 *-----------------------------------------------------------------------------
 */

int cyc_zbus_reset_uart(volatile UART_REGS *uart)
{
	unsigned char buffer = 0;
	int cnt = 0;

	uart->w.ier = IER_CLEAR;
	uart->w.mcr = MCR_LLOOPBACK;	//To disable Rx

	while ((uart->r.lsr) & LSR_RECEIVE_DATA_READY) {
		buffer += uart->r.rhr;
		if (cnt++ > UART_FIFO_SIZE) {
			break;
		}
	}

	buffer += uart->r.lsr;	//clear registers
	buffer += uart->r.msr;
	buffer += uart->r.isr;

	uart->w.mcr = MCR_CLEAR;
		
	uart->w.lcr = LCR_EFR_ENABLE;
	uart->s.efr = 0x10;	 
	uart->w.lcr = 0x00;

	uart->w.mcr = 0x00;		//reset CLOCK_SELECT bit;
	uart->w.ier = 0x00;		//reset Xoff/RTS interrupts

	uart->w.lcr = LCR_EFR_ENABLE;
	uart->s.xon1 = 0x00;
	uart->s.xon2 = 0x00;
	uart->s.xoff1 = 0x00;
	uart->s.xoff2 = 0x00;
	uart->s.efr = 0x00;	 
	uart->w.lcr = 0x00;

	uart->w.lcr = LCR_DLAB;	// Set LCR bit 7 to define baud rate
	uart->s.dll = 0x00;
	uart->s.dlm = 0x00;
	uart->w.lcr = 0x00;

	cnt = 0;
	do {
		buffer += uart->r.lsr;	//clear registers
		buffer += uart->r.msr;
		buffer += uart->r.isr;
		if (buffer);
		buffer = (uart->r.isr & 1);
		if (++cnt > 100) {
			return(1);
		}
	} while (buffer == 0);
	return(0);
}

/*--------------------------------------------------------------------------
 * Routine:       cyc_zbus_initialize ()
 *
 * Parameters:
 *    nboard: number of board (0, 1, 2);
 *    board_addr: board address
 *
 * Description:
 *   Initializes ZBus boards performing: Reset, Power Suplly sense,
 *   Number of SEX's connected, Number of ports of each SEX,
 *   Program ZBus Wait States
 *
 * Return:
 *  number of Serial Expander detected.
 *
 *---------------------------------------------------------------------------
 */
int cyc_zbus_initialize (int nboard, unsigned char *pboard)
{
	volatile ZBUS_REGS *zbus = (ZBUS_REGS *) pboard;
	volatile SEX_REGS *sexr;
	int i, j, power, nsex;
	ZBUS_INFO *pzinfo = &zInfo[nboard];

	zbus->irq |= 0xF0;

	pzinfo->fw_version = ZBUS_FW_VERSION;

	pzinfo->base_address = pboard;

	if (zbus->id != ZBUS_ID) {
		printk ("Card on slot: %d is not a ZBUS! Initialization failed...\n", nboard+1);
		return (0);
	}

	zbus->reset = 0;
	udelay (200);

	printk ("ZBus card on slot %d\n", nboard + 1);

	zbus->reset = 1; 
	udelay (500);
	zbus->reset = 0;
	udelay (500);

	power = cyc_zbus_detect_power(pboard);	// PWRMON - EHF - DEC/14/99
	if (power & Z_EXTERNAL_POWER_ACTIVE) {
		zbus->power = Z_DISABLE_INTERNAL_POWER;
		udelay (100);
		printk("Using external power.\n");
		pzinfo->power = Z_EXTERNAL_POWER_ACTIVE;// PWRMON - EHF - DEC/14/99
	} else {
		zbus->power = Z_ENABLE_INTERNAL_POWER;
		udelay (100);
		printk ("Using internal power.\n");
		pzinfo->power = Z_NO_EXTERNAL_POWER;	// PWRMON - EHF - DEC/14/99
	}
	

	for (i = 0, nsex = 0; i < MAX_SEXS_PER_SLOT; i++) {
		sexr = (SEX_REGS *) (pboard + 0x10000L + i * 0x1000L);

		// Forces changes in ZBUS high address
		dummy = cyc_read_byte(pboard + 0x10000L + 0x400L);

		if (sexr->id == SEX_ID && sexr->version == 0x04) {
			nsex++;
			for(j=0; j < MAX_PORTS_PER_SEX; j++) {
				cyc_zbus_reset_uart(&sexr->uart[j]);
			}
		}
	}

	if (nsex == 0) {
		printk (" No Serial Expanders were found...\n");
		zbus->power = Z_DISABLE_INTERNAL_POWER;
		return(0);
	}
	
	if ((power == Z_NO_EXTERNAL_POWER) && (nsex > 2)) {
		zbus->power = Z_DISABLE_INTERNAL_POWER;
		printk ("Too many active Serial Expanders.\n"
				"Internal Power Supply deactivated. Cannot proceed!\n");
		return (0);
	}

	printk ("Wait States: %d\n", wait_table[nsex]);
	zbus->wait_states = wait_table[nsex]; 
	udelay (1);

	pzinfo->nsex = nsex;
	pzinfo->nirq = 1 + nboard;
	pzinfo->status_flag = 1;	//Initiated	
	pzinfo->first_tty = 0xffff;

	if (rs_initialize_zbus_tty(pzinfo) != nsex * 16) {
		printk ("Too many serial ports active.\n"
			"Only 64 ports are supported!!!\n");
	}

	if (pzinfo->first_tty == 0xffff) {
		return(0);
	}

	zbus->irq &= (pzinfo->nirq<<4); //bits 7-4: 0, 1, 2 or 3

	printk ("Ports: %d IRQ: %d ttyS%d to ttyS%d\n", nsex * 16,
		pzinfo->nirq, pzinfo->first_tty, pzinfo->last_tty);

	return (nsex);
}

/*-----------------------------------------------------------------------------
 *	cyc_loadfpga()
 *-----------------------------------------------------------------------------
 *	Description: this routine loads the daughter board FPGA.
 *	Parameters:
 *
 *	board_addr   -	board address
 *	fpga_program -	FPGA program data address
 *	size		 -	size of the program data
 *-----------------------------------------------------------------------------
 */

void cyc_loadfpga(unsigned char *board_addr, unsigned char *fpga_program, unsigned int size)
{
	while (size--) {
		*board_addr = *fpga_program++;
		udelay(40);
	}
}

/*-----------------------------------------------------------------------------
 *	cyc_start_stop_SCC()
 *-----------------------------------------------------------------------------
 *	Description: this routine start/stop the daughter board
 *	Parameters:
 *
 *	nboard - board number [0,1,2]
 *	flag - Start/stop
 *-----------------------------------------------------------------------------
 */

void cyc_start_stop_scc(int nboard, int flag)
{
	scc_t *scc_regs;

	scc_regs = &((immap_t *)IMAP_ADDR)->im_cpm.cp_scc[nboard];

	if (flag) {
        	scc_regs->scc_gsmrl |= (3 << 4);
	} else {
        	scc_regs->scc_gsmrl &= ~(3 << 4);
	}

}

/*-----------------------------------------------------------------------------
 *	prxk_detect_board()
 *-----------------------------------------------------------------------------
 *	Description: this routine detects the daughter boards and loads the 
 *	daughter boards FPGA according to the board detected. If the board is
 *	detected and the FPGA is successfully loaded, board_detect[i] is
 *	assigned the board ID.
 *-----------------------------------------------------------------------------
 */
void prxk_detect_board(void)
{
	unsigned char *pboard[NUMBER_OF_BOARDS];
	int bdn;
	ZBUS_REGS *zbus;
	unsigned long flags;

	//save_flags(flags); cli();

	fpga->rcr = 0;

	udelay(200);

	/*
	 * Verify if there are daughter boards in the connectors 
	 */

	/* Internal transfer acknowledge */
	((immap_t *)IMAP_ADDR)->im_memctl.memc_or6 &= ~0x00000008;

	for (bdn = 0; bdn < NUMBER_OF_BOARDS; bdn++) {
		/* Read an address of the board */
		pboard[bdn] = board_addr + DB_FPGA_ADDR_OFF(bdn);

		dummy = cyc_read_byte (pboard[bdn]);

		/* Read the daughter board presence detect register */
		board_detect[bdn] = fpga->pdr & 0xf8;
	}

	udelay(200);

	for (bdn = 0; bdn < NUMBER_OF_BOARDS; bdn++) {
		switch (board_detect[bdn]) {
			case ZBUS_CARD_DET:
				save_flags(flags); cli();
				cyc_start_stop_scc(bdn+1, 0);
				cyc_loadfpga(pboard[bdn], zbus_program, ZBUS_PROG_SIZE);
				restore_flags(flags);
				break;
			default:
				break;
		}
	}

	udelay(1000);

	/* External transfer acknowledge */
	((immap_t *)IMAP_ADDR)->im_memctl.memc_or6 |= 0x00000008;
	
	/*
	 * Verify if the fpgas were successfully changed.
	 */
	for (bdn = 0; bdn < NUMBER_OF_BOARDS; bdn++) {
		switch (board_detect[bdn]) {
			case ZBUS_CARD_DET:
				zbus = (ZBUS_REGS *)(pboard[bdn]);
				if (zbus->id != ZBUS_ID) {
					board_detect[bdn] = NO_CARD_DET;
					break;
				}
				if (!cyc_zbus_initialize(bdn, pboard[bdn])) {
					board_detect[bdn] = NO_CARD_DET;
					break;
				}
				cyc_start_stop_scc(bdn+1, 1);
				break;
			default:
				board_detect[bdn] = NO_CARD_DET;
				break;
		}
	}
	//restore_flags(flags);
}
