/* -*- linux-c -*- */
/*
    This file is part of cacpd

    Copyright (C) 2005 Cyclades Corp.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/* Author: Oliver Kurth <oliver.kurth@cyclades.com> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "databuf.h"

#define MIN_BUF_SIZE 1024
#define MAX_BUF_SIZE 8192
#define MAX_READ 1024

#undef max
#define max(x,y) ((x) > (y) ? (x) : (y))
#undef min
#define min(x,y) ((x) < (y) ? (x) : (y))

struct databuf *create_databuf()
{
	struct databuf *dbuf;

	dbuf = malloc(sizeof(struct databuf));
	if(dbuf){
		memset(dbuf, 0, sizeof(struct databuf));
		dbuf->size = MIN_BUF_SIZE;
		dbuf->buf = (char *)malloc(dbuf->size);
		if(dbuf->buf == NULL){
			destroy_databuf(dbuf);
			dbuf = NULL;
		}
	}
	return dbuf;
}

void destroy_databuf(struct databuf *dbuf)
{
	if(dbuf){
		if(dbuf->buf) free(dbuf->buf);
		free(dbuf);
	}
}

static
int databuf_reserve(struct databuf *dbuf, int len)
{
	if(dbuf->buf == NULL){
		dbuf->size = len + MIN_BUF_SIZE;
		dbuf->buf = (char *)malloc(dbuf->size);
		dbuf->end = dbuf->start = 0;
	}else if(dbuf->end + len > dbuf->size){
		/* does not fit in buffer */
		if(dbuf->start > 0){
			/* moving data to the beginning may help: */
			memmove(dbuf->buf, dbuf->buf + dbuf->start, dbuf->end - dbuf->start);
			dbuf->end -= dbuf->start;
			dbuf->start = 0;
		}
		if(dbuf->end + len > dbuf->size){
			/* No, did not help. We have to reallocate: */
			dbuf->size = dbuf->end + len + MIN_BUF_SIZE;
			dbuf->buf = (char *)realloc(dbuf->buf, dbuf->size);
		}
	}
	return dbuf->size - dbuf->end;
}

int databuf_add(struct databuf *dbuf, const char *data, int len)
{
	databuf_reserve(dbuf, len);

	memcpy(dbuf->buf + dbuf->end, data, len);
	dbuf->end += len;

	return 0;
}

int databuf_add_string(struct databuf *dbuf, const char *str)
{
	return databuf_add(dbuf, str, strlen(str));
}

int databuf_remove(struct databuf *dbuf, int len)
{
//	printf("databuf_remove(), len=%d\n", len);

	dbuf->start += len;
	/* prevent buffer from getting too large. */
	if(dbuf->start > MAX_BUF_SIZE){
		char *newbuf;

		dbuf->size = (dbuf->end - dbuf->start) + MIN_BUF_SIZE;
		newbuf = malloc(dbuf->size);
		memcpy(newbuf, dbuf->buf + dbuf->start, dbuf->end - dbuf->start);
		dbuf->end -= dbuf->start;
		dbuf->start = 0;
		free(dbuf->buf);
		dbuf->buf = newbuf;
	}
	return 0;
}

int databuf_flushout(struct databuf *dbuf, int fd)
{
	int sz = 0;

	if(databuf_length(dbuf) > 0){
		/* attempt to write as much as possible */
		sz = write(fd,
			   databuf_get(dbuf),
			   databuf_length(dbuf));
		if(sz > 0)
			databuf_remove(dbuf, sz);
	}
	return sz;
}

#if 0
int databuf_flushin(struct databuf *dbuf, int fd)
{
	int sz;
	char data[MAX_READ];

	sz = read(fd, data, sizeof(data));
	if(sz > 0)
		databuf_add(dbuf, data, sz);
	return sz;
}
#else
int databuf_flushin(struct databuf *dbuf, int fd)
{
	int sz;

	databuf_reserve(dbuf, MAX_READ);

	sz = read(fd, databuf_put(dbuf), MAX_READ);
	if(sz > 0)
		dbuf->end += sz;

	return sz;
}
#endif
