/* -*- linux-c -*- */
/*
    This file is part of logpost

    Copyright (C) 2006 Cyclades Corp.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/* Author: Oliver Kurth <oliver.kurth@avocent.com> */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/ip.h> 
#include <arpa/inet.h>
#include <sys/stat.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <sys/wait.h>

#include <libdaemon/dfork.h>
#include <libdaemon/dsignal.h>
#include <libdaemon/dlog.h>
#include <libdaemon/dpid.h>
#include <libdaemon/dexec.h>

#include <llconf/nodes.h>
#include <llconf/entry.h>
#include <llconf/modules.h>
#include <llconf/allmodules.h>
#include <llconf/cyconf.h>

#include "debug.h"

#include "loop.h"
#include "lines.h"
#include "databuf.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#else
#define SYSCONFDIR "/etc/logpipe/"
#endif

volatile int g_debug_level = 0;

struct line *g_line_list = NULL;

#define READ_BUF_SIZE 16384
#define WRITE_BUF_LIMIT (1048576-1)

#define CONN_TYPE_IN 0x01
#define CONN_TYPE_OUT 0x02
#define CONN_TYPE_MON 0x04
#define CONN_TYPE_SAVE 0x08
#define CONN_TYPE_URGENT 0x10

#define CONN_SOCK_TYPE_PIPE 0x01
#define CONN_SOCK_TYPE_TCP 0x02
#define CONN_SOCK_TYPE_FILE 0x03

struct connection{
	struct connection *next;
	struct proxy *proxy;
	char *name;

	int type;
	int fd;
	char *spec;
	char *sub_spec;
	int (*open)(const char *spec);

	struct databuf *databuf;
	int max_buf;

	/* we may need a process which sets up a tunnel
	   (most likely ssh): */
	char *tunnel_cmd;
	int tunnel_cmd_pid;
	int tunnel_timeout;
	int tunnel_threshold;

	int alert_low, alert_med, alert_high;
	char *alert_cmd;
	char *alert_pipe;
	int alerted;

	struct fd_callback *cb;

	/* for tcp write, we try to read, to check for the
	   other end closing the connection: */
	struct fd_callback *cb_mon;

	/* file stuff: */
	int bytes_written;
	int max_file_size;

	time_t last_write;
};

struct proxy{
	struct line *line_buf;
	struct connection *conn_list;
	int interval;
};

static
struct proxy *create_proxy(void)
{
	struct proxy *proxy;

	if((proxy = (struct proxy *)malloc(sizeof(struct proxy)))){
		memset(proxy, 0, sizeof(struct proxy));
	}
	return proxy;
}

static
void proxy_add_connection(struct proxy *proxy, struct connection *conn)
{
	proxy->conn_list = (struct connection *)slist_append(
		(struct slist_hdr *)(proxy->conn_list), (struct slist_hdr *)conn);
	conn->proxy = proxy;
}

static
struct connection *create_connection(const char *spec, int type)
{
	struct connection *conn;

	if((conn = (struct connection *)malloc(sizeof(struct connection)))){
		memset(conn, 0, sizeof(struct connection));
		conn->spec = strdup(spec);
		conn->type = type;
		conn->databuf = create_databuf();
		conn->last_write = time(NULL);
	}
	return conn;
}

static
void conn_close(struct connection *conn)
{
	if(conn->fd >= 0)
		close(conn->fd);
	conn->cb->type = 0;
	conn->cb->fd = -1;
	if(conn->cb_mon){
		conn->cb_mon->type = 0;
		conn->cb_mon->fd = -1;
	}
}

static
int conn_open(struct connection *conn)
{
	int fd;

	fd = conn->open(conn->sub_spec);
	if(fd >= 0){
		conn->fd = conn->cb->fd = fd;
		if(conn->type & CONN_TYPE_OUT)
			conn->cb->type |= FDCB_WRITE;
		else
			conn->cb->type |= FDCB_READ;
		if(conn->cb_mon){
			conn->cb_mon->fd = fd;
			conn->cb_mon->type |= FDCB_READ;
		}
	}
	return fd;
}

static
int listen_sock_in(const char *ip, int port)
{
	int s;
	struct sockaddr_in sa;
	int yes = 1;
	int ret;

	s = socket(PF_INET, SOCK_STREAM, 0);
	if(s < 0){
		daemon_log(LOG_ERR, "socket: %s\n", strerror(errno));
		exit(1);
	}
	ret = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &yes, sizeof (yes));
	if(ret != 0){
		daemon_log(LOG_ERR, "setsockopt: %s\n", strerror(errno));
	}
	
	/* set close on exec flag: */
	ret = fcntl(s, F_SETFD, FD_CLOEXEC);
	if(ret != 0){
		daemon_log(LOG_ERR, "fcntl: %s\n", strerror(errno));
	}

	sa.sin_family = AF_INET;
	inet_aton(ip, &sa.sin_addr);
	sa.sin_port = htons(port);
	if(bind(s, (struct sockaddr *)&sa, sizeof(sa)) != 0){
		daemon_log(LOG_ERR, "bind: %s\n", strerror(errno));
		exit(1);
	}

	listen(s, 1);

	return s;
}

static
int connect_sock_in(const char *ip, int port)
{
	int s = -1;
	struct sockaddr_in sa;
	int yes = 1;
	int ret;

	errno = 0;
	s = socket(PF_INET, SOCK_STREAM, 0);
	if(s < 0){
		daemon_log(LOG_ERR, "socket: %s\n", strerror(errno));
		return -1;
	}
	ret = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &yes, sizeof (yes));
	if(ret != 0){
		daemon_log(LOG_ERR, "setsockopt: %s\n", strerror(errno));
		return -1;
	}
	
	/* set close on exec flag: */
	ret = fcntl(s, F_SETFD, FD_CLOEXEC);
	if(ret != 0){
		daemon_log(LOG_ERR, "fcntl: %s\n", strerror(errno));
		return -1;
	}

	sa.sin_family = AF_INET;
	inet_aton(ip, &sa.sin_addr);
	sa.sin_port = htons(port);

	errno = 0;
	if(connect(s, (struct sockaddr *)&sa, sizeof(sa)) != 0){
		daemon_log(LOG_INFO, "connect: %s\n", strerror(errno));
		close(s);
		return -1;
	}

	return s;
}

static
int parse_tcp_spec(const char *spec, char *ip, int *pport)
{
	const char *p = spec;
	char *q = ip;

	while(*p && (*p != ':') && (q < ip+15))
		*q++ = *p++;
	*q = 0;
	if(*p == ':'){
		p++;
		*pport = atoi(p);
		return 0;
	}
	return -1;
}

static
int open_read_pipe(const char *pipename)
{
	DEBUG(5) printf("%s\n", __FUNCTION__);

	mkfifo(pipename, 0666);
	chmod(pipename, 0666);

	return open(pipename, O_RDONLY|O_NONBLOCK);
}

static
int open_write_pipe(const char *pipename)
{
	DEBUG(5) printf("%s\n", __FUNCTION__);

	mkfifo(pipename, 0600);
	return open(pipename, O_WRONLY);
}

static
int open_write_file(const char *filename)
{
	DEBUG(5) printf("%s\n", __FUNCTION__);

	return open(filename, O_WRONLY|O_CREAT|O_APPEND, 0644);
}

static
int open_write_tcp(const char *spec)
{
	int fd;
	char ip[16];
	int port;

	DEBUG(5) printf("%s\n", __FUNCTION__);

	parse_tcp_spec(spec, ip, &port);
	fd = connect_sock_in(ip, port);

	DEBUG(5) printf("%s, fd = %d, error = %s\n", __FUNCTION__, fd, strerror(errno));

	/* sleeping increases the chance that we get an error on write,
	   if no one listens on the other side of the tunnel. */
	usleep(10000);

	return fd;
}

static
int conn_setup_tunnel(struct connection *conn)
{
	pid_t pid;

	DEBUG(5) printf("%s\n", __FUNCTION__);
	DEBUG(1) printf("starting tunnel '%s'\n", conn->tunnel_cmd);

	if(conn->tunnel_cmd_pid > 0){
		printf("tunnel already running with pid %d\n", conn->tunnel_cmd_pid);
		return 0;
	}

	pid = fork();
	if(pid == 0){
#if 1
                int fd1, fd2;
		char *args[16], *p, *q, buf[256];
		int i;
		
		i = 0;
		p = conn->tunnel_cmd;
		while(*p && i < 15){
			while(*p && isspace(*p)) p++;
			q = buf;
			while(*p && !isspace(*p) && (q - buf) < 255) *q++ = *p++;
			*q = 0;
			args[i] = strdup(buf);
			DEBUG(3) printf("args[%d] = %s\n", i, args[i]);
			i++;
		}
		args[i] = NULL;

                close(1); close(2);
                fd1 = open("/dev/null", O_WRONLY);
                fd2 = open("/dev/null", O_WRONLY);

                dup2(fd1, 1);
                dup2(fd2, 2);

//                execl("/bin/sh", "/bin/sh", "-c", conn->tunnel_cmd, NULL);
                execv(args[0], args);
		/* no need to free args - we have terminated */
#else
		int ret;
		daemon_exec("/", &ret, "/bin/sh", "-c", conn->tunnel_cmd, NULL);
#endif
		exit(0);
	}else if(pid > 0){
		conn->tunnel_cmd_pid = pid;
		DEBUG(2) printf("tunnel_pid = %d\n", pid);
		sleep(1);
		return 0;
	}
	return -1;
}

void conn_save_counter(struct connection *conn)
{
	char fname[256];
	FILE *fptr;

	snprintf(fname, sizeof(fname), "%s.counter", conn->sub_spec);
	fptr = fopen(fname, "wt");
	if(fptr){
		fprintf(fptr, "%d\n", conn->bytes_written);
		fclose(fptr);
	}
}

static void conn_alert(struct connection *conn, int level)
{
	int size = databuf_length(conn->databuf);

	if(conn->alert_cmd)
		system(conn->alert_cmd);

	if(conn->alert_pipe){
		FILE *fptr;
		if(fptr = fopen(conn->alert_pipe, "wt")){
			switch(level){
			case 1:
				DEBUG(2) printf("sending EVT18 %d %d \n", time(NULL), size);
				fprintf(fptr, "EVT18 %d %d \n", time(NULL), size);
				break;
			case 2:
				DEBUG(2) printf("sending EVT19 %d %d \n", time(NULL), size);
				fprintf(fptr, "EVT19 %d %d \n", time(NULL), size);
				break;
			case 3:
				DEBUG(2) printf("sending EVT20 %d \n", time(NULL));
				fprintf(fptr, "EVT20 %d \n", time(NULL));
				break;
			}
			fclose(fptr);
		}
	}

	conn->alerted = level;
}

static
void conn_check_alerts(struct connection *conn)
{
	int size = databuf_length(conn->databuf);

	if((conn->alert_high > 0) && (size >= conn->alert_high)){
		if(conn->alerted != 3) conn_alert(conn, 3);
	}else if((conn->alert_med > 0) && (size >= conn->alert_med)){
		if(conn->alerted != 2) conn_alert(conn, 2);
	}else if((conn->alert_low > 0) && (size >= conn->alert_low)){
		if(conn->alerted != 1) conn_alert(conn, 1);
	}
}

static
void proxy_push(struct proxy *proxy, const char *buf)
{
	struct connection *conn;

	for(conn = proxy->conn_list; conn; conn = conn->next){
		if(conn->type & CONN_TYPE_OUT){

			conn_check_alerts(conn);

			if(databuf_length(conn->databuf) < conn->max_buf){
				databuf_add(conn->databuf, buf, strlen(buf));

			}else{
				if(conn->alert_cmd && !conn->alerted){
					system(conn->alert_cmd);
					conn->alerted = 1;
					DEBUG(1) printf("buffer full for %s: %d >= %d\n",
							conn->name, databuf_length(conn->databuf), conn->max_buf);
				}
			}
			if((databuf_length(conn->databuf) > 0) && (conn->fd >= 0))
				conn->cb->type |= FDCB_WRITE;
			if(databuf_length(conn->databuf) > conn->tunnel_threshold){
				/* check if tunnel is running, if we need one: */
				if(conn->tunnel_cmd && (conn->tunnel_cmd_pid == 0)){
					DEBUG(2) printf("%s calling conn_setup_tunnel()\n", __FUNCTION__);
					conn_setup_tunnel(conn);
				}
			}
		}
	}
}

static
int mon_callback(int fd, void *data)
{
	char trash[1024];
	int sz;
	struct connection *conn = (struct connection *)data;

	sz = read(fd, trash, sizeof(trash));
	if(sz <= 0){
		close(fd);
		conn->fd = -1;
		conn->cb->type = conn->cb_mon->type = 0;
		return 1;
	}
	return 0;
}

static
int read_callback(int fd, void *data)
{
	int sz = 0;
	struct connection *conn = (struct connection *)data;
	struct proxy *proxy = conn->proxy;
	struct databuf *ibuf = conn->databuf;
	int ret;

	DEBUG(5) printf("read_callback()\n");

	sz = databuf_flushin(ibuf, fd);

	if(sz <= 0){
		int save_errno = errno;
		if(sz < 0){
			daemon_log(LOG_ERR, "session_flushin() failed: %s; closing session", strerror(save_errno));
		}else{ /* sz == 0 */
			DEBUG(5) printf("client closed connection");
		}
		conn_close(conn);
		conn_open(conn);
		return -1;
	}

	while(1){
		char buf[READ_BUF_SIZE];
		char *p = databuf_get(ibuf);
		char *q = buf;

		while((p < databuf_get(ibuf) + databuf_length(ibuf)) &&
		      (*p != '\n') && (q < buf + sizeof(buf) - 1)){
			*q++ = *p++;
		}

		/* add trailing linefeed: */
		if((*p == '\n') && (q < buf + sizeof(buf) - 1))
			*q++ = *p;

		/* if the line is too long, we slam the door: */
		if(q >= (buf + sizeof(buf) - 1)){
			daemon_log(LOG_ERR, "line too long, closing session");
			return -1;
		}
		if(p >= databuf_get(ibuf) + databuf_length(ibuf))
			break;

		*q = 0;

		proxy_push(conn->proxy, buf);
		databuf_remove(ibuf, (p - databuf_get(ibuf)) + 1);
	}

	if(conn->type & CONN_TYPE_URGENT){
		struct connection *conn_out;
		for(conn_out = proxy->conn_list; conn_out; conn_out = conn_out->next){
			if(conn_out->type & CONN_TYPE_OUT){
				/* check if tunnel is running, if we need one: */
				if(conn_out->tunnel_cmd && (conn_out->tunnel_cmd_pid == 0)){
					DEBUG(2) printf("%s calling conn_setup_tunnel()\n", __FUNCTION__);
					conn_setup_tunnel(conn_out);
				}
				if(conn_out->fd < 0){
					conn_open(conn_out);
				}
			}
		}
	}

	return 0;
}

static
int write_callback(int fd, void *data)
{
	struct connection *conn = (struct connection *)data;
	int sz, sz_done;

	sz = databuf_length(conn->databuf);
	sz_done = databuf_flushout(conn->databuf, fd);

	DEBUG(5) printf("write_callback() %d bytes written\n", sz);

	if(sz_done == sz){
		conn->cb->type &= ~FDCB_WRITE;
		conn->bytes_written += sz_done;
	}else if(sz_done <= 0){
		conn_close(conn);
	}
	if((conn->max_file_size > 0) && (conn->bytes_written >= conn->max_file_size)){
		conn_close(conn);
	}
	conn->last_write = time(NULL);
	conn->alerted = 0;

	conn_save_counter(conn);
}

static
int time_callback(int fd, void *data)
{
	struct proxy *proxy = (struct proxy *)data;
	struct connection *conn_save, *conn;
	time_t now = time(NULL);

	DEBUG(5) printf("%s\n", __FUNCTION__);

	for(conn = proxy->conn_list; conn; conn = conn->next){

		DEBUG(5){
			printf(" spec: %s\n", conn->spec);
			printf(" type: %d\n", conn->type);
			printf(" fd: %d\n", conn->fd);
			printf(" bytes written: %d\n", conn->bytes_written);
			printf(" pending bytes: %d\n", databuf_length(conn->databuf));
			printf(" last write: %d seconds ago\n", now - conn->last_write);
			printf("\n");
		}

		if(conn->type & CONN_TYPE_OUT){
			/* tunnel maintanance: */
			/* check for health only if we have data to send */
			if(databuf_length(conn->databuf) > 0){

				/* check if tunnel is running, if we need one: */
				if(conn->tunnel_cmd && (conn->tunnel_cmd_pid == 0)){
					DEBUG(2) printf("%s calling conn_setup_tunnel()\n", __FUNCTION__);
					conn_setup_tunnel(conn);
				}

				/* check for valid fd if tunnel is okay, or not needed: */
				if((!conn->tunnel_cmd) || (conn->tunnel_cmd_pid != 0)){
					if(conn->fd < 0){
						int fd;

						fd = conn->open(conn->sub_spec);
						if(fd >= 0){
							conn->fd = conn->cb->fd = fd;
							conn->cb->type |= FDCB_WRITE;
							if(conn->type & CONN_TYPE_MON){
								conn->cb_mon->fd = fd;
								conn->cb_mon->type |= FDCB_READ;
							}
						}else{
							daemon_log(LOG_INFO, "conn->open(%s) failed: %s\n",
								   conn->sub_spec, strerror(errno));
						}
					}
				}
			}
			
			/* close tunnel if idle: */
			if(conn->tunnel_cmd && conn->tunnel_cmd_pid > 0)
				if(((now - conn->last_write) > conn->tunnel_timeout) &&
				   (databuf_length(conn->databuf) == 0)){
					DEBUG(1) printf("killing tunnel (pid=%d)\n", conn->tunnel_cmd_pid);
					kill(conn->tunnel_cmd_pid, SIGQUIT);
			}
		} /* if(conn->type & CONN_TYPE_OUT) */
		if(conn->type & CONN_TYPE_IN){
			if(conn->fd < 0)
				conn_open(conn);
		}
	}

	for(conn_save = proxy->conn_list; conn_save; conn_save = conn_save->next){
		if(conn_save->type & CONN_TYPE_SAVE){
			/* cycle file */
			
			if((conn_save->max_file_size > 0) && (conn_save->bytes_written >= conn_save->max_file_size)){
				int offset = conn_save->bytes_written;
				char filename[256];
				
				snprintf(filename, sizeof(filename), "%s.0", conn_save->sub_spec);
				
				conn_close(conn_save);
				rename(conn_save->sub_spec, filename);
				conn_open(conn_save);
				
				conn_save->bytes_written = 0;
				
				for(conn = proxy->conn_list; conn; conn = conn->next){
					if((conn->type & CONN_TYPE_OUT) && !(conn->type & CONN_TYPE_SAVE)){
						conn->bytes_written -= offset;
					}
				}
			}
		}
	}

	return 0;
}

static
int signal_callback(int fd, void *data)
{
	struct proxy *proxy = (struct proxy *)data;
	int sig;

	while((sig = daemon_signal_next()) > 0){

		daemon_log(LOG_DEBUG, "received signal %d", sig);
		
		switch (sig) {
		case SIGINT:
		case SIGQUIT:
		{
			struct fd_callback *cb;
			struct connection *conn;
			
			daemon_log(LOG_DEBUG, "marking sockets for close");
			for(cb = g_cb_list; cb; cb = cb->next){
				cb->state = FDCB_CLOSING;
			}
			
			for(conn = proxy->conn_list; conn; conn = conn->next){
				if(conn->tunnel_cmd && conn->tunnel_cmd_pid > 0){
					DEBUG(1) printf("killing tunnel (pid=%d)\n", conn->tunnel_cmd_pid);
					kill(conn->tunnel_cmd_pid, SIGQUIT);
				}
			}
			
			break;
		}
		case SIGUSR1:
			g_debug_level = 5;
			break;
		case SIGUSR2:
			g_debug_level = 0;
			break;
		case SIGCHLD:
		{
			int status;
			struct connection *conn;
			
			for(conn = proxy->conn_list; conn; conn = conn->next){
				if(conn->tunnel_cmd != NULL){
					if(conn->tunnel_cmd_pid != 0){
						DEBUG(2) printf("waiting for %d\n", conn->tunnel_cmd_pid);
						if(waitpid(conn->tunnel_cmd_pid, &status, WNOHANG) == conn->tunnel_cmd_pid){
							/* sometimes we get a SIGCHLD on the ACS, although the command hasn't terminated.
							   We check for that: */
/*							if(WIFEXITED(status)){ */
							/* But, on the TS, WIFEXITED always return false, although the
							   command HAS exited. Sigh. */
							if(1){
								DEBUG(2) printf("%d has exited\n", conn->tunnel_cmd_pid);
								conn->tunnel_cmd_pid = 0;
								DEBUG(5) printf("%s: closing: spec = %s, tunnel_cmd = %s\n",
										__FUNCTION__, conn->spec, conn->tunnel_cmd);
								conn_close(conn);
							}else
								DEBUG(2) printf("WIFEXITED() returned false\n");
						}
					}
				}
			}
			break;
		}
		}
	}
	return 0;
}

static
void conn_setup(struct connection *conn)
{
	char *spec = conn->spec;

	if(conn->type & CONN_TYPE_IN){
		if(strncmp(spec, "pipe:", 5) == 0){
			conn->open = open_read_pipe;
			conn->sub_spec = spec + 5;
		}
#if 0
		else if(strncmp(spec, "tcp:", 4) == 0){
			conn->open = open_read_tcp;
			conn->sub_spec = spec + 4;
		}
#endif
	}else if(conn->type & CONN_TYPE_OUT){
		if(strncmp(spec, "pipe:", 5) == 0){
			conn->open = open_write_pipe;
			conn->sub_spec = spec + 5;
		}
		else if(strncmp(spec, "file:", 5) == 0){
			conn->open = open_write_file;
			conn->sub_spec = spec + 5;
		}
		else if(strncmp(spec, "tcp:", 4) == 0){
			conn->type |= CONN_TYPE_MON;
			conn->open = open_write_tcp;
			conn->sub_spec = spec + 4;
		}
	}
}

static
int setup_cb_write(struct connection *conn)
{
	struct fd_callback *cb = NULL;
	int fd;

	DEBUG(5) printf("%s\n", __FUNCTION__);

	fd = conn->open(conn->sub_spec);
	conn->fd = fd;

	/* open() may have failed, but we do not care for now.
	   If the fd is < 0, we will try to (re)open later. */

	cb = create_fd_callback(fd, 0, write_callback, conn);
	loop_add_callback(cb);
	
	conn->cb = cb;

	return 0;
}

static
int setup_cb_read(struct connection *conn)
{
	struct fd_callback *cb = NULL;
	int fd;

	DEBUG(5) printf("%s\n", __FUNCTION__);

	fd = conn->open(conn->sub_spec);
	conn->fd = fd;

	/* open() may have failed, but we do not care for now.
	   If the fd is < 0, we will try to (re)open later. */

	cb = create_fd_callback(fd, fd >= 0 ? FDCB_READ : 0, read_callback, conn);
	loop_add_callback(cb);
	
	conn->cb = cb;

	return 0;
}

static
int setup_cb_mon(struct connection *conn)
{
	struct fd_callback *cb = NULL;
	int fd;

	DEBUG(5) printf("%s\n", __FUNCTION__);

	fd = conn->fd;

	DEBUG(5) printf("%s, fd = %d\n", __FUNCTION__, fd);

	cb = create_fd_callback(fd, fd >= 0 ? FDCB_READ : 0, mon_callback, conn);
	loop_prepend_callback(cb);
	
	conn->cb_mon = cb;

	return 0;
}

static
int setup_cb_time(struct proxy *proxy)
{
	struct fd_callback *time_cb = NULL;
	int fd;

	DEBUG(5) printf("%s\n", __FUNCTION__);

	time_cb = create_fd_callback(-1, 0,
				     time_callback, proxy);
	loop_add_time_callback(time_cb);
	
	return 0;
}	

int conn_get_saved_counter(struct connection *conn)
{
	int counter = 0;
	char fname[256];
	FILE *fptr;

	snprintf(fname, sizeof(fname), "%s.counter", conn->sub_spec);
	fptr = fopen(fname, "rt");
	if(fptr){
		char buf[256];
		fgets(buf, 256, fptr);
		counter = atoi(buf);
		fclose(fptr);
	}
	return counter;
}

int conn_recover_data(struct connection *conn, const char *save_file, int offset)
{
	int sz, sz_tot = 0;
	int fd;
	char buf[1024];

	/* Later: if offset is negative, we can read from
	   cycled file with the .0 extension. */

	fd = open(save_file, O_RDONLY);
	if(fd >= 0){
		lseek(fd, offset, SEEK_SET);
		while((sz = read(fd, buf, sizeof(buf))) > 0){
			sz_tot += sz;
			databuf_add(conn->databuf, buf, sz);
		}
		close(fd);
	}
	return sz_tot;
}

static
void proxy_setup(struct proxy *proxy)
{
	struct connection *conn, *conn_save;

	for(conn_save = proxy->conn_list; conn_save; conn_save = conn_save->next){
		if(conn_save->type & CONN_TYPE_SAVE){
			for(conn = proxy->conn_list; conn; conn = conn->next){
				if((conn->type & CONN_TYPE_OUT) && !(conn->type & CONN_TYPE_SAVE)){
					int offset;

					offset = conn_get_saved_counter(conn);
					conn_recover_data(conn, conn_save->sub_spec, offset);
				}
			}
		}
	}

	for(conn = proxy->conn_list; conn; conn = conn->next){
		int fd;
		struct fd_callback *cb;
		int fd_type;

		if(conn->tunnel_cmd){
			DEBUG(2) printf("%s calling conn_setup_tunnel()\n", __FUNCTION__);
			conn_setup_tunnel(conn);
		}
		conn_setup(conn);

		if(conn->type & CONN_TYPE_IN)
			setup_cb_read(conn);
		if(conn->type & CONN_TYPE_OUT)
			setup_cb_write(conn);
		if(conn->type & CONN_TYPE_MON)
			setup_cb_mon(conn);
	}

	setup_cb_time(proxy);

	/* signal callback: */
	{
		struct fd_callback *signal_cb = NULL;
		int sig_fd;

		sig_fd = daemon_signal_fd();
		if(sig_fd >= 0){
			signal_cb = create_fd_callback(sig_fd, FDCB_READ, signal_callback, proxy);
			loop_add_callback(signal_cb);
		}
	}
}

static
void proxy_loop(struct proxy *proxy)
{
	struct timeval timeout;

	signal(SIGPIPE, SIG_IGN);

	timeout.tv_sec = proxy->interval;
	timeout.tv_usec = 0;

	loop(&timeout);
}

int read_conf(struct proxy *proxy, const char *conffile)
{
	struct cnfnode *cn_top, *cn_root, *cn_pipe, *cn;
	struct cnfmodule *mod_cyconf;
	struct connection *conn;

	mod_cyconf = find_cnfmodule("cyconf");
	if(!mod_cyconf)
		return -1;

	cn_root = cnfmodule_parse_file(mod_cyconf, conffile);
	if(cn_root){

		for(cn_top = cn_root->first_child; cn_top; cn_top = cn_top->next){
			if(strcmp(cnfnode_getname(cn_top), "global") == 0){
				for(cn = cn_top->first_child; cn; cn = cn->next){
					const char *name = cnfnode_getname(cn);
					const char *value = cnfnode_getval(cn);
					
					if(strcmp(name, "interval:") == 0){
						proxy->interval = atoi(value);
					}
				}
			}else if(strcmp(cnfnode_getname(cn_top), "pipe") == 0){
				for(cn_pipe = cn_top->first_child; cn_pipe; cn_pipe = cn_pipe->next){
					const char *conn_name = cnfnode_getname(cn_pipe);
					const char *spec = NULL;
					const char *tunnel_cmd = NULL;
					const char *alert_cmd = NULL;
					const char *alert_pipe = NULL;
					int type = 0, max_size = 0;
					int count = 1, tunnel_timeout = 15;
					int max_buf = 1000000, tunnel_threshold = 10000;
					int alert_low = 0, alert_med = 0, alert_high = 0;
					
					for(cn = cn_pipe->first_child; cn; cn = cn->next){
						const char *name = cnfnode_getname(cn);
						const char *value = cnfnode_getval(cn);
						
						if(strcmp(name, "spec:") == 0){
							spec = value;
						}else if(strcmp(name, "type:") == 0){
							if(strcmp(value, "write") == 0)
								type |= CONN_TYPE_OUT;
							else if(strcmp(value, "read") == 0)
								type |= CONN_TYPE_IN;
						}else if(strcmp(name, "save:") == 0){
							if(strcmp(value, "yes") == 0){
								type |= CONN_TYPE_SAVE;
							}
						}else if(strcmp(name, "tunnel_cmd:") == 0){
							tunnel_cmd = value;
						}else if(strcmp(name, "tunnel_timeout:") == 0){
							tunnel_timeout = atoi(value);
						}else if(strcmp(name, "tunnel_threshold:") == 0){
							tunnel_threshold = atoi(value);
						}else if(strcmp(name, "max_size:") == 0){
							max_size = atoi(value);
						}else if(strcmp(name, "max_buf:") == 0){
							max_buf = atoi(value);
						}else if(strcmp(name, "count:") == 0){
							count = atoi(value);
						}else if(strcmp(name, "urgent:") == 0){
							if(strcmp(value, "yes") == 0){
								type |= CONN_TYPE_URGENT;
							}
						}else if(strcmp(name, "alert_low:") == 0){
							alert_low = atoi(value);
						}else if(strcmp(name, "alert_med:") == 0){
							alert_med = atoi(value);
						}else if(strcmp(name, "alert_high:") == 0){
							alert_high = atoi(value);
						}else if(strcmp(name, "alert_cmd:") == 0){
							alert_cmd = value;
						}else if(strcmp(name, "alert_pipe:") == 0){
							alert_pipe = value;
						}
					}
					if(spec && (type & (CONN_TYPE_OUT|CONN_TYPE_IN))){
						struct connection *conn;
						int i;

						for(i = 1; i <= count; i++){
							if(count > 1){
								char spec_n[256];

								snprintf(spec_n, sizeof(spec_n), spec, i);
								conn = create_connection(spec_n, type);
							}else
								conn = create_connection(spec, type);

							conn->name = strdup(conn_name);
							proxy_add_connection(proxy, conn);

							conn->max_buf = max_buf;
							conn->max_file_size = max_size;
							if(tunnel_cmd)
								conn->tunnel_cmd = strdup(tunnel_cmd);
							if(alert_cmd)
								conn->alert_cmd = strdup(alert_cmd);
							if(alert_pipe)
								conn->alert_pipe = strdup(alert_pipe);
							conn->tunnel_timeout = tunnel_timeout;
							conn->tunnel_threshold = tunnel_threshold;
							conn->alert_low = alert_low;
							conn->alert_med = alert_med;
							conn->alert_high = alert_high;
						}
					}
				}
			}
		}
		destroy_cnftree(cn_root);
	}
}

int main(int argc, char *argv[])
{
	struct proxy *proxy;
	struct connection *last_conn = NULL;
	char *conffile = SYSCONFDIR"/logpipe.conf";

	daemon_log_ident = daemon_ident_from_argv0(argv[0]);
	daemon_log_use = DAEMON_LOG_SYSLOG;

	register_all();

	proxy = create_proxy();

	while (1){
		int c;

		c = getopt(argc, argv, "c:i:o:d:t:s:");
		if (c == -1)
			break;
		switch(c){
		case 'c':
			conffile = optarg;
			break;
		case 'd':
			g_debug_level = atoi(optarg);
			break;
		case 't':
			if(last_conn) last_conn->tunnel_cmd = strdup(optarg);
			break;
		case 'i':
			proxy_add_connection(proxy, last_conn = create_connection(optarg, CONN_TYPE_IN));
			break;
		case 'o':
			proxy_add_connection(proxy, last_conn = create_connection(optarg, CONN_TYPE_OUT));
			break;
		case 's':
			proxy_add_connection(proxy, last_conn = create_connection(optarg, CONN_TYPE_OUT|CONN_TYPE_SAVE));
			break;
		default:
			printf("unknown option %c\n", c);
			return 1;
		}
	}

	read_conf(proxy, conffile);

	if((argc - optind) == 0){
		/* Initialize signal handling */
		if (daemon_signal_init(SIGINT, SIGQUIT, SIGUSR1, SIGUSR2, SIGCHLD, 0) < 0) {
			daemon_log(LOG_ERR, "could not register signal handlers (%s).", strerror(errno));
			exit(1);
		}

		proxy_setup(proxy);
		proxy_loop(proxy);
	}else{
		fprintf(stderr, "usage: %s <spec_in> <spec_out>\n", argv[0]);
		exit(1);
	}
}
