/*
 *  Module:  modbusaux.c (Modbus/TCP auxiliary functions)
 *
 *  Author:  Marcelo Peccin <marcelo.peccin@cyclades.com>
 *
 *  Copyright:  (c) 2002 Cyclades Corp.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *  Date:  June 13, 2002
 *
 *  Comments:  
 *
 */

#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <asm/frame_mode.h>
#include "modbus.h"

extern int  debug_mode;

static int  fd  = -1;
static int  ffd = -1;

struct sigaction sa;


/*************************************************************/
/**********     Print caught signal to log file     **********/
/*************************************************************/
static void modbus_got_signal(int sig)
{
	if (debug_mode) mpfprint("Got signal # %d !!!\n", sig);

	if (sig == SIGINT) exit(0);

	if (sig == SIGALRM) resend();
}


/***********************************************************/
/**********     Prepare to catch some signals     **********/
/***********************************************************/
void prepare_signals()
{

	bzero((void*)&sa, sizeof(struct sigaction));

	sa.sa_handler = modbus_got_signal;

	sigaction(SIGINT,  &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGHUP,  &sa, NULL);
	sigaction(SIGALRM, &sa, NULL);

}
/***********************************************************/


/******************************************************************/
/*****   Print to ttyS0 (the sole serial port on TS100 !!!)   *****/
/******************************************************************/
void mpprint(const char *fmt, ...)
{
	va_list ap;
	char buf[1024];

	if (fd < 0) {
		if ((fd = open("/dev/ttyS0", O_WRONLY, 0)) <= 0) {
			return;
		}
	}

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	write(fd, buf, strlen(buf));
	va_end(ap);
}
/*******************************************************************/


/*********************************************/
/**********   Print to a log file   **********/
/*********************************************/
void mpfprint(const char *fmt, ...)
{
	va_list ap;
	char buf[1024];

	if (ffd < 0) {
		if ((ffd = open("/root/modbus.log", O_WRONLY | O_CREAT | O_TRUNC, 0)) <= 0) {
			mpprint("mpfprint problems !!!\r\n");
			return;
		}
	}

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	write(ffd, buf, strlen(buf));
	va_end(ap);
}
/**********************************************/


int check_CRC(const uchar *buff, int *len)
{
	ushort crc, calc = 0xFFFF;
	int    i, flen;

	*len -= 2;
	flen = *len;

	crc = ntohs(SWAPcrc(*(ushort*)&buff[*len]));

	while (flen--) {

		calc ^= *buff++;
		
		for (i=0; i < 8; i++) {

			if (calc & 0x0001) {
				
				calc >>= 1;
				calc ^= 0xA001;
				
			} else {
				
				calc >>= 1;
				
			}
		}
	}

	if (crc != calc) return ERROR;

	return OK;
}


void build_CRC(uchar *buff, int *len)
{
	ushort crc = 0xFFFF;
	int    i, flen = *len;

	*len += 2;

	while (flen--) {

		crc ^= *buff++;
		
		for (i=0; i < 8; i++) {

			if (crc & 0x0001) {
				
				crc >>= 1;
				crc ^= 0xA001;
				
			} else {
				
				crc >>= 1;
				
			}
		}
	}

	*(ushort*)buff = htons(SWAPcrc(crc));
}


int  check_LRC(uchar *buff, int *len, const uchar *lrcin)
{
	uchar lrc, calc = 0;
	int   flen = *len;

	*len += HOFFSET;

	ASCIIread(*(ushort*)lrcin, lrc);

	while (flen--) calc += *buff++;

	if ((uchar)(lrc + calc)) return ERROR;

	return OK;
}


void build_LRC(uchar *buff, int *len, uchar *res)
{
	uchar lrc = 0;
	int   flen = *len;

	*len = (flen * 2) + 2;

	while (flen--) lrc += *buff++;

	ASCIIwrite(*(ushort*)res, (0xFF-lrc+1));
}


/*****************************************************************************
 * This routine turn on/off and select the frame-mode to be used on the given
 * fd. The possible values are:
 * 		status: FRAMEMODE_ON or FRAMEMODE_OFF
 * 		mode:	FRAMEMODE_ASCII or FRAMEMODE_RTU
 ****************************************************************************/
int set_frame_mode(int fd, int status, int mode)
{
	struct frame_mode_cfg frame_cfg;

	bzero( (void *) &frame_cfg, sizeof(struct frame_mode_cfg));
	frame_cfg.status = status;
	frame_cfg.mode = mode;
		
	return (ioctl(fd, FRAMEMODESCFG, (void *)&frame_cfg));
}


/*****************************************************************************
 * This routine configures the Serial Port
 ****************************************************************************/
int config_port(int fd, sconf *pconf)
{ 
	struct termios term;
	int aux;

	
	// Getting the port configuration
	if (tcgetattr(fd, &term) == ERROR) {
		return ERROR;
	}

	// Setting media type
	aux = pconf->media | CYC_SET_MEDIA; //[RK]Jul/29/03
	if (ioctl(fd, TIOCSERCYC, &aux) == ERROR) {
		return ERROR;
	}

	// Setting raw mode
	cfmakeraw(&term);

	// Setting speed
	switch (pconf->speed) {
	case 50:
		aux = B50;
		break;
	case 75:
		aux = B75;
		break;
	case 110:
		aux = B110;
		break;
	case 134:
		aux = B134;
		break;
	case 150:
		aux = B150;
		break;
	case 200:
		aux = B200;
		break;
	case 300:
		aux = B300;
		break;
	case 600:
		aux = B600;
		break;
	case 1200:
		aux = B1200;
		break;
	case 1800:
		aux = B1800;
		break;
	case 2400:
		aux = B2400;
		break;
	case 4800:
		aux = B4800;
		break;
	case 14400:
		aux = B14400;
		break;
	case 19200:
		aux = B19200;
		break;
	case 28800:
		aux = B28800;
		break;
	case 38400:
		aux = B38400;
		break;
	case 57600:
		aux = B57600;
		break;
	case 76800:
		aux = B76800;
		break;
	case 115200:
		aux = B115200;
		break;
	case 230400:
		aux = B230400;
		break;
	case 460800:
		aux = B460800;
		break;
	case 921600:
		aux = B921600;
		break;
	default:
		aux = B9600;
	}
	cfsetospeed(&term, aux);
	cfsetispeed(&term, aux);

	// Setting data size
	term.c_cflag &= ~CSIZE;
	switch (pconf->datasize) {
	case 5:
		term.c_cflag |= CS5;
		break;
	case 6:
		term.c_cflag |= CS6;
		break;
	case 7:
		term.c_cflag |= CS7;
		break;
	case 8:
	default:
		term.c_cflag |= CS8;
	}

	// Setting parity
	switch (pconf->parity) {
	case PARITY_ODD:
		term.c_cflag |= (PARENB | PARODD);
		break;
	case PARITY_EVEN:
		term.c_cflag |= (PARENB);
		term.c_cflag &= ~(PARODD);
		break;
	default:
		term.c_cflag &= ~(PARENB); //none
	}

	// Setting number of stop bits
	if (pconf->stopbits == 2) {
		term.c_cflag |= (CSTOPB);
	}
	else {
		term.c_cflag &= ~(CSTOPB);
	}

	// Setting flow control
	switch (pconf->flowctrl) {
	case FLOW_HARD:
		term.c_iflag &= ~(IXON | IXOFF);
		term.c_cflag |= (CRTSCTS);
		break;
	case FLOW_SOFT:
		term.c_iflag |= (IXON | IXOFF);
		term.c_cflag &= ~(CRTSCTS);
		break;
	default:
		term.c_iflag &= ~(IXON | IXOFF); //none
		term.c_cflag &= ~(CRTSCTS);
	}

	// not canonical
	term.c_lflag &= ~(ICANON);

	// Disabling modem signal
	term.c_cflag |= CLOCAL;

	// Enabling modem hang up
	term.c_cflag |= HUPCL;

	// flush the port
	tcflush(fd, TCIOFLUSH);

	// Setting the port configuration
	if (tcsetattr(fd, TCSANOW, &term) == ERROR) {
		return ERROR;
	}

	return OK;
}	


/*****************************************************************************
 * This routine writes to the Serial Port, considering media type
 ****************************************************************************/
/*  int  write_serial(int fd, uchar *sbuff, int count, int media) */
/*  { */
/*  	int sent, aux; */

/*  	if (media) { */
/*  		aux = TIOCM_DTR | TIOCM_RTS; */
/*  		ioctl(fd, TIOCMSET, &aux); */
/*  	} */

/*  	sent = write(fd, sbuff, count); */

/*  	if (media) { */
/*  		alarm(4); */
/*  		tcdrain(fd); */
/*  		alarm(0); */
/*  		aux = TIOCM_DTR; */
/*  		ioctl(fd, TIOCMSET, &aux); */
/*  	} */
	
/*  	return sent; */
/*  } */


/* End of modbusaux.c */
