/*
 *  Module:  modem-replace.c (end to end DCE-like behavior to replace MODEMs)
 *
 *  Author:  Marcelo Peccin <marcelo.peccin@cyclades.com>
 *
 *  Copyright:  (c) 2003 Cyclades Corp.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *  Date:  July 22, 2003
 *
 *  Comments:  The first release of this module is intended to solve the
 *             ISDN Modem substitution using TS family for NSW RTA (Australia).
 *             The project involves video cameras that communicate to 
 *             another equipment in a remote central office.
 */

#include "rot_nports.h"
#include <asm/prxk.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/ioctl.h>
#include "server.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include "modem-replace.h"


// Local Variables
static int     version = 4;
static int     debug = FALSE;
static int     directconn = FALSE;
static int     firstpass = TRUE;
static int     fd = -1;
static int     ffd = -1;
static int     serial_port, txdisp;
static int     fdserial, fdsock = NO_SOCK;
static uchar  *sbuffp;
static int     exitstatus = PROBLEM;

static struct sigaction    sa;
static struct MachineType  mt;

static struct itimerval  almtimeout = {{0,0},{0,0}};


/*************************************************************/
/**********     Print caught signal to log file     **********/
/*************************************************************/
static void  modemreplace_got_signal(int sig)
{
	if (debug && !(sig == SIGALRM))
		mpfprint("Got signal - %s !!!\n\n", strsignal(sig));
	
	switch(sig) {
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		exit(OK);
	case SIGALRM:
		if (txdisp) { //send received data to Socket Connection, if any
			if (tcpsend() == ERROR) {
				mpfprint("error writing to tcp connection: %s (%d)\n", strerror(errno), errno);
				if (close(fdsock) == ERROR) exitstatus = ERRCLOSE;
				if (close(fdserial) == ERROR) exitstatus = ERRCLOSE;
				exit(exitstatus);
			}
			if (debug) mpfprint("TX due %s...\n", strsignal(sig));
		}
	}
}


/***********************************************************/
/**********     Prepare to catch some signals     **********/
/***********************************************************/
static void  prepare_signals()
{

	bzero((void*)&sa, sizeof(struct sigaction));

	sa.sa_handler = modemreplace_got_signal;

	sigaction(SIGINT,  &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGHUP,  &sa, NULL);
	sigaction(SIGALRM, &sa, NULL);

}
/***********************************************************/



/******************************/
int  main(int argc, char **argv)   // MODEM Replacement feature entry point
/******************************/
{
	modemrepconf        config;
	int                 arg, signals;
	uchar               sdevbase[] = "/dev/ttyS";
	int                 fdlsock = NO_SOCK, flags;
	struct sockaddr_in  sock;
	int                 sal = sizeof(sock);
	fd_set              inputs, testfds;
	int                 fdready, event = MODEMREP_NO_EVENT;
	struct timeval      timeout;
	uchar               dest_ip[25];
	uchar               ok_str[] = "OK\r";
	int                 rec, disp = 0, cmd = FALSE;
	uchar               tcpbuff[TCPBUFFSIZE];
    uchar               serialbuff[SBUFFSIZE];


	// get machine type
	get_machine_type(&mt);
	

	// take command line arguments
	if ((argc < 2) || ((config.serial.port = atoi(argv[1])) < 1) || (config.serial.port > mt.nports)) {
		mpfprint("MODEM Replacement must be called with a port (1..%d) as first argument.\n", mt.nports);
		goto end_modemreplace;		
	}
	serial_port = config.serial.port; //serial_port has a bigger scope
	if ((mt.board_type == BOARD_TS1H) || (mt.board_type == BOARD_TS110)) {
		// TS100/110 has its own zImage...
		config.serial.port = 0; //the shared console and serial port
		sprintf(config.serial.dev, "%s%d", sdevbase, config.serial.port);
	} else {
		sprintf(config.serial.dev, "%s%d", sdevbase, config.serial.port);
		config.serial.port--;
	}
	for (arg = 2; arg < argc; arg++) {
		if (!strcmp(argv[arg], "-debug")) {
			debug = TRUE;
  		} else if (!strcmp(argv[arg], "-connect")) {
			directconn = TRUE;
			if (++arg == argc) {
				mpfprint("-connect option must be followed by an IP number as next argument\n");
				goto end_modemreplace;
			}
			strcpy(dest_ip, argv[arg]);
		}
	}	
	if (debug)
		mpfprint("\nMODEM Replacement running for serial port #%02d (v%d)...\n", serial_port, version);


    // initialize Shared Memory
	if (shm_init(-1) < 0)
		goto end_modemreplace;
	while (shm_get_set_val(0, &cy_shm->st_cy_ras, 0, CY_SEM_CONFIG) == 0)
		sleep(1);


	// prepare to catch some Signals and sbufp
	prepare_signals();
	sbuffp = serialbuff;


	// copy Configuration Information
	config.serial.speed    = cy_shm->line_conf[config.serial.port].speed; 
	config.serial.datasize = cy_shm->line_conf[config.serial.port].DataSize;
	config.serial.parity   = cy_shm->line_conf[config.serial.port].Parity;
	config.serial.stopbits = cy_shm->line_conf[config.serial.port].StopBits;
	config.serial.flowctrl = cy_shm->line_conf[config.serial.port].flow;
	config.serial.media	   = cy_shm->line_conf[config.serial.port].media;
	config.socket_ip	   = cy_shm->line_conf[config.serial.port].ipno;
	config.tx_interval	   = cy_shm->line_conf[config.serial.port].tx_interval;

	// info printing
	if (debug)
	    mpfprint("\
config.serial.speed    = %d\n\
config.serial.datasize = %d\n\
config.serial.parity   = %d (1,odd,even)\n\
config.serial.stopbits = %d\n\
config.serial.flowctrl = %d (0,Hw,Sw)\n\
config.serial.media    = %d (232=0)\n\
config.tx_interval     = %d ms\n",
				 config.serial.speed,
				 config.serial.datasize,
				 config.serial.parity,
				 config.serial.stopbits,
				 config.serial.flowctrl,
                 config.serial.media,
                 config.tx_interval);


    // open and configure Serial Port
	if ((fdserial = open(config.serial.dev, O_RDWR | O_NOCTTY | O_NONBLOCK)) == ERROR) {
		mpfprint("error opening serial port: %s (%d)\n", strerror(errno), errno);
		goto end_modemreplace;
	}
	if (config_port(fdserial, &config.serial) == ERROR) {
		mpfprint("problems configuring serial port\n");
		goto end_closeserial;
	}
	// drop DTR (DCD to the camera side)
	ioctl(fdserial,TIOCMGET,&signals);
	signals &= ~TIOCM_DTR;
	ioctl(fdserial,TIOCMSET,&signals);
	ioctl(fdserial,TIOCMGET,&signals); //not sure if "signals" is preserved

	if (directconn)
		mpfprint("...serial port is ready.\n");


 // Main Loop
  	while (1) {

		int flag = FALSE;
		while (!(signals & TIOCM_DSR)) { //serial cable isn't plugged
			if (debug && !flag) {
				mpfprint("waiting for DSR...\n");
				flag = TRUE;
			}
			sleep(1);
			ioctl(fdserial,TIOCMGET,&signals); //read DSR
		}

		if (!directconn) {
			if (debug && !firstpass)
				mpfprint("\nentering Command Mode\n");

			// open Listen Socket
			if ((fdlsock = socket(AF_INET, SOCK_STREAM, 0)) == ERROR) {
				mpfprint("error creating listen socket: %s (%d)\n", strerror(errno), errno);
				goto end_closeserial;
			}
			// must be non blocking
			flags = fcntl(fdlsock, F_GETFL);
			fcntl(fdlsock, F_SETFL, flags | O_NONBLOCK);
			
			bzero(&sock, sizeof(sock));
			sock.sin_family = AF_INET;
			sock.sin_addr.s_addr = config.socket_ip;
			sock.sin_port = htonl(MODEMREP_PORT);
			
			// finish info printing
			if (debug && firstpass)
				mpfprint("config.socket_ip       = %s (htoa)\n", inet_ntoa(sock.sin_addr));
			
			// "name" Listen Socket
			flag = FALSE;
			do {
				if (bind(fdlsock, (struct sockaddr*)&sock, sal) == ERROR) {
					if (errno != EADDRINUSE) {
						mpfprint("error binding listen socket: %s (%d)\n", strerror(errno), errno);
						goto end_closelsocket;
					}
					if (debug && !flag) {
						mpfprint("waiting socket to be available...\n");
						flag = TRUE;
					}
					sleep(1);
				}
				else
					break;
			} while (1);
			
			// put Listen Socket to listen
			if (listen(fdlsock, 1) == ERROR) {
				mpfprint("error putting listen socket to listen: %s (%d)\n", strerror(errno), errno);
				goto end_closelsocket;
			}
		
			if (debug) {
				if (firstpass) {
					mpfprint("...serial port and listen socket are ready.\n\n");
					firstpass = FALSE;
				} else {
					mpfprint("listen socket ready (%s)\n", inet_ntoa(sock.sin_addr));
				}
			}

		} else {

			if (debug) mpfprint("\n");

		}

		FD_ZERO(&inputs);
		FD_SET(fdserial, &inputs);

	    // wait for an EVENT to happen
		while (!event && !directconn && (signals & TIOCM_DSR)) {
			                            //serial port plugged
			testfds = inputs;
			timeout.tv_sec = 0;
			timeout.tv_usec = 30 * 1000;  // 30ms

			if ((fdready = select(fdserial+1, &testfds, NULL, NULL, &timeout)) == ERROR) {
				mpfprint("error in select: %s (%d)\n", strerror(errno), errno);
				goto end_closelsocket;
			}

			if (fdready) { //command received

				int i;
				
				if ((rec = read(fdserial, (serialbuff + disp), (SBUFFSIZE - disp))) == ERROR) {
					mpfprint("error reading serial port: %s (%d)\n", strerror(errno), errno);
					goto end_closelsocket;
				}
				for (i=0; i < rec; i++) {
					if (serialbuff[disp+i] == CHAR_CR) {
						rec = disp + i + 1;
						cmd = TRUE;
						break;
					}
				}
				if (cmd) {
					if (debug) {
						mpfprint("command mode: %d chars received from serial port\n", rec);
						for (i=0; i < rec; i++) {
							if (serialbuff[i] != CHAR_CR)
								mpfprint("%c", serialbuff[i]);
							else
								mpfprint("<CR>");
						}
						mpfprint("\n");
					}

					// look for "ATDT[ip]<CR>" command
					if (!strncmp(serialbuff, "ATDT", 4)) {
						
                        // save IP
						serialbuff[rec-1] = '\0';
						for (i=4; serialbuff[i] == ' '; i++);
						if ((rec - i) < 8) {
							if (debug)
								mpfprint("received ATDT: IP is too short\n");
						} else {
							strcpy(dest_ip, serialbuff + i);
							if (debug)
								mpfprint("received ATDT command (IP = %s)\n", dest_ip);
							event = SER_CONN_REQUEST;
						}
					
					}
				
					//send OK for this command
					if (ssend(fdserial, ok_str, sizeof(ok_str)) == ERROR) {
						mpfprint("error writing to serial port: %s (%d)\n", strerror(errno), errno);
						goto end_closelsocket;
					}
					if (debug)
						mpfprint("answered with \"OK<CR>\"\n");

					disp = 0; cmd = FALSE;
			
					if (event) break;

				} else {

					disp += rec;
					if (disp > MAX_CMD_LENGTH) {
						disp = 0;
						if (debug)
							mpfprint("command mode: command too long\n");
					}
						
				}
				
			} else { //timeout
				
				if ((fdsock = accept(fdlsock, (struct sockaddr*)&sock, &sal)) == ERROR) {
					if (errno != EAGAIN) {
						mpfprint("error accepting socket connection: %s (%d)\n", strerror(errno), errno);
						goto end_closelsocket;
					}
				} else { //remote socket connection accepted
					event = SOCK_CONN_ACCEPT;
					if (debug)
						mpfprint("socket connection accepted   >>>>>>>>>>\n");
					break;
				}
			}
			
			ioctl(fdserial,TIOCMGET,&signals); //read DSR

		} /* while (!event) */

		if (!directconn) {
			if (close(fdlsock) == ERROR) { //stop listen
				exitstatus = ERRCLOSE;
				if (fdsock != NO_SOCK) close(fdsock);
				goto end_closeserial;
			}
			fdlsock = NO_SOCK;
		}

		if (!(signals & TIOCM_DSR)) { //cable was unplugged
			if (debug)
				mpfprint("detected DSR down\n");
			if (fdsock != NO_SOCK) {
				if (close(fdsock) == ERROR) {
					exitstatus = ERRCLOSE;
					goto end_closeserial;
				}
				fdsock = NO_SOCK;
			}
			continue;
		}

		if ((event == SER_CONN_REQUEST) || (directconn)) {
            //serial side asked for socket connection or option -connect was issued
			
			if ((fdsock = socket(AF_INET, SOCK_STREAM, 0)) == ERROR) {
				mpfprint("error opening socket connection: %s (%d)\n", strerror(errno), errno);
				goto end_closeserial;
			}

			bzero(&sock, sizeof(sock));

			if (!inet_aton(dest_ip, &sock.sin_addr)) {
				mpfprint("problems setting remote host\n");
				if (directconn)
					goto end_closesocket;
				if (close(fdsock) == ERROR) {
					exitstatus = ERRCLOSE;
					goto end_closeserial;
				}
				fdsock = NO_SOCK;
				event = MODEMREP_NO_EVENT; //prepare to next pass
				sleep(3/*sec*/); //before recreating the Listen Socket

				// SHOULDN'T SEND AN ERROR MESSAGE TO THE CALLER?

				continue;
			}

			sock.sin_family = AF_INET;
			sock.sin_port = htonl(MODEMREP_PORT);

			if (connect(fdsock,(struct sockaddr*)&sock, sizeof(sock)) == ERROR) {
				mpfprint("error connecting to remote host: %s (%d)\n", strerror(errno), errno);
				if (directconn)
					goto end_closesocket;
				if (close(fdsock) == ERROR) {
					exitstatus = ERRCLOSE;
					goto end_closeserial;
				}
				fdsock = NO_SOCK;
				event = MODEMREP_NO_EVENT; //prepare to next pass
				sleep(3/*sec*/); //before recreating the Listen Socket

				// SHOULDN'T SEND AN ERROR MESSAGE TO THE CALLER?

				continue;
			}

			if (debug)
				mpfprint("socket connection established   >>>>>>>>>>\n");

            // MAY SEND "CONNECT<CR>" TO (CALLING) SERIAL PORT

		} else { //remote socket connection was accepted

            // MAY SEND "RING<CR>" AND "CONNECT<CR>" TO (CALLED) SERIAL PORT

		}

		event = MODEMREP_NO_EVENT; //prepare to next pass

		// assert DTR (DCD to the camera side)
		ioctl(fdserial,TIOCMGET,&signals);
		signals |= TIOCM_DTR;
		ioctl(fdserial,TIOCMSET,&signals);
		ioctl(fdserial,TIOCMGET,&signals); //not sure if "signals" is preserved

		// start timer
		txdisp = 0;
		if (config.tx_interval) {
			almtimeout.it_value.tv_usec = config.tx_interval * 1000;
			setitimer(ITIMER_REAL, &almtimeout, NULL);
		}


		FD_SET(fdsock, &inputs);
		/*****************************/
		/* Receive and Transmit data */
		/*****************************/
		while (signals & TIOCM_DSR) {

			testfds = inputs;
			timeout.tv_sec = 0;
			timeout.tv_usec = config.tx_interval * 1000;  // us
			if (!timeout.tv_usec)
				timeout.tv_usec = 30 * 1000;  // 30ms

			if ((fdready = select(Max(fdserial, fdsock) + 1, &testfds, NULL, NULL, &timeout)) == ERROR) {
				if (errno == EINTR) { //Interrupted system call (due to max tx_interval)
					fdready = 0;
				} else {
					mpfprint("error in select: %s (%d)\n", strerror(errno), errno);
					goto end_closesocket;
				}
			}

			if (fdready) {

				/*****************************/
				if (FD_ISSET(fdsock, &testfds)) {   // Activity on ethernet port
				/*****************************/
			
					// read data
					if ((rec = recv(fdsock, tcpbuff, TCPBUFFSIZE, 0)) == ERROR) {
						mpfprint("error reading socket: %s (%d)\n", strerror(errno), errno);
						goto end_closesocket;
					}

					if (!rec) { //remote socket has closed
						if (debug)
							mpfprint("connection reset by peer   <<<<<<<<<<\n");
						break;
					}

					// received data is ready to be sent to Serial Port
					if (ssend(fdserial, tcpbuff, rec) == ERROR) {
						mpfprint("error writing to serial port: %s (%d)\n", strerror(errno), errno);
						goto end_closesocket;
					}

				}

				/*******************************/
				if (FD_ISSET(fdserial, &testfds)) {   // Activity on serial port
				/*******************************/

					// read data
					if ((rec = read(fdserial, (serialbuff + txdisp), (SBUFFSIZE - txdisp))) == ERROR) {
						mpfprint("error reading serial port: %s (%d)\n", strerror(errno), errno);
						goto end_closesocket;
					}

					// check buffer limit and immediate TX
					if (((txdisp += rec) >= SBUFFLIMIT) || !config.tx_interval){

						// send received data to Socket Connection
						if (tcpsend() == ERROR) {
							mpfprint("error writing to tcp connection: %s (%d)\n", strerror(errno), errno);
							goto end_closesocket;
						}

					}

				}
			
			} else { //timeout

				if (txdisp) { //send received data to Socket Connection, if any
					if (tcpsend() == ERROR) {
						mpfprint("error writing to tcp connection: %s (%d)\n", strerror(errno), errno);
						goto end_closesocket;
					}

				}

			}

			ioctl(fdserial,TIOCMGET,&signals); //read DSR
			if (!(signals & TIOCM_DSR)) {
				
				//WOULDN'T IT BE GOOD SEE DSR/DCD DOWN FOR MORE THAN XXXms FIRST?
				
				if (debug)
					mpfprint("detected DSR down (DTR from device)   <<<<<<<<<<\n");
			}

		}


		// stop timer
		if (config.tx_interval) {
			almtimeout.it_value.tv_usec = 0;
			setitimer(ITIMER_REAL, &almtimeout, NULL);
		}

		// tear down connection
		if (debug)
			mpfprint("tearing down TCP connection\n");
		if (close(fdsock) == ERROR) {
			exitstatus = ERRCLOSE;
			goto end_closeserial;
		}
		fdsock = NO_SOCK;

		// drop DTR (DCD to the camera side)
		if (debug)
			mpfprint("dropping our DTR (DCD to device)\n");
		ioctl(fdserial,TIOCMGET,&signals);
		signals &= ~TIOCM_DTR;
		ioctl(fdserial,TIOCMSET,&signals);


		sleep(3/*sec*/); //before recreating the Listen Socket or trying a new connection


	} /* Main Loop */


end_closesocket:
	if (close(fdsock) == ERROR) exitstatus = ERRCLOSE;
	goto end_closeserial;

end_closelsocket:
	if (close(fdlsock) == ERROR) exitstatus = ERRCLOSE;

end_closeserial:
	if (close(fdserial) == ERROR) exitstatus = ERRCLOSE;

end_modemreplace:
	if (debug) mpfprint("................................... exit status: %d\n\n", exitstatus);


	exit(exitstatus);
	

} /* main() */



/*****************************************************************/
/**********   This routine configures the Serial Port   **********/
/*****************************************************************/
static int  config_port(int fd, sconf *pconf)
{ 
	struct termios  term;
	int             aux;

	
	// Getting the port configuration
	if (tcgetattr(fd, &term) == ERROR) {
		return ERROR;
	}

	// Setting media type
	if ((mt.board_type == BOARD_TS1H) || (mt.board_type == BOARD_TS110)) {
		aux = pconf->media;
		if (ioctl(fd, TIOCSERCONFIG, &aux) == ERROR) {
			return ERROR;
		}
	}

	// Setting raw mode
	cfmakeraw(&term);

	// Setting speed
	switch (pconf->speed) {
	case 50:
		aux = B50;
		break;
	case 75:
		aux = B75;
		break;
	case 110:
		aux = B110;
		break;
	case 134:
		aux = B134;
		break;
	case 150:
		aux = B150;
		break;
	case 200:
		aux = B200;
		break;
	case 300:
		aux = B300;
		break;
	case 600:
		aux = B600;
		break;
	case 1200:
		aux = B1200;
		break;
	case 1800:
		aux = B1800;
		break;
	case 2400:
		aux = B2400;
		break;
	case 4800:
		aux = B4800;
		break;
	case 14400:
		aux = B14400;
		break;
	case 19200:
		aux = B19200;
		break;
	case 28800:
		aux = B28800;
		break;
	case 38400:
		aux = B38400;
		break;
	case 57600:
		aux = B57600;
		break;
	case 76800:
		aux = B76800;
		break;
	case 115200:
		aux = B115200;
		break;
	case 230400:
		aux = B230400;
		break;
	case 460800:
		aux = B460800;
		break;
	default:
		aux = B9600;
	}
	cfsetospeed(&term, aux);
	cfsetispeed(&term, aux);

	// Setting data size
	term.c_cflag &= ~CSIZE;
	switch (pconf->datasize) {
	case 5:
		term.c_cflag |= CS5;
		break;
	case 6:
		term.c_cflag |= CS6;
		break;
	case 7:
		term.c_cflag |= CS7;
		break;
	case 8:
	default:
		term.c_cflag |= CS8;
	}

	// Setting parity
	switch (pconf->parity) {
	case PARITY_ODD:
		term.c_cflag |= (PARENB | PARODD);
		break;
	case PARITY_EVEN:
		term.c_cflag |= (PARENB);
		term.c_cflag &= ~(PARODD);
		break;
	default:
		term.c_cflag &= ~(PARENB); //none
	}

	// Setting number of stop bits
	if (pconf->stopbits == 2)
		term.c_cflag |= (CSTOPB);
	else
		term.c_cflag &= ~(CSTOPB);

	// Setting flow control
	switch (pconf->flowctrl) {
	case FLOW_HARD:
		term.c_iflag &= ~(IXON | IXOFF);
		term.c_cflag |= (CRTSCTS);
		break;
	case FLOW_SOFT:
		term.c_iflag |= (IXON | IXOFF);
		term.c_cflag &= ~(CRTSCTS);
		break;
	default:
		term.c_iflag &= ~(IXON | IXOFF); //none
		term.c_cflag &= ~(CRTSCTS);
	}

	// Not canonical
	term.c_lflag &= ~(ICANON);

	// Disabling modem signal
	term.c_cflag |= CLOCAL;

	// Enabling modem hang up
	term.c_cflag |= HUPCL;

	// Flush the port
	tcflush(fd, TCIOFLUSH);

	// Setting the port configuration
	if (tcsetattr(fd, TCSANOW, &term) == ERROR) {
		return ERROR;
	}

	return OK;
}


/****************************************/
/**********   Print to ttyS0   **********/
/****************************************/
static void  mpprint(const char *fmt, ...)
{
	va_list   ap;
	uchar     buf[1024];

	if (fd < 0) {
		if ((fd = open("/dev/ttyS0", O_WRONLY, 0)) <= 0) {
			return;
		}
	}

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	write(fd, buf, strlen(buf));
	va_end(ap);
}


/*********************************************/
/**********   Print to a log file   **********/
/*********************************************/
static void  mpfprint(const char *fmt, ...)
{
	va_list   ap;
	uchar     buf[1024];
	uchar     filen[30];

	if (serial_port)
		sprintf(filen, "/root/modem-replace-%02d.log", serial_port);
	else
		sprintf(filen, "/root/modem-replace.log");

	if (ffd < 0) {
		if ((ffd = open(filen, O_WRONLY | O_CREAT | O_TRUNC, 0)) <= 0) {
			mpprint("error opening mpfprint default file: %s !!!\r\n", strerror(errno));
			return;
		}
	}

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	write(ffd, buf, strlen(buf));
	va_end(ap);
}


/******************************************************/
/**********   Send data to the Serial Port   **********/
/******************************************************/
static int  ssend(int fds, uchar *buff, int tosend)
{
	int disp = 0, sent;

	do {
		if ((sent = write(fds, (buff + disp), (tosend - disp))) >= 0)
			disp += sent;
		else
			if (errno != EAGAIN)
				break;
		if (disp != tosend)
			usleep(1000);  // 1ms
	} while (disp < tosend);

	if (sent < 0)
		return ERROR;
	else
		return OK;

}


/*************************************************************/
/**********   Send chars to the Socket Connection   **********/
/*************************************************************/
static int  tcpsend()
{
	int disp = 0, sent;

	do {
		if ((sent = send(fdsock, (sbuffp + disp), (txdisp - disp), 0)) >= 0)
			disp += sent; 
		else
			if (errno != EAGAIN)
				break;
		if (disp != txdisp)
			usleep(100);
	} while (disp < txdisp);

	txdisp = 0;

	if (sent < 0)
		return ERROR;
	else
		return OK;

}


/* End of modem-replace.c */
