#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <termios.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/cy_sm.h>

/* access type */
#define READ_DEV 			0x0001	/* read access */
#define WRITE_DEV 			0x0002	/* write access */
#define ACCESS_TYPE_MASK	0x000f	/* mask */
#define GPIO_ACCESS			0x0010	/* /dev/gpio device */
#define ADC1_ACCESS			0x0020	/* /dev/adc1 device */
#define ADC2_ACCESS			0x0040	/* /dev/adc2 device */
#define DEVICE_TYPE_MASK	0x00f0	/* length type access mask */
#define LIST_DIG_PORT		0x0100	/* list digital port status */
#define SINGLE_DIG_PORT		0x0200	/* list single digital port status */
#define DEC_OUTPUT			0x0400	/* list single digital port status */
#define TEMP_OUTPUT			0x0800	/* display temperature */
#define INTERVAL			0x1000	/* interval to sample */

/* devices supported */
#define DEV_GPIO		"gpio"
#define DEV_ADC1		"adc1"
#define DEV_ADC2		"adc2"

#if 0
static void msleep(uint t) 
{
	uint i;
	/* sleep t miliseconds */
	for (i=0; i<t; i++) {
		usleep(1000);
	}
}
#endif

void unblock(int sig)
{
	sigset_t set;

	sigemptyset(&set);
	sigaddset(&set, sig);
	sigprocmask(SIG_UNBLOCK, &set, NULL);
}

#if 0
static void term_handler(int sig)
{
	got_sig_term = 1;
	signal(sig, SIG_IGN);
}
#endif

void usage(char *appl, char *extra)
{
    if (extra != 0) {
		printf("%s\n", extra);
    }
    printf("Usage:\n");
    printf("  %s -<r|[w data]> [-d device] [-p digital_port] -l -o -h \n", appl);
    printf("   Options:\n");
    printf("    -r               read from device.\n");
    printf("    -w data          write data to device.\n");
    printf("    -d device        defines device to use.\n");
    printf("    -p digital_port  defines digital port [1-8] to use.\n");
    printf("    -l               list digital port status.\n");
    printf("    -o               output in decimal format.\n");
    printf("    -h               for help.\n");
}

int main(int argc, char **argv)
{
	int fw, arg, action = 0;
	char fio[32] = {0}, ch[32] = {0}, i=0, port=0;
    extern char *optarg;

	/* Disconnect from the controlling tty. */ 
    if ((fw = open("/dev/tty", O_RDWR)) >= 0) {
        signal(SIGHUP, SIG_IGN);
        ioctl(fw, TIOCNOTTY, (char *)1);
        signal(SIGHUP, SIG_DFL);
        close(fw);
    }

	/*
	 *    Check the syntax.
	 */

    while ((arg = getopt(argc, argv, "d:w:p:i:rhlot")) > 0) {
		switch (arg) {
		    case 'd':
				strcpy(fio, optarg);
				if (strstr(fio, DEV_GPIO)) { 
					/* command for device /dev/gpio */
					action |= GPIO_ACCESS;
					break;
				}
				if (strstr(fio, DEV_ADC1)) { 
					/* command for device /dev/adc1 */
					action |= ADC1_ACCESS;
					break;
				}
				if (strstr(fio, DEV_ADC2)) { 
					/* command for device /dev/adc2 */
					action |= ADC2_ACCESS;
				}
				break;
		    case 'h':
				usage(argv[0], 0);
				exit(1);
				break;
			case 'l':
				action |= LIST_DIG_PORT;
				break;
			case 'o':
				action |= DEC_OUTPUT;
				break;
		    case 'p':
				port = *optarg - '1';
				if ((port >= 0) && (port < 8)) { 
					/* port range 0-7 */
					action |= SINGLE_DIG_PORT;
				}
				break;
		    case 'r':
				action |= READ_DEV;
				break;
		    case 'w':
				action |= WRITE_DEV;
				strcpy(&ch[0], optarg);
				i = strlen(ch);
				break;
		    default:
				usage(argv[0], "Syntax error");
				exit(1);
				break;
		}
	}

    if (!(action & ACCESS_TYPE_MASK) || !fio[0]) {
    	/* no action or no device was informed */
		printf(	"No action. Use '%s -h' for help\r\n", argv[0]);
		exit(1);
    }

	if ((fw = open(fio, O_RDWR|O_NONBLOCK)) < 0) {
		printf(	"%s : ERROR can't open %s\n", argv[0], fio);
		return -1;
	}

	switch (action & ACCESS_TYPE_MASK) {
		case READ_DEV:
			switch (action & DEVICE_TYPE_MASK) {
				case GPIO_ACCESS:
					/* read all 8 digital ports */
					read(fw, &ch[0], 1);
					if (action & SINGLE_DIG_PORT) {
						if (action & LIST_DIG_PORT) {
							printf("Digital Port Status:\r\n");
							printf("	port %d - ", (port+1));
							if (ch[0] & (1<<port)) {
								printf("closed\r\n");
							}
							else {
								printf("open\r\n");
							}
						}
						else {
							if (ch[0] & (1<<port)) {
								printf("1\r\n");
							}
							else {
								printf("0\r\n");
							}
						}	
						break;
					}
					if (action & LIST_DIG_PORT) {
						printf("Digital Port Status: %02x\r\n", ch[0]);
						for (port=0; port<8; port++) {
							printf("	port %d - ", (port+1));
							if (ch[0] & (1<<port)) {
								printf("closed\r\n");
							}
							else {
								printf("open\r\n");
							}
						}
					}
					else {
						printf("%02x\r\n", ch[0]);
					}
					break;
				case ADC1_ACCESS:
					/* read analog port 1 */
					//ch[0] = ADC1_I2C_ADDR;
					//ch[1] = ADC_CONF_REG;
					//write(fw, &ch[0], 2);
					//ch[0] = ADC1_I2C_ADDR;
					//msleep(100);
					read(fw, &ch[0], 3); 		/* 16-bit output reg + 8-bit config reg */
					if (action & DEC_OUTPUT) {
						printf("%d\r\n", (int)((ch[0]<<8)|ch[1]));	/* print just output reg */
					}
					else {
						printf("%02x%02x\r\n", ch[0],ch[1]);	/* print just output reg */
					}
					break;
				case ADC2_ACCESS:
					/* read analog port 2 */
					//ch[0] = ADC2_I2C_ADDR;
					//ch[1] = ADC_CONF_REG;
					//write(fw, &ch[0], 2);
					//ch[0] = ADC2_I2C_ADDR;
					//msleep(100);
					read(fw, &ch[0], 3); 		/* 16-bit output reg + 8-bit config reg */
					if (action & DEC_OUTPUT) {
						printf("%d\r\n", (int)((ch[0]<<8)|ch[1]));	/* print just output reg */
					}
					else {
						printf("%02x%02x\r\n", ch[0],ch[1]);	/* print just output reg */
					}
					break;
			}				
			break;
		case WRITE_DEV:
			if (i) {
				write(fw, &ch[0], i);
			}
			break;
	}

	close(fw);
	return(0);
}

