/*
 *  Module:  puconnect.c (TCP remote connection for digital inputs)
 *
 *  Authors:  Marcelo Peccin <marcelo.peccin@cyclades.com>
 *            Sulamita Garcia <sulamita.garcia@cyclades.com>
 *
 *  Copyright:  (c) 2004 Cyclades Corp.
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *  Date:  October 7, 2004
 *
 *  Comments:  The first release of this module is intended to the New York
 *             Fire Department. They need to communicate over the network
 *             when the monitored devices close a "dry contact".
 *             This will be achieved by raising the DTR signal on a remote
 *             RS232 serial interface (on a remote console server).
 *
 *  Author:        Marcelo Peccin <marcelo.peccin@cyclades.com>
 *  Date:          December 28, 2004
 *  Enhancements:  Sockets now have the "keep alive" option and a select() loop
 *                 to detect socket related exceptions. To quickly detect network
 *                 related problems, a "peer monitoring" mechanism was implemented,
 *                 as well as a "recovery mechanism" to detect when connectivity
 *                 is reestablished. Also, both mechanisms described above are
 *                 optimized to allow resynchronization and to avoid duplication
 *                 when the destination server is the same for many inputs.
 */

#include "rot_nports.h"
#include <asm/prxk.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include "server.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include "puconnect.h"

extern char *optarg;

// Local Variables
static int    version = 3;
static int exitstatus = ERROR;
static int      debug = FALSE;
static int terminator = FALSE;
static int        ffd = -1;
static int         fw = -1;

static int               fdsock[9]; //position 0 is not used
static struct sockaddr_in  sock[9]; //position 0 is not used
static int          remote_conn[9]; //position 0 is not used
static int          recover_pid[9]; //position 0 is not used
static int          peerMon_pid[9]; //position 0 is not used

static puconf  config;

static struct timeval  debounce_start[9]; //position 0 is not used
static int	               debounce = 75; //miliseconds

static struct sigaction    sa;
static struct MachineType  mt;


/***************************************************************************/
/**********     Print caught signal to log file and terminate     **********/
/***************************************************************************/
static void  puconn_got_signal(int sig)
{
	if (debug && !terminator && (sig != SIGCHLD))
		mpfprint("Got signal - %s (pid:%d) !!!\n\n", strsignal(sig), getpid());
	
	if (terminator)
		printf("All puconnect instances were terminated.\n");
	
	switch(sig) {
	case SIGCHLD:
		wait(NULL); //no zombies...
		break;
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		exit(OK);
	}
}


/***********************************************************/
/**********     Prepare to catch some signals     **********/
/***********************************************************/
static void  prepare_signals()
{

	bzero((void*)&sa, sizeof(struct sigaction));

	sa.sa_handler = puconn_got_signal;

	sigaction(SIGCHLD, &sa, NULL);
	sigaction(SIGINT,  &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);
	sigaction(SIGHUP,  &sa, NULL);

}


/***************************************************************/
/**********     Control contacts deboucing period     **********/
/***************************************************************/
static int  debounce_complete(int input)
{
	struct timeval timenow;

	gettimeofday(&timenow, NULL);

	if (((timenow.tv_sec - debounce_start[input].tv_sec) * 1000000 +
		 timenow.tv_usec - debounce_start[input].tv_usec) / 1000 >= debounce)

		return TRUE;
	else
		return FALSE;

}


/*************************************************/
/**********     Change process name     **********/
/*************************************************/
void changeTitle(int argc, char **argv, char *fmt, ...)
{
	int i;
	va_list ap;

	if (!argv) goto exit_rot;

	va_start(ap, fmt);
	for (i=1; i<argc; i++)
		memset(argv[i], '\0', strlen(argv[i])+1);
	vsnprintf(argv[0], strlen(argv[0])+1, fmt, ap);

exit_rot:
	va_end(ap);
}


/*******************************************************************/
/*****************       puconnect Main Loop       *****************/
/*******************************************************************/
/*
 *    Here we read all digital inputs (after debouncing contacts) and
 *    maintain the corresponding remote TCP connections as needed,
 *    according to information read from "/etc/puconnect.conf" file.
 */
void  puconn_mainloop(int argc, char **argv)
{
	uchar di;
	int i, j, maxsock, flags, fdready;
//	int reg, lastreg = 0, corr = 0;
//	struct timeval last = {0,0}, timenow;
	fd_set          socks;
	struct timeval  timeout;
	char            recbuff[50];

	//finish info printing
	if (debug)
		mpfprint("pid: %d\n\n", getpid()); 

	setsid();
	//now this process is the leader of a new session, the process group leader of the
	//new process group, and has no controlling tty

	while (1) {

		FD_ZERO(&socks);
		maxsock = 0;

        //prepare socket and select vectors 
		for (i=1; i<=8; i++) {
			if (fdsock[i] == NO_SOCK) {
				int aliveON = 1;
				if ((fdsock[i] = socket(AF_INET, SOCK_STREAM, 0)) == ERROR) {
					mpfprint("error creating socket for input %d: %s (%d)\n", i, strerror(errno), errno);
					return;
				}
				// prepare to monitor network connection
				if (setsockopt(fdsock[i], SOL_SOCKET, SO_KEEPALIVE, &aliveON, sizeof(int)) == ERROR) {
					mpfprint("error seting socket options for input %d: %s (%d)\n", i, strerror(errno), errno);
					return;
				}
				// must be non-blocking
				if (((flags = fcntl(fdsock[i], F_GETFL)) == ERROR) || 
					(fcntl(fdsock[i], F_SETFL, flags | O_NONBLOCK) == ERROR)) {
					mpfprint("error seting non-blocking socket for input %d: %s (%d)\n", i, strerror(errno), errno);
					return;
				}
				bzero(&sock[i], sizeof(struct sockaddr_in));
				sock[i].sin_family = AF_INET;
				sock[i].sin_addr = config.destip[i];
				sock[i].sin_port = config.destport[i];
			} else  if (remote_conn[i] == PUCONN_ON) {
				FD_SET(fdsock[i], &socks);
				if (fdsock[i] > maxsock) maxsock = fdsock[i];
			}
		}

		timeout.tv_sec = 0;
		timeout.tv_usec = 95 * 1000;  //with this we have approx. 100ms between reads for each port

		if ((fdready = select(maxsock+1, &socks, NULL, NULL, &timeout)) == ERROR) {
			if (errno != EINTR) {
				mpfprint("error in select: %s (%d)\n", strerror(errno), errno);
				return;
			}
		}
		if (fdready) { //we must have problems with at least one connection
			for (i=1; i<=8; i++) {
				if (FD_ISSET(fdsock[i], &socks)) {
					int rec;
					if ((rec = recv(fdsock[i], recbuff, 50, 0)) <= 0) {
						if (!rec || (errno != EAGAIN)) {
							if (debug)
								mpfprint("error on socket for input %d: %s (%d)\n", i, strerror(errno), errno);
							remote_conn[i] = NET_DOWN;
						}
					} else {
						if (debug)
							mpfprint("!!! received %d chars from socket for input %d !!!\n", rec, i);
					}
				}
			}
		}

		/* read all 8 digital ports */
		if (read(fw, &di, 1) == ERROR) {
			mpfprint("error reading firmware port: %s (%d)\n", strerror(errno), errno);
			return;
		}

		for (i=1; i<=8; i++) {

/*  			if (i == 3) { */
/*  				if (last.tv_sec == 0) */
/*  					gettimeofday(&last, NULL); */
/*  				else { */
/*  					gettimeofday(&timenow, NULL); */
/*  					if ((reg = (((timenow.tv_sec - last.tv_sec) * 1000000 + */
/*  								 timenow.tv_usec - last.tv_usec) / 1000)) > lastreg) { */
/*  						printf("################################# interval: %d ms\n", reg); */
/*  						lastreg = reg; */
/*  					} else { */
/*  						if (corr++ >= 100) { */
/*  							lastreg = 0; */
/*  							corr = 0; */
/*  						} */
/*  					} */
/*  					gettimeofday(&last, NULL); */
/*  				}				 */
/*  			} */

			switch (remote_conn[i]) {
			case CONNECTING:
				goto connecting;
			case NET_DOWN:
				//is there another input port already on WAIT_RECOVER?
				for (j=1; j<=8; j++) {
					if (remote_conn[j] == WAIT_RECOVER) {
						char mydestip[20];
						strcpy(mydestip, inet_ntoa(config.destip[i]));
						//is the IP address the same as for this input port?
						if (!strcmp(mydestip, inet_ntoa(config.destip[j]))) {
							//don't need to fork
							if (debug)
								mpfprint("input port %d going directly to WAIT_RECOVER state (pid:%d)\n", i, recover_pid[j]);
							remote_conn[i] = WAIT_RECOVER;
							recover_pid[i] = recover_pid[j];
							break;
						}
					}
				}
				if (remote_conn[i] == WAIT_RECOVER) continue;
				//shouldn't be a Peer Monitor started...
				if (peerMon_pid[i]) {
					kill(peerMon_pid[i], SIGKILL);
					if (debug)
						mpfprint("input port %d stopped Peer Monitor (pid:%d)\n", i, peerMon_pid[i]);
				}
				if ((recover_pid[i] = fork())) {
					if (debug)
						mpfprint("input port %d going to WAIT_RECOVER state (pid:%d)\n", i, recover_pid[i]);
					remote_conn[i] = WAIT_RECOVER;
					continue;
				} else {
					// The recovery code...
					char cmd[40];
					changeTitle(argc, argv, "purecover");
					for (j=1; j<=8; j++)
						if (fdsock[j] != NO_SOCK)
							close(fdsock[j]);
					sprintf(cmd, "ping -c 3 %s > /dev/null", inet_ntoa(config.destip[i]));
					while (1) {
						sleep(5);
						if (!system(cmd))
							exit(0);
					}
				}
			case WAIT_RECOVER:
				if (kill(recover_pid[i], 0) == ERROR) {
					//network is recovered
					if (debug)
						mpfprint("WAIT_RECOVER finished for input port %d (pid:%d)\n", i, recover_pid[i]);
					if (peerMon_pid[i]) {
						if (debug)
							mpfprint("stopping remote signalling for Digital Input #%d\n", i);
						if (close(fdsock[i]) == ERROR) {
							exitstatus = ERRCLOSE;
							return;
						}
						fdsock[i] = NO_SOCK;
						peerMon_pid[i] = 0;
					}
					remote_conn[i] = PUCONN_OFF;
				}
				continue;
			case GOING_ON:
				goto make_tcpconn;
			case GOING_OFF:
				goto drop_tcpconn;
			case PUCONN_ON:
				//is there another input port on PUCONN_ON?
				for (j=1; j<=8; j++) {
					if ((remote_conn[j] == PUCONN_ON) && (j != i)) {
						char mydestip[20];
						strcpy(mydestip, inet_ntoa(config.destip[i]));
						//is the IP address the same as for this input port?
						if (!strcmp(mydestip, inet_ntoa(config.destip[j]))) {
							//don't need to fork
							if (!peerMon_pid[i] && peerMon_pid[j]) {
								peerMon_pid[i] = peerMon_pid[j];
								if (debug)
									mpfprint("input port %d using Peer Monitor (pid:%d)\n", i, peerMon_pid[i]);
							}
							break;
						}
					}
				}
				if (!peerMon_pid[i]) {
					if ((peerMon_pid[i] = fork())) {
						if (debug)
							mpfprint("input port %d started Peer Monitor (pid:%d)\n", i, peerMon_pid[i]);
						continue;
					} else {
						// The peer monitoring code...
						char cmd[40];
						changeTitle(argc, argv, "peermntor");
						for (j=1; j<=8; j++)
							if (fdsock[j] != NO_SOCK)
								close(fdsock[j]);
						sprintf(cmd, "ping -c 3 %s > /dev/null", inet_ntoa(config.destip[i]));
						while (1) {
							sleep(5);
							if (system(cmd))
								exit(0);
						}
					}
				} else {
					if (kill(peerMon_pid[i], 0) == ERROR) {
						//network is down
						if (debug)
							mpfprint("network down detected by Peer Monitor for input port %d (pid:%d)\n", i, peerMon_pid[i]);
						remote_conn[i] = NET_DOWN;
						continue;
					}
				}
			}

			j = 0;
			if (di & (1<<(i-1))) j = 1;
			if (j ^ config.dref[i]) { //check input pin
			/*** state differs from expected reference ***/

				if (remote_conn[i] == PUCONN_OFF)
					goto make_tcpconn;

			} else {
			/*** state is the expected reference ***/

				if (remote_conn[i] == PUCONN_ON)
					goto drop_tcpconn;

			}
			continue;

make_tcpconn:
			if (remote_conn[i] != GOING_ON) {
				gettimeofday(&debounce_start[i], NULL); //start debouncing
				remote_conn[i] = GOING_ON;
				continue;
			}
			if (debounce_complete(i)) {
connecting:
				j = 0;
				if (di & (1<<(i-1))) j = 1;
				if (j ^ config.dref[i]) { //check if input still differs from reference
					if (connect(fdsock[i],(struct sockaddr*)&sock[i], sizeof(struct sockaddr_in)) == ERROR) {
						if (errno != EINPROGRESS) {
							if (debug)						
								mpfprint("error connecting socket for input %d: %s (%d)\n", i, strerror(errno), errno);
							if (close(fdsock[i]) == ERROR) {
								exitstatus = ERRCLOSE;
								return;
							}
							fdsock[i] = NO_SOCK;

							remote_conn[i] = NET_DOWN;
						} else {
							remote_conn[i] = CONNECTING;
						}
					} else {
						if (debug)
							mpfprint("starting remote signalling for Digital Input #%d\n", i);

						remote_conn[i] = PUCONN_ON;
					}
				} else {
					if (remote_conn[i] == CONNECTING) {
						if (close(fdsock[i]) == ERROR) {
							exitstatus = ERRCLOSE;
							return;
						}
						fdsock[i] = NO_SOCK;
					}
					remote_conn[i] = PUCONN_OFF;
				}
			}
			continue;

drop_tcpconn:
			if (remote_conn[i] != GOING_OFF) {
				gettimeofday(&debounce_start[i], NULL); //start debouncing
				remote_conn[i] = GOING_OFF;
				continue;
			}
			if (debounce_complete(i)) {
				j = 0;
				if (di & (1<<(i-1))) j = 1;
				if (!(j ^ config.dref[i])) { //check if input is still the same as reference
					if (debug)
						mpfprint("stopping remote signalling for Digital Input #%d\n", i);
					if (close(fdsock[i]) == ERROR) {
						exitstatus = ERRCLOSE;
						return;
					}
					fdsock[i] = NO_SOCK;
					for (j=1; j<=8; j++) {
						if ((peerMon_pid[j] == peerMon_pid[i]) && (j != i)) {
							peerMon_pid[i] = 0;
							break;
						}
					}
					if (peerMon_pid[i]) {
						kill(peerMon_pid[i], SIGKILL);
						if (debug)
							mpfprint("input port %d stopped Peer Monitor (pid:%d)\n\n", i, peerMon_pid[i]);
						peerMon_pid[i] = 0;
					} else {
						if (debug)
							mpfprint("\n");
					}
					remote_conn[i] = PUCONN_OFF;
				} else {
					remote_conn[i] = PUCONN_ON;
				}
			}

		} /* for() */

	} /* while() */

} /* puconn_mainloop */


/******************************/
int  main(int argc, char **argv)   // puconnect application entry point
/******************************/
 {
	FILE   *file;
	char    buffer[1024], par[20], value[20];
	int     port, lineno=0, i, opt;
	puconf  *cp = &config;


	// check machine type and prepare to catch some signals
	get_machine_type(&mt);
	if (mt.board_type != BOARD_TS110) {
		printf("This application only runs on TS110 hardware...\n");
		exit(ERROR);
	}
	prepare_signals();


	// take command line arguments
	while ((opt = getopt(argc, argv, "xd:gh-")) != EOF) {
		switch (opt) {
		case 'x':
			terminator = TRUE;
			system("killall puconnect");
			break;
		case 'd':
			if (!optarg) exit(ERROR);
			debounce = atoi(optarg);
			break;
		case 'g':
			debug = TRUE;
			break;
		default:
			printf("\nThis application is intended to monitor devices that\n"
				   "use a \"dry contact\" to communicate its state.\n"
				   "This is achieved by connecting TS110's digital inputs to\n"
				   "those contacts so a TCP connection to a remote server is\n"
				   "done when the reference value for any digital input is no\n"
				   "longer verified. The remote DTR then indicates the status change.\n"
				   "See \"/etc/puconnect.conf\" for datails on configuration.\n"
                   "\nUSAGE: \"puconnect [-x] [-d <interval>] [-g] [-h]\", where:\n\n"
				   "-x: terminates the application\n"
				   "-d: changes default contacts debouncing interval (miliseconds)\n"
				   "-g: generates a log file on \"/var/log\" directory\n"
				   "-h: show this message\n\n");
			exit(OK);
		}
	}


	// check there is no other instance running
	system("ps | grep puconnect | grep -v grep | grep -v conf > /tmp/puconn");
	if ((file = fopen("/tmp/puconn", "r")) == NULL) {
		printf("Could not open temporary file.\n");
		exit(ERROR);                                                           
	}
	if (fgets(buffer, sizeof(buffer), file) && fgets(buffer, sizeof(buffer), file)) {
		//at least two lines there...
		printf("puconnect is alreay running; use \"puconnect -x\" to terminate it...\n");
		exit(ERROR);                                                           
	}
	fclose(file);
	if (debug)
		mpfprint("\npuconnect(v%d)\n", version);


	// set some default values
	for (i=2, cp->destport[1]=FIRSTDESTPORT; i<=8; i++)
		cp->destport[i] = cp->destport[i-1] + 1;
	for (i=1; i<=8; i++)
		fdsock[i] = NO_SOCK;


	// read configuration
	if ((file = fopen(PUCONFFILE, "r")) == NULL) {
		printf("Could not open %s file.\n", PUCONFFILE);
		goto end_puconn;                                                           
	}
	while (fgets(buffer, sizeof(buffer), file)) {
		lineno++;
		if (buffer[0] == '#' || buffer[0] == '\n') continue;
		if (buffer[0] == 'D' ) {
			sscanf(buffer,"D%d.%s %s",&port,par,value);
			if(strcmp(par,"destipaddr")==0) {
				if (!inet_aton(value,&(cp->destip[port]))) {
					printf("Wrong IP format on line %d.\n", lineno);
					goto end_puconn;
				}
				continue;
			}
			else if(strcmp(par,"destport")==0) {
				cp->destport[port]=atoi(value);
				continue;
			}
			else if(strcmp(par,"ref")==0) {
				cp->dref[port]=atoi(value);
				if(cp->dref[port]!=0 && cp->dref[port]!=1) {
					printf("Invalid ref value on line %d.\n", lineno);
					goto end_puconn;
				}
				continue;
			}
		}
		printf("Syntax error on line %d.\n", lineno);
		goto end_puconn;
	}
	fclose(file);
	for (i=1; i<=8; i++) {
		if (!cp->destip[i].s_addr) {
			printf("Destination IP addresses must be specified for all digital inputs.\n");
			goto end_puconn;
		}
	}
	// info printing
	if (debug) {
		mpfprint("debouncing period is %d\n", debounce);
		mpfprint("input #1: dref=%d  destipaddr=%s  destport=%d\n", 
				 cp->dref[1], inet_ntoa(cp->destip[1]), cp->destport[1]);
		mpfprint("input #2: dref=%d  destipaddr=%s  destport=%d\n", 
				 cp->dref[2], inet_ntoa(cp->destip[2]), cp->destport[2]);
		mpfprint("input #3: dref=%d  destipaddr=%s  destport=%d\n", 
				 cp->dref[3], inet_ntoa(cp->destip[3]), cp->destport[3]);
		mpfprint("input #4: dref=%d  destipaddr=%s  destport=%d\n", 
				 cp->dref[4], inet_ntoa(cp->destip[4]), cp->destport[4]);
		mpfprint("input #5: dref=%d  destipaddr=%s  destport=%d\n", 
				 cp->dref[5], inet_ntoa(cp->destip[5]), cp->destport[5]);
		mpfprint("input #6: dref=%d  destipaddr=%s  destport=%d\n", 
				 cp->dref[6], inet_ntoa(cp->destip[6]), cp->destport[6]);
		mpfprint("input #7: dref=%d  destipaddr=%s  destport=%d\n", 
				 cp->dref[7], inet_ntoa(cp->destip[7]), cp->destport[7]);
		mpfprint("input #8: dref=%d  destipaddr=%s  destport=%d\n", 
				 cp->dref[8], inet_ntoa(cp->destip[8]), cp->destport[8]);
	}


	if (fork()) {

		exit(OK); // return to command line

	} else {

		//open firmware port
		if ((fw = open("/dev/gpio", O_RDWR|O_NONBLOCK)) == ERROR) {
			mpfprint("error opening firmware port: %s (%d)\n", strerror(errno), errno);
			goto end_puconn;
		}

		puconn_mainloop(argc, argv);

	}


end_puconn:
	if (debug) mpfprint(".......................... exit status: %d\n\n", exitstatus);


	exit(exitstatus);

} /* main() */


/*********************************************/
/**********   Print to a log file   **********/
/*********************************************/
void mpfprint(const char *fmt, ...)
{
	va_list ap;
	char buf[1024];
	char filen[] = "/var/log/puconnect.log";
//	char filen[] = "/root/puconnect.log";

	if (ffd < 0) {
		if ((ffd = open(filen, O_WRONLY | O_CREAT | O_TRUNC, 0)) < 0) {
			printf("!!! error opening log file: %s !!!\r\n", strerror(errno));
			return;
		}
	}

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	write(ffd, buf, strlen(buf));
	va_end(ap);
}
/**********************************************/


/* End of puconnect.c */
